/*	@(#)AR.c	1.5 11/17/92
 *	Author:	Roger A. Cole
 *	Date:	03-05-92
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1992, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 *  .01 03-05-92 rac	initial version
 *  .02 03-09-92 rac	continued development
 *  .03 08-25-92 rac	continued development
 *  .04 09-30-92 rac	print time stamps for interval test; add handling
 *			for status channels;
 *  .05 11-17-92 rac	make output to status channels more robust
 * AR.c,v
 * Revision 1.2  1994/05/06 19:05:14  mrk
 * conversion from sccs to cvs
 *
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	AR - mouse and menu archiver
*
* DESCRIPTION
*
*	% AR options
*
*   options:		(options starting with -D allow changing defaults;
*			the "canned" default is shown in ()'s )
*	-Dret dir	default directory for retrieval files (./arReq)
*	-I file		specifies a file with initialization commands
*	-R reqFile	specifies an archive request file to be opened
*	-E reqFile	specifies an archive request file to be edited
*
*	initialization file:
*		MAIN ReqEdit		add a request edit btn on main
*		MAIN NEWLINE		starts new line of buttons
*		MAIN X pos		sets x position on current line
*		MAIN STRING string	places "string" at current position
*		EDIT Glossary		add a glossary button on edit window
*		DIR directory		default directory
*		Dret dir		directory for retrieval files
*		REQ reqFile		archive request file to be opened
*		GLOSSARY directory file	place glossary resides
*		...
*
* BUGS
* o	doesn't handle REQ_STATUS and SAMP_STATUS channels
* o	doesn't give a good variety of menu items for status
* o	doesn't have an exit handler to wrapup `by channel' operations
* o	`allow partial samples' crashes on L-net for snap1c.req
*
* WISH LIST
* o	a log file and log window which will show `what's happened'
*
*-***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <varargs.h>
#include <sys/file.h>	/* for O_RDWR */

#define XVIEW
#if defined XVIEW
#   include <xview/xview.h>
#   include <xview/frame.h>
#   include <xview/openmenu.h>
#   include <xview/panel.h>
#   include <xview/textsw.h>
#endif

#include <genDefs.h>
#include <arAccessLib.h>
#include <guiSubr.h>
#include <sydDefs.h>
#include <cadef.h>
#include <reqSubr.h>
#include <sydPlotDefs.h>
#include <PL.h>
#include <PR.h>
#include <enumShared.h>

void arQuit();
void arReqCheckConn();
void arReqCheckAct();
void arReqCheckInact();
void arReqCheckSample();
void arReqClose_mi();
void arReqClose();
void arReqEdit_mi();
void arReqEdit();
void arReqOpen();
void arReqMon_mi();
void arReqMonStop_mi();
void arReqSampleMonStart();
void arReqSampleMonHandler();
void arReqReadFile();
void arReqSnap_mi();
void arReqSnapStop_mi();
void arReqSampleSnapStart();
void arReqStatus_mi();
void arReqStatus();
void arTimer();
void edit();
void editExtras();
void editValPuts();
void editOpen_pb();
void editValSyntax_pb();
void editValConn_pb();
void editValInterval_pb();
void edit_pb();
RAW_REQ *editValGetReq();
void editValidateInitCF();
void glossary_pb();
void quit_pb();


typedef enum {
    ADD_TY_BTN=1,
    ADD_TY_NEWLINE,
    ADD_TY_X,
    ADD_TY_REQ,
    ADD_TY_EDIT,
    ADD_TY_STR,
} ADD_TY;

typedef struct {
    ADD_TY	type;
    char	*keyword;
    char	*label;
    void	(*fn)();
} ADD_CANNED;

#define ADD_CANNED_DIM 2

static ADD_CANNED configAddItems[ADD_CANNED_DIM]={
    ADD_TY_BTN, "ReqEdit", "Archive requests...", edit_pb,
    ADD_TY_BTN, "Glossary", "Channel glossary...", glossary_pb,
};

typedef struct addOn {
    struct addOn *pNext;
    struct addOn *pPrev;
    ADD_TY	type;
    void	*item;
    char	*place;
    int		intVal;
    char	strVal[GUI_TDIM];
} ADDON;

typedef enum {
    REQ_ST_CONN=1,
    REQ_ST_NOT_CONN,
    REQ_ST_ENABLED,
    REQ_ST_ACTIVE,
    REQ_ST_INACTIVE,
} REQ_ST;

/*-----------------------------------------------------------------------------
*    enum for status codes to arReqStatus
*----------------------------------------------------------------------------*/
typedef enum {
    AR_INF_STATE,		/* show request state */
    AR_INF_CONN,		/* show connection information */
    AR_INF_SAMP,		/* show sample state */
    AR_INF_STEP,		/* show `step a channel' state */
} AR_INF;

typedef struct arChan {
    struct arChan *pNext;	/* pointer to next channel */
    struct arChan *pPrev;	/* pointer to previous channel */
    AR_CHAN_DESC *pCfChan;	/* pointer to `bcf' channel descriptor */
    SYD_CHAN	*pSChan;	/* pointer to syd channel descriptor */
    struct arReq *pReq;		/* pointer to archiver request for channel */
    short	getGR;		/* 1 if need to get graphics info */
} AR_CHAN;

typedef struct arReq {
    struct arReq *pNext;
    struct arReq *pPrev;
    struct arCtx *pCtx;
    PR_CTX	*pPrCtx;	/* pointer to data printing context */
    PL_CTX	*pPlCtx;	/* pointer to data plotting context */
    char	name[db_name_dim];/* name of archive request */
    char	file[120];	/* file name for request */
    char	dir[120];	/* directory for request */
    REQ_ST	status;		/* status of the request */
    short	nDis;		/* number of channels disconnected */
    short	nCtrlDis;	/* number of control channels disconnected */
    short	sampling;	/* 0 if no sampling in progress;
				   1 if actual sampling is in progress;
				   -1 if pre-trigger queueing is in progress */
    Panel	panel;		/* panel for this request */
    Panel_item	status_PM;	/* status panel message item */
    SYD_SPEC	*pSspec;
    REQ_TRIG	actTrig;	/* activation trigger */
    short	allowDiscon;	/* allow disconnected channels */
    REQ_TRIG	inactTrig;	/* inactivation trigger */
    REQ_SAMPLE	sample;		/* sample method structure */
    AR_CF_DESC	*pCfDesc;	/* by channel file handle */
    AR_CHAN	*pChanHead;	/* head of channel list */
    AR_CHAN	*pChanTail;	/* tail of channel list */
    REQ_STAT	reqStat;	/* channel for writing archive request status */
    REQ_STAT	sampleStat;	/* channel for writing sampling status */
    char	lockPath[2*GUI_TDIM+8];	/* lock on open */
    short	getGR;		/* 1 if need to get graphics info */
    TS_STAMP	firstTs;	/* time of first data, while monitoring */
    TS_STAMP	lastTs;		/* time of last data, while monitoring */
    unsigned long monitorCount;	/* count of values, while monitoring */
    REQ_STEP	step;		/* `step a channel' structure */
} AR_REQ;

typedef struct arCtx {
    GUI_CTX	guiCtx;
    GUI_EDIT	*pEditor;	/* pointer to editor for archive requests */
    GUI_EDIT	*pGlossary;	/* pointer to editor for channel glossary */
    GUI_PRT	*pGuiPrt;	/* pointer to "printer" from guiSubr */
    ADDON	*pAddonHead;	/* list of items on main panel */
    ADDON	*pAddonTail;
    AR_REQ	*pReqHead;	/* list of opened archive requests */
    AR_REQ	*pReqTail;
    char	Req_subdir[GUI_TDIM];/* default request file sub-dir */
    char	*config;	/* configuration file (from -I) */
    Panel	panel;		/* main panel */
    int		x,y,ht;		/* main panel item position information */
    char	glossDir[GUI_TDIM];/* directory for channel glossary */
    char	glossFile[GUI_TDIM];/* file name for channel glossary */

    Frame	validate_CF;	/* command frame for req validation */
    Textsw	validate_TSW;	/* text subwindow for validation messages */
    int		validatePos;	/* position in text subwindow */
    Panel_item	interval_PT;
    Panel_item	jitter_PT;
    Panel_item	testTime_PT;
} AR_CTX;

/*-----------------------------------------------------------------------------
* globals
*----------------------------------------------------------------------------*/
static TS_STAMP intervalTs;
static double intervalTestTime;
static RAW_REQ	*interval_pRaw=NULL;

static TS_STAMP connectTs;
static double connectTestTime;
static RAW_REQ	*connect_pRaw=NULL;

static char	nowText[28];

#define REQ_PRINTF guiMessagePrintf(pReq->status_PM,
#define NOW arStampToText_hh_mm_ss(&nowTs, nowText)

/*+/subr**********************************************************************
* NAME	arStampToText_hh_mm_ss
*
*-*/
char *
arStampToText_hh_mm_ss(pStamp, text)
TS_STAMP *pStamp;
char	*text;
{
    char	tempText[28];
    tsStampToText(pStamp, TS_TEXT_MMDDYY, tempText);
    tempText[17] = '\0';
    strcpy(text, &tempText[9]);
    return text;
}


/*+/subr**********************************************************************
* NAME	main
*
*-*/
main(argc, argv)
int	argc;
char	*argv[];
{
    Panel	panel;
    static unsigned short icon_bits[]={
#include "AR.icon"
    };
    AR_CTX	arCtx, *pCtx=&arCtx;

    int		i;
    FILE	*pFile;
    char	dir[GUI_TDIM];

    bzero((char *)pCtx, sizeof(arCtx));
    pCtx->pAddonHead = NULL;
    pCtx->pAddonTail = NULL;
    pCtx->pReqHead = NULL;
    pCtx->pReqTail = NULL;
    strcpy(pCtx->Req_subdir, "arReq");
    pCtx->pEditor = NULL;
    pCtx->pGlossary = NULL;
    strcpy(pCtx->glossDir, "../log");
    strcpy(pCtx->glossFile, "channel.glossary");
    pCtx->config = NULL;
    pCtx->validate_CF = NULL;
    pCtx->validate_TSW = NULL;
    configAddCannedItem(pCtx, "Glossary", "EDIT");
    configAddCannedItem(pCtx, "ReqEdit", "MAIN");

    panel = guiInit(&pCtx->guiCtx, &argc, argv, "AR", 100, 100, 0,0);

/*-----------------------------------------------------------------------------
*    handle command line options
*----------------------------------------------------------------------------*/
    for (i=1; i<argc; i++) {
	if (strcmp(argv[i], "-Dret") == 0)
	    strcpy(pCtx->Req_subdir, argv[++i]);
	else if (strcmp(argv[i], "-I") == 0)
	    pCtx->config = argv[++i];
	else if (strcmp(argv[i], "-E") == 0)
	    configAddEdit(pCtx, argv[++i], "REQ_EDIT");
	else if (strcmp(argv[i], "-R") == 0)
	    configAddReq(pCtx, argv[++i], "REQ");
	else {
	    printf("invalid command line keyword: %s", argv[i]);
	    exit(1);
	}
    }

    if (pCtx->config != NULL) {
	if ((pFile = fopen(pCtx->config, "r")) == NULL) {
	    printf("couldn't open input config file: %s", pCtx->config);
	    exit(1);
	}
	else {
	    configAddonsClear(pCtx);	/* clear out default panel items */
	    configParse(pFile, pCtx);
	    fclose(pFile);
	}
    }

    pCtx->panel = panel;
    guiIconCreate(pCtx->guiCtx.baseFrame, icon_bits);

    editValidateInitCF(pCtx);

    pCtx->pGuiPrt = guiPrinter(&pCtx->guiCtx,
		pCtx->guiCtx.baseFrame, "AR: printer setup", 1);
    sprintf(dir, "%s/%s", pCtx->guiCtx.pwd, pCtx->Req_subdir);
    pCtx->pEditor = guiEditor(&pCtx->guiCtx, "AR: edit archive requests",
			dir, NULL, pCtx->pGuiPrt, editExtras, pCtx);

    pCtx->y = 10; pCtx->x = 5; pCtx->ht = 0;

    guiButton("Quit", panel,
		&pCtx->x,&pCtx->y,&pCtx->ht, quit_pb, NULL, ARCTX,pCtx,0,NULL);

    /* actually realize the panel items */
    configAddRealize(pCtx, panel, &pCtx->x,&pCtx->y,&pCtx->ht, "MAIN");
    pCtx->y += pCtx->ht + 10; pCtx->x = 5; pCtx->ht = 0;

    window_fit_height(panel);
    window_fit(pCtx->guiCtx.baseFrame);
    configAddRealize(pCtx, panel, &pCtx->x,&pCtx->y,&pCtx->ht, "REQ");
    configAddRealize(pCtx, panel, &pCtx->x,&pCtx->y,&pCtx->ht, "REQ_EDIT");

    guiTimer(.1, arTimer, pCtx);

    xv_main_loop(pCtx->guiCtx.baseFrame);
    arQuit(pCtx, 0);

    return 0;
}

/*+/subr**********************************************************************
* NAME	arTimer - handle notifier timer calls
*
*-*/
void
arTimer(pCtx, dummy)
AR_CTX	*pCtx;
int     dummy;
{
    AR_REQ	*pReq;
    SYD_CHAN	*pChan;
    int		nCon, nDis, nCtrlDis;
    TS_STAMP	nowTs;
    double	diff;
    static TS_STAMP last1secTs={0,0}, last15secTs={0,0};
    long	stat;

    ca_pend_event(.001);
    tsLocalTime(&nowTs);
    TsDiffAsDouble(&diff, &nowTs, &last1secTs);
    if (diff >= 1.) {
/*-----------------------------------------------------------------------------
*    1 second processes
*    o  check channel connections, and update status on changes
*    o  check to activate request
*    o  check to inactivate request
*----------------------------------------------------------------------------*/
	last1secTs = nowTs;
	for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	    arReqCheckConn(pReq, &nCon, &nDis, &nCtrlDis, 0);
	    if (nDis != pReq->nDis) {
		if (nDis > 0) {
		    REQ_PRINTF "connects pending: total=%d, control=%d",
				nDis, nCtrlDis);
		    reqStatPrintf(&pReq->reqStat, "%d conn pend", nDis);
		}
		else {
		    REQ_PRINTF "connections are complete");
		    reqStatPrintf(&pReq->reqStat, "conn done");
		}
		if (nDis == 0 || (nCtrlDis == 0 && pReq->allowDiscon)) {
		    pReq->status = REQ_ST_CONN;
		    reqTrigArm(&pReq->actTrig, nowTs);
		}
		pReq->nDis = nDis;
		pReq->nCtrlDis = nCtrlDis;
	    }
	    arReqCheckAct(pReq, nowTs);
	    arReqCheckInact(pReq, nowTs);
	}
	TsDiffAsDouble(&diff, &nowTs, &last15secTs);
	if (diff >= 15.) {
/*-----------------------------------------------------------------------------
*    15 second processes
*    o  if awaiting graphics information for `by channel' output, check for
*	its arrival and store it when available
*    o  sync `by channel' files
*----------------------------------------------------------------------------*/
	    last15secTs = nowTs;
	    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
		if (pReq->getGR)
		    arReqCheckGraphics(pReq);
		if (pReq->sample.outputType == REQ_OUT_BCF) {
		    stat = arCFSyncWrite(pReq->pCfDesc);
		    assert(stat == OK);
		}
	    }
	}
    }
/*-----------------------------------------------------------------------------
*    .1 second processes
*    o  ca_pend_event		(called above)
*    o  check for sampling activities for request (if the request has
*	inactivation type of REQ_TYPE_DONE and sampling is complete, then
*	this check is skipped)
*    o  check for `step a channel' activities.  Most of the work is done
*	here, but SnapStart and MonStart help with detecting when sampling
*	actually starts (for TRIGGER: after each step).
*    o  check Validate activities:
*	"connect check"
*	"interval check"
*----------------------------------------------------------------------------*/
    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	if (pReq->status == REQ_ST_ACTIVE && (pReq->inactTrig.doneFlag == 0 ||
				pReq->inactTrig.type != REQ_TYPE_DONE)) {
	    arReqCheckSample(pReq, nowTs);
	}
	if (pReq->step.active) {
	    if (pReq->step.sampleActive) {
		if (!pReq->sampling)
		    pReq->step.sampleActive = 0, pReq->step.sampleDone = 1;
	    }
	    if (reqStep_stepDoneCheck(&pReq->step, &nowTs)) {
		if (pReq->sample.trigger.type == REQ_TYPE_AFTER_STEP &&
						pReq->step.sampleDone == 0) {
		    pReq->sample.trigger.demand = 1;
		    pReq->step.samplePending = 1;
		    if (pReq->status == REQ_ST_ACTIVE &&
				(pReq->inactTrig.doneFlag == 0 ||
				pReq->inactTrig.type != REQ_TYPE_DONE)) {
			arReqCheckSample(pReq, nowTs);
		    }
		}
		else if (!pReq->step.complete) {
		    reqStep_stepNext(&pReq->step);
		    pReq->step.sampleDone = 0;
		}
		reqStep_seqDoneCheck(&pReq->step);
	    }
	}
    }
    if (connect_pRaw != NULL) {
	TsDiffAsDouble(&diff, &nowTs, &connectTs);
	if (diff >= connectTestTime) {
	    rawReqCheckConnect_wrap(connect_pRaw, editValPuts, pCtx);
	    rawReqFree(connect_pRaw);
	    connect_pRaw = NULL;
	}
    }
    if (interval_pRaw != NULL) {
	TsDiffAsDouble(&diff, &nowTs, &intervalTs);
	if (diff >= intervalTestTime) {
	    rawReqCheckInterval_wrap(interval_pRaw, editValPuts, pCtx);
	    rawReqFree(interval_pRaw);
	    interval_pRaw = NULL;
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckAct - check for need to activate a request
*
* DESCRIPTION
*	Checks to see if it's OK to activate a request.  If it is OK, and
*	if the activation condition is met, the request is activated.
*
*	There are several special cases which keep a request from activating:
*	o   the request is waiting to receive graphics information to be
*	    stored in a `by channel' archive file
*	o   all the channels in the request aren't connected; this includes
*	    trigger channels
*	o   an event trigger for a DBF_ENUM channel is waiting to receive
*	    graphics information (to get the state strings)
*	o   an event trigger for a DBF_ENUM channel has a comparison value
*	    which isn't one of the state strings for the channel
*
* BUGS
* o	doesn't put an "activated" item in `by channel' file
*
*-*/
void
arReqCheckAct(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    AR_CHAN	*pRChan;

    if (pReq->getGR) {
	arReqCheckGraphics(pReq);
	if (pReq->getGR)
	    return;
    }
    if (pReq->status == REQ_ST_CONN || pReq->status == REQ_ST_INACTIVE) {
	if (pReq->actTrig.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->actTrig) != 0)
		return;
	}
	if (pReq->inactTrig.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->inactTrig) != 0)
		return;
	}
	if (pReq->sample.forEvent.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->sample.forEvent) != 0)
		return;
	}
	if (pReq->sample.trigger.enumGetGR != 0) {
	    if (arReqCheckEnum(pReq, &pReq->sample.trigger) != 0)
		return;
	}
	if (reqTrigTest(&pReq->actTrig, nowTs)) {
	    pReq->status = REQ_ST_ACTIVE;
	    REQ_PRINTF "%s  activated", NOW);
	    reqStatPrintf(&pReq->reqStat, "activated");
	    reqStatPrintf(&pReq->sampleStat, "not sampling");
	    reqTrigArm(&pReq->inactTrig, nowTs);
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckConn - check connections for a request's channels
*
*-*/
void
arReqCheckConn(pReq, pNcon, pNdiscon, pNCtrlDiscon, printFlag)
AR_REQ	*pReq;
int	*pNcon, *pNdiscon, *pNCtrlDiscon, printFlag;
{
    SYD_CHAN	*pChan;

    *pNcon=0;
    *pNdiscon=0;
    *pNCtrlDiscon=0;
    for (pChan=pReq->pSspec->pChanHead; pChan!= NULL; pChan=pChan->pNext){
	if (pChan->discon) {
	    (*pNdiscon)++;
	    if (printFlag)
		printf("      %s\n", pChan->name);
	}
	else
	    (*pNcon)++;
    }
    if (pReq->actTrig.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->actTrig.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->actTrig.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->inactTrig.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->inactTrig.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->inactTrig.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->sample.forEvent.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->sample.forEvent.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->sample.trigger.type == REQ_TYPE_EVENT) {
	if (ca_field_type(pReq->sample.trigger.event_pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->sample.trigger.event_pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->step.type == REQ_TYPE_STEP) {
	if (ca_field_type(pReq->step.pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", ca_name(pReq->step.pCh));
	}
	else
	    (*pNcon)++;
    }
    if (pReq->reqStat.pCh != NULL) {
	if (ca_field_type(pReq->reqStat.pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", pReq->reqStat.name);
	}
	else
	    (*pNcon)++;
	
    }
    if (pReq->sampleStat.pCh != NULL) {
	if (ca_field_type(pReq->sampleStat.pCh)==TYPENOTCONN) {
	    (*pNdiscon)++;
	    (*pNCtrlDiscon)++;
	    if (printFlag)
		printf("      %s\n", pReq->sampleStat.name);
	}
	else
	    (*pNcon)++;
	
    }
}

/*+/internal******************************************************************
* NAME	arReqCheckEnum - check enumerated trigger channel
*
* DESCRIPTION
*	Checks to see if graphics info for enumerated channels has been
*	received.  If so, checks to see that the comparison value is legal.
*
* RETURNS
*	0	if comparison value is OK, or
*	-1	if graphics info hasn't yet been received, or
*	-2	if comparison value is illegal
*
* BUGS
* o	failure should set an invalid state or disable the request;
*	presently, the only action is to give a notice and let the
*	operator do "whatever", including nothing.
*
*-*/
arReqCheckEnum(pReq, pTrig)
AR_REQ	*pReq;
REQ_TRIG *pTrig;
{
    AR_CTX	*pCtx=pReq->pCtx;
    int		i;

    assert(pTrig->enumGetGR != 0);
    if (pTrig->enumGetGR != 2)
	return -1;
    pTrig->enumGetGR = 0;
    for (i=0; i<pTrig->enumGr.no_str; i++) {
	if (strcmp(pTrig->eventValCmp.strVal, pTrig->enumGr.strs[i]) == 0)
	    return 0;
    }
    guiNoticePrintf(&pCtx->guiCtx,
"condition test is invalid; \"%s\" not a legal value for \"%s\"",
pTrig->eventValCmp.strVal, pTrig->eventName);
    return -2;
}

/*+/internal******************************************************************
* NAME	arReqCheckGraphics
*
* DESCRIPTION
*    if graphics information was requested for a channel being written to
*    a `by channel' file, and if the graphics information has arrived,
*    it is written to the file.
*
*-*/
arReqCheckGraphics(pReq)
AR_REQ	*pReq;
{
    AR_CHAN	*pRChan;
    struct event_handler_args arg;
    long	stat;

    for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext) {
	if (pRChan->getGR && pRChan->pSChan->gotGr) {
	    arg.dbr = &pRChan->pSChan->grBuf;
	    arg.count = pRChan->pSChan->elCount;
	    arg.chid = (chid)pRChan->pSChan->pHandle;
	    arg.type = pRChan->pSChan->dbrGrType;
	    stat = arCFChanWriteGR(pRChan->pCfChan, &arg);
	    assert(stat == OK);
	    pRChan->getGR = 0;
	    pReq->getGR--;
	}
    }
}

/*+/subr**********************************************************************
* NAME	arReqCheckInact - check for need to inactivate a request
*
* DESCRIPTION
*	o   check for inactivation condition
*
*-*/
void
arReqCheckInact(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    int		doIt=0;

    if (pReq->status != REQ_ST_ACTIVE)
	return;
    if (pReq->nCtrlDis != 0 || (pReq->allowDiscon == 0 && pReq->nDis != 0)) {
	REQ_PRINTF "%s  inactivated; %d broken connections", NOW, pReq->nDis);
	reqStatPrintf(&pReq->reqStat, "inactivated--disCon");
	doIt = 1;
	pReq->status = REQ_ST_NOT_CONN;
    }
    else {
	if (reqTrigTest(&pReq->inactTrig, nowTs)) {
	    REQ_PRINTF "%s  inactivated", NOW);
	    reqStatPrintf(&pReq->reqStat, "inactivated");
	    doIt = 1;
	    pReq->status = REQ_ST_INACTIVE;
	    reqTrigArm(&pReq->actTrig, nowTs);
	}
    }
    if (doIt == 0)
	return;
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);
}

/*+/subr**********************************************************************
* NAME	arReqCheckSample - check sampling activities
*
* DESCRIPTION
*	o   check for sample trigger
*	o   check for sample condition
*	o   check for "until" condition
*	o   check for need to handle "before trigger" condition
*
* BUGS
* o	for "n samples before trigger", actual sample can start as much
*	as 10 samples prior to desired spot.  This is due to the fact that
*	the trigger comes directly from a CA monitor, and the sync samples
*	are buffered some time later.  This could be improved some by
*	moving the test to follow the sydInputGet loop.  An exact solution
*	requires that the trigger channel be part of the sync set and the
*	test based on the values in the syd buffers.
* o	for "until condition", sampling stops as soon as the condition is
*	satisfied.  This may leave several samples in the syd buffers if
*	filled and sync data are being intermixed.
*
*-*/
void
arReqCheckSample(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;

/*-----------------------------------------------------------------------------
*    Check to see if any sampling activities need to be started.
*
*	If ordinary sampling is being done, then no activities are needed
*	until the trigger occurs.
*
*	If pre-trigger sampling is being done, then the pre-trigger queueing
*	needs to be started.
*----------------------------------------------------------------------------*/
    if (pReq->sampling == 0) {
	if (pReq->sample.trigger.armed != 1) {
	    reqTrigArm(&pReq->sample.trigger, nowTs);
	}
	if (reqTrigTest(&pReq->sample.trigger, nowTs) == 0) {
	    if (pReq->sample.startWhen != REQ_TYPE_BEFORE)
		return;
	    if (pReq->sample.trigger.type != REQ_TYPE_EVENT ||
				pReq->sample.trigger.everyPrimed == 1) {
		arReqSampleSnapStart(pReq, nowTs);
	    }
	    else
		return;
	}
	else if (pReq->sample.type == REQ_TYPE_SNAP) {
	    arReqSampleSnapStart(pReq, nowTs);
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	}
	else {
	    arReqSampleMonStart(pReq, nowTs);
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	}
    }
    else if (pReq->sampling == -1) {
	if (reqTrigTest(&pReq->sample.trigger, nowTs) == 1) {
/*-----------------------------------------------------------------------------
*	A trigger has finally occurred in pre-trigger queueing mode.  Set
*	up the proper starting sample and switch into normal sampling mode.
*
*	First, figure out which sample to start with, depending on whether
*	pre-trigger is based on number of samples or on time.  Then, adjust
*	the deltaSec in the sample set spec, and also adjust firstData and
*	sampleCount.  If sampling is "for hh:mm:ss", adjust forStamp.
*----------------------------------------------------------------------------*/
	    TS_STAMP	ts;
	    short	samp=pSspec->firstData;

	    if (pSspec->sampleCount <= 0)
		;	/* no samples yet; just switch into normal mode */
	    else if (pReq->sample.startType == REQ_TYPE_NTH) {
		if (pSspec->sampleCount <= pReq->sample.startCount)
		    ;	/* current count <= desired; just go to normal mode */
		else {
		    samp = pSspec->lastData - pReq->sample.startCount + 1;
		    if (samp < 0)
			samp += pSspec->dataDim;
		}
	    }
	    else {
		TsDiffAsStamp(&ts,
			&pReq->sample.trigger.dataTs, &pReq->sample.startStamp);
		while (1) {
		    if (TsCmpStampsGE(&pSspec->pTimeStamp[samp], &ts))
			break;
		    if (samp == pSspec->lastData)
			break;
		    else if (++samp >= pSspec->dataDim)
			samp = 0;
		}
	    }
	    if (samp != pSspec->firstData) {
		ts = pSspec->pTimeStamp[samp];
		ts.nsec = 0;
		pSspec->refTs = ts;
		pSspec->firstData = samp;
		pSspec->sampleCount = 0;
		while (1) {
		    pSspec->sampleCount++;
		    TsDiffAsDouble(&pSspec->pDeltaSec[samp],
						&pSspec->pTimeStamp[samp], &ts);
		    if (samp == pSspec->lastData)
			break;
		    else if (++samp >= pSspec->dataDim)
			samp = 0;
		}
	    }
	    if (pReq->sample.forType == REQ_TYPE_TIMED) {
		double	d=-1.*pSspec->pDeltaSec[pSspec->lastData];
		TsAddStamps(&pReq->sample.forStop,
						&nowTs, &pReq->sample.forStamp);
		TsAddDouble(&pReq->sample.forStop, &pReq->sample.forStop, d);
	    }
	    if (pReq->sample.forEvent.type == REQ_TYPE_EVENT)
		reqTrigArm(&pReq->sample.forEvent, nowTs);
	    pReq->sampling = 1;
	    REQ_PRINTF "%s  snapshot sampling...", NOW);
	    reqStatPrintf(&pReq->sampleStat, "snapshot sampling");
	}
    }
    if (pReq->sample.halt == 1)
	goto checkSample_stop;
    if (pReq->sample.type == REQ_TYPE_SNAP) {
	while (1) {
	    if ((stat = sydInputGet(pSspec, NULL)) == S_syd_noDataNow)
		break;
	    sydInputStoreInSet(pSspec, pReq->sample.ignorePartial);
	    if (pReq->sampling==1 && pReq->sample.forType==REQ_TYPE_SAMPLES) {
		if (pSspec->sampleCount >= pReq->sample.forCmp)
		    goto checkSample_stop;
	    }
	    if (pSspec->sampleCount >= pSspec->reqCount)
		goto checkSample_stop;
	}
    }
    else if (pReq->sample.type == REQ_TYPE_MONITOR) {
	if (pReq->sample.forType == REQ_TYPE_SCHED) {
	    if (TsCmpStampsGT(&pReq->lastTs, &pReq->sample.forStop))
		goto checkSample_stop;
	}
    }
    if (pReq->sampling == 1) {
/*-----------------------------------------------------------------------------
*    Check to see if sampling should be stopped.  This might be for any of:
*	o   for n samples  (only for SNAP, checked above)
*	o   for hh:mm:ss
*	o   until chanName condition
*
*	These checks aren't made if pre-trigger queueing is going on.
*----------------------------------------------------------------------------*/
	if (pReq->sample.forType == REQ_TYPE_TIMED &&
			TsCmpStampsGT(&nowTs, &pReq->sample.forStop)) {
	    goto checkSample_stop;
	}
	if (pReq->sample.forType == REQ_TYPE_EVENT) {
	    if (reqEventTest_sm(&pReq->sample.forEvent, NULL) == 1) {
		goto checkSample_stop;
	    }
	}
    }
    return;

checkSample_stop:
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);
}

/*+/subr**********************************************************************
* NAME	arReqDemandXxx_mi - handle "demand" from operator
*
*-*/
void arReqDemandAct_mi(menu, item)
Menu	menu;
Menu_item item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    pReq->actTrig.demand = 1;
}
void arReqDemandInact_mi(menu, item)
Menu	menu;
Menu_item item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    pReq->inactTrig.demand = 1;
}
void arReqDemandSample_xvo(item)
Xv_opaque item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    pReq->sample.trigger.demand = 1;
}
void arReqDemandSample_mi(menu, item)
Menu	menu; Menu_item item; {    arReqDemandSample_xvo(item); }
void arReqDemandStep_xvo(item)
Xv_opaque item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    AR_CTX	*pCtx=pReq->pCtx;
    if (pReq->step.active) {
	guiNotice(&pCtx->guiCtx, "stepping already in progress");
	return;
    }
    if (ca_field_type(pReq->step.pCh) == TYPENOTCONN) {
	guiNoticePrintf(&pCtx->guiCtx,
			"stepping channel disconnected: %s\n", pReq->step.name);
	return;
    }
    reqStep_seqStart(&pReq->step);
}

/*+/subr**********************************************************************
* NAME	arReqClose
*
*-*/
void
arReqClose_mi(menu, item)
Menu	menu;
Menu_item item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    arReqClose(pReq, 1);
}
void
arReqClose(pReq, destroyFlag)
AR_REQ	*pReq;
int	destroyFlag;
{
    AR_CTX	*pCtx=pReq->pCtx;
    AR_CHAN	*pReqChan;
    TS_STAMP	nowTs;

    tsLocalTime(&nowTs);
    if (pReq->sample.type == REQ_TYPE_SNAP)
	arReqSampleSnapStop(pReq, nowTs);
    else
	arReqSampleMonStop(pReq, nowTs);
    reqStatPrintf(&pReq->reqStat, "closed");

    reqTrigWrap(&pReq->actTrig);
    reqTrigWrap(&pReq->inactTrig);
    reqTrigWrap(&pReq->sample.forEvent);
    reqTrigWrap(&pReq->sample.trigger);
    reqStepWrap(&pReq->step);
    reqStatWrap(&pReq->reqStat);
    reqStatWrap(&pReq->sampleStat);

    if (destroyFlag)
	arReqClose_destroy(pCtx, pReq);
    DoubleListRemove(pReq, pCtx->pReqHead, pCtx->pReqTail);
    for (; pReq->pChanHead!=NULL; ) {
	pReqChan = pReq->pChanHead;
	DoubleListRemove(pReqChan, pReq->pChanHead, pReq->pChanTail);
	if (pReqChan->pCfChan != NULL)
	    arCFChanClose(&pReqChan->pCfChan);
	if (pReq->pSspec != NULL && pReqChan->pSChan != NULL)
	    sydChanClose(pReq->pSspec, pReqChan->pSChan);
	genFree((char *)pReqChan);
    }
    if (pReq->pSspec != NULL)
	sydClose(pReq->pSspec);
    if (pReq->pCfDesc != NULL)
	arCFClose(&pReq->pCfDesc);
    unlink(pReq->lockPath);
}

/*+/subr**********************************************************************
* NAME	arReqClose_destroy - destroy panel for a request
*
*-*/
static
arReqClose_destroy(pCtx, pReq)
AR_CTX	*pCtx;
AR_REQ	*pReq;
{
    AR_REQ	*pReq1;
    int		y=(int)xv_get(pReq->panel, XV_Y);

    if (pReq->pPrCtx != NULL)
	PRquit(pReq->pPrCtx);
    if (pReq->pPlCtx != NULL)
	PLquit(pReq->pPlCtx);

    if (pReq->panel != NULL)
	xv_destroy_safe(pReq->panel);
    for (pReq1=pReq->pNext; pReq1!= NULL; pReq1=pReq1->pNext) {
	if (pReq1->panel != NULL) {
	    xv_set(pReq1->panel, XV_Y, y, NULL);
	    y += (int)xv_get(pReq1->panel, XV_HEIGHT);
	}
    }
    pCtx->y = y;
    xv_set(pCtx->guiCtx.baseFrame, XV_HEIGHT, y, NULL);
}

/*+/subr**********************************************************************
* NAME	arReqEdit
*
*-*/
void
arReqEdit_mi(menu, item)
Menu	menu;
Menu_item item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    arReqEdit(pReq);
}
void
arReqEdit(pReq)
AR_REQ	*pReq;
{
    AR_CTX	*pCtx=pReq->pCtx;

    edit(pCtx, NULL, pReq->file, 1);
}

/*+/internal******************************************************************
* NAME	arReqOpen - open an archive request from a raw request structure
*
*-*/
static void
arReqOpen(pCtx, pRaw, path)
AR_CTX	*pCtx;		/* pointer to context block */
RAW_REQ	*pRaw;		/* pointer to raw request */
char	*path;		/* path to file */
{
    AR_REQ	*pReq;
    enum sydChanSync sync;
    long	stat;
    int		x=5, y=0, ht=0;
    Panel	panel=pCtx->panel;
    Menu	menu;
    RAW_CHAN	*pChan;
    SYD_CHAN	*pSChan;
    AR_CHAN_DESC *pCfChan;
    AR_CHAN	*pReqChan;
    int		inactMi;

    if (pRaw->totalErrors > 0) {
	guiNoticePrintf(&pCtx->guiCtx, "%s not opened: %d errors encountered",
			pRaw->name, pRaw->totalErrors);
	rawReqFree(pRaw);
	guiCFshowPin(pCtx->validate_CF);
	return;
    }
    for (pReq=pCtx->pReqHead; pReq!=NULL; pReq=pReq->pNext) {
	if (strcmp(pReq->name, pRaw->name) == 0) {
	    guiNoticePrintf(&pCtx->guiCtx, "%s already opened", pRaw->name);
	    rawReqFree(pRaw);
	    return;
	}
    }
    pReq = (AR_REQ *)genMalloc(sizeof(AR_REQ));
    assertAlways(pReq != NULL);
    strcpy(pReq->name, pRaw->name);
    strcpy(pReq->dir, pRaw->dir);
    strcpy(pReq->file, pRaw->file);
    sprintf(pReq->lockPath, "%s.open", path);
    if (guiLock(&pCtx->guiCtx, pReq->name, pReq->lockPath, 1) != 1) {
	rawReqFree(pRaw);
	genFree((char *)pReq);
	return;
    }
    pReq->pCtx = pCtx;
    pReq->pPrCtx = NULL;
    pReq->pPlCtx = NULL;
    pReq->pCfDesc = NULL;
    pReq->pChanHead = pReq->pChanTail = NULL;
    pReq->pNext = pReq->pNext = NULL;
    pReq->sample = pRaw->sample;
    pReq->actTrig = pRaw->actTrig;
    pReq->allowDiscon = pRaw->allowDiscon;
    pReq->inactTrig = pRaw->inactTrig;
    strcpy(pReq->reqStat.name, pRaw->reqStatChan.name);
    strcpy(pReq->sampleStat.name, pRaw->sampleStatChan.name);
    pReq->pSspec = NULL;
    pReq->status = REQ_ST_ENABLED;
    pReq->nDis = pReq->nCtrlDis = -1;
    pReq->sampling = 0;
    pReq->step = pRaw->step;
    pReq->getGR = 0;
    pReq->panel = panel = (Panel)xv_create(pCtx->guiCtx.baseFrame,PANEL,NULL);
    xv_set(panel, WIN_WIDTH, WIN_EXTEND_TO_EDGE, XV_X, 0, XV_Y, pCtx->y, NULL);

/*-----------------------------------------------------------------------------
*  description......
*  control v        status v
*     activate         show state
*     inactivate       show IOC connections
*     start sampling   edit...
*     close            printing...	(snapshot only)
*                      plotting...	(snapshot only)
*  status message.......
*----------------------------------------------------------------------------*/
    guiMessage(pRaw->desc, panel, &x,&y,&ht);
    x = 5; y += ht; ht = 0;
    menu = guiMenu(NULL, 0,NULL,0,NULL);
    inactMi = pReq->actTrig.type == REQ_TYPE_DEMAND ? 0 : 1;
    guiMenuItem("activate", menu,
		arReqDemandAct_mi, inactMi, 0, ARREQ,pReq,0,NULL);
    inactMi = pReq->inactTrig.type == REQ_TYPE_DEMAND ? 0 : 1;
    guiMenuItem("inactivate", menu,
		arReqDemandInact_mi, inactMi, 0, ARREQ,pReq,0,NULL);
    inactMi = pReq->sample.trigger.type == REQ_TYPE_DEMAND ? 0 : 1;
    guiMenuItem("start sampling", menu,
		arReqDemandSample_mi, inactMi, 0, ARREQ,pReq,0,NULL);
    if (!inactMi) {
	if (pReq->sample.type == REQ_TYPE_SNAP) {
	    guiButton("take a snapshot",panel,&x,&y,&ht,arReqDemandSample_xvo,
			NULL, ARREQ,pReq,0,NULL);
	}
	else if (pReq->sample.type == REQ_TYPE_MONITOR) {
	    guiButton("start monitoring",panel,&x,&y,&ht,arReqDemandSample_xvo,
			NULL, ARREQ,pReq,0,NULL);
	}
    }
    if (pReq->step.type == REQ_TYPE_STEP) {
	guiButton("start stepping",panel,&x,&y,&ht,arReqDemandStep_xvo,
			NULL, ARREQ,pReq,0,NULL);
    }
    guiMenuItem("close", menu, arReqClose_mi, 0, 0, ARREQ,pReq,0,NULL);
    guiButton("control",panel,&x,&y,&ht,NULL, menu, 0,NULL,0,NULL);
    menu = guiMenu(NULL, 0,NULL,0,NULL);

    guiMenuItem("show state", menu, arReqStatus_mi, 0, 1,
			ARREQ, pReq, ARINFO, AR_INF_STATE);
    guiMenuItem("show sampling state", menu, arReqStatus_mi, 0, 0,
			ARREQ, pReq, ARINFO, AR_INF_SAMP);
    guiMenuItem("show IOC connections", menu, arReqStatus_mi, 0, 0,
			ARREQ, pReq, ARINFO, AR_INF_CONN);
    guiMenuItem("show channel stepping", menu, arReqStatus_mi, 0, 0,
			ARREQ, pReq, ARINFO, AR_INF_STEP);
    guiMenuItem("edit...", menu, arReqEdit_mi, 0, 0, ARREQ,pReq,0,NULL);
    if (pReq->sample.type == REQ_TYPE_SNAP) {
	PRinitMenu(&pReq->pPrCtx, &pCtx->guiCtx, pCtx->pGuiPrt,
		"AR: printing...", menu, 0, &pReq->pSspec);
	PLinitMenu(&pReq->pPlCtx, &pCtx->guiCtx, pCtx->pGuiPrt,
		"AR: plotting...", menu, 0, &pReq->pSspec);
    }

    guiButton("status",panel,&x,&y,&ht,NULL, menu, 0,NULL,0,NULL);

    x = 5; y += ht; ht = 0;
    pReq->status_PM = guiMessage(pReq->name, panel, &x,&y,&ht);
    x = 5; y += ht; ht = 0;
    window_fit_height(panel);
    window_fit(pCtx->guiCtx.baseFrame);
    pCtx->x = 0; pCtx->y += (int)xv_get(panel, XV_HEIGHT); pCtx->ht = 0; 

    if (pReq->sample.outputType == REQ_OUT_BCF) {
	FILE	*pFile;
	char	path[GUI_TDIM*2];

	sprintf(path, "%s/%s", pReq->sample.outDir, pReq->sample.outFile);
	if ((pFile = fopen(path, "r")) == NULL) {
	    pReq->pCfDesc = arCFCreate(path, pReq->sample.outSize);
	    if (pReq->pCfDesc == NULL) {
		guiNoticeFile(&pCtx->guiCtx, "couldn't create file", path);
		rawReqFree(pRaw);
		arReqClose_destroy(pCtx, pReq);
		unlink(pReq->lockPath);
		genFree((char *)pReq);
		return;
	    }
	}
	else {
	    fclose(pFile);
	    pReq->pCfDesc = arCFOpen(path, O_RDWR);
	    if (pReq->pCfDesc == NULL) {
		guiNoticeFile(&pCtx->guiCtx, "couldn't open file", path);
		rawReqFree(pRaw);
		arReqClose_destroy(pCtx, pReq);
		unlink(pReq->lockPath);
		genFree((char *)pReq);
		return;
	    }
	}
    }

    DoubleListAppend(pReq, pCtx->pReqHead, pCtx->pReqTail);
    reqTrigInit(&pReq->actTrig);
    reqTrigInit(&pReq->inactTrig);
    reqTrigInit(&pReq->sample.forEvent);
    reqTrigInit(&pReq->sample.trigger);
    reqStepInit(&pReq->step);
    reqStatInit(&pReq->reqStat);
    reqStatInit(&pReq->sampleStat);

    stat = sydOpenCA(&pReq->pSspec, NULL);
    assertAlways(stat == S_syd_OK);
    pReq->pSspec->roundNsec = pRaw->sample.roundMsec*1000000;
    if (pReq->sample.type == REQ_TYPE_MONITOR)
	sydSetAttr(pReq->pSspec, SYD_ATTR_MON_FN, 0, arReqSampleMonHandler);
    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext) {
	pReqChan = (AR_CHAN *)genMalloc(sizeof(AR_CHAN));
	assertAlways(pReqChan != NULL);
	pReqChan->pReq = pReq;
	pReqChan->pCfChan = NULL;
	pReqChan->pSChan = NULL;
	if (pChan->filled)	sync = SYD_SY_FILLED;
	else			sync = SYD_SY_NONF;
	stat = sydChanOpen(pReq->pSspec, &pSChan, pChan->name,sync,pReqChan,0);
	assertAlways(stat == S_syd_chanNotConn);
	if (pReq->sample.outputType == REQ_OUT_BCF) {
	    pCfChan = arCFChanOpen(pReq->pCfDesc, pChan->name, O_RDWR);
	    if (pCfChan == NULL) {
		guiNoticePrintf(&pCtx->guiCtx,
			"couldn't do arCFChanOpen for %s", pChan->name);
		genFree(pReqChan);
		rawReqFree(pRaw);
		arReqClose_destroy(pCtx, pReq);
		return;
	    }
	    pReqChan->pCfChan = pCfChan;
	    pReqChan->pSChan = pSChan;
	    if ((pCfChan->flags & AR_CDESC_NEW) != 0)	pReqChan->getGR = 1;
	    else					pReqChan->getGR = 0;
	    if (pReqChan->getGR)
		pReq->getGR++;
	    DoubleListAppend(pReqChan, pReq->pChanHead, pReq->pChanTail);
	}
    }
    REQ_PRINTF "opened");
    reqStatPrintf(&pReq->reqStat, "opened");
    rawReqFree(pRaw);
}

/*+/subr**********************************************************************
* NAME	arReqReadFile - open an archive request from a file
*
*-*/
static void
arReqReadFile(pCtx, file)
AR_CTX	*pCtx;
char	*file;
{
    RAW_REQ	*pRaw;
    char	path[GUI_TDIM*2];
    FILE	*pFile;

    sprintf(path, "%s/%s", pCtx->Req_subdir, file);
    if ((pFile = fopen(path, "r")) == NULL) {
	guiNoticeFile(&pCtx->guiCtx, "couldn't open file", path);
	return;
    }
    fclose(pFile);
    if ((pRaw = rawReqReadFile(pCtx->Req_subdir, file, NULL, NULL)) == NULL) {
	return;
    }
    arReqOpen(pCtx, pRaw, path);
}

/*+/subr**********************************************************************
* NAME	arReqSampleMonHandler - handle Channel Access monitors for a request
*
*-*/
void arReqSampleMonHandler(arg)
struct event_handler_args arg;
{
    AR_CHAN	*pChan=(AR_CHAN *)arg.usr;
    AR_REQ 	*pReq=pChan->pReq;
    long	stat;
    TS_STAMP	*pStamp;

    if (pReq->sampling == 0)
	return;
    pStamp = &((struct dbr_time_float *)(arg.dbr))->stamp;
    if (pReq->pSspec->roundNsec > 0)
	sydTsRound(pStamp, pReq->pSspec->roundNsec);
    if (pReq->firstTs.secPastEpoch == 0)
	pReq->firstTs = *pStamp;
    if (TsCmpStampsGT(pStamp, &pReq->lastTs))
	pReq->lastTs = *pStamp;
    stat = arCFChanWrite(pChan->pCfChan, &arg);
    assert(stat == OK);
    pReq->monitorCount++;
}

/*+/subr**********************************************************************
* NAME	arReqSampleMonStart - start sampling in monitor mode
*
*-*/
void arReqSampleMonStart(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;
    AR_CHAN	*pRChan;

    if (pReq->status != REQ_ST_ACTIVE) {
	REQ_PRINTF "can't monitor--state isn't `active'");
	return;
    }
    REQ_PRINTF "%s  monitor sampling...", NOW);
    reqStatPrintf(&pReq->sampleStat, "monitor sampling");
    pReq->sample.halt = 0;
    pReq->sampling = 1;
    pReq->firstTs.secPastEpoch = pReq->lastTs.secPastEpoch = 0;
    pReq->firstTs.nsec = pReq->lastTs.nsec = 0;
    pReq->monitorCount = 0;
    if (pReq->sample.forType == REQ_TYPE_TIMED)
	TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);
    if (pReq->step.samplePending) {
	pReq->step.samplePending = 0;
	pReq->step.sampleActive = 1;
    }

    for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext)
	arCFChanWriteByte(pRChan->pCfChan, AR_DAT_IC_MON_BEGIN);
    sydMonitorStart(pReq->pSspec);
}

/*+/internal******************************************************************
* NAME	arReqSampleMonStop - stop sampling in monitor mode
*
*-*/
arReqSampleMonStop(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    char	fromTxt[28], toTxt[28];

    if (pReq->sampling == 0)
	return;
    sydMonitorStop(pReq->pSspec);
    pReq->sampling = 0;
    if (pReq->sample.halt)
	REQ_PRINTF "%s  monitoring halted; ", NOW);
    else
	REQ_PRINTF "%s  monitoring stopped; ", NOW);
    reqTrig_setDone(&pReq->inactTrig);
    tsStampToText(&pReq->firstTs, TS_TEXT_MMDDYY, fromTxt);
    tsStampToText(&pReq->lastTs, TS_TEXT_MMDDYY, toTxt);
    REQ_PRINTF "%d values from %s to %s", pReq->monitorCount, fromTxt, toTxt);
    reqStatPrintf(&pReq->sampleStat, "%.5s-%.5s n=%d",
			    &fromTxt[12], &toTxt[12], pReq->monitorCount);
    stat = arCFSyncWrite(pReq->pCfDesc);
    assert(stat == OK);
}

/*+/internal******************************************************************
* NAME	arReqSampleSnapStart - start sampling in snapshot mode
*
*-*/
void arReqSampleSnapStart(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;

    if (pReq->status != REQ_ST_ACTIVE) {
	REQ_PRINTF "can't snapshot--state isn't `active'");
	return;
    }
    if (pReq->sample.trigger.eventHappened ||
				pReq->sample.startWhen != REQ_TYPE_BEFORE) {
	REQ_PRINTF "%s  snapshot sampling...", NOW);
	reqStatPrintf(&pReq->sampleStat, "snapshot sampling");
	if (pReq->sample.forType == REQ_TYPE_TIMED)
	    TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);
	pReq->sampling = 1;
    }
    else {
	REQ_PRINTF "%s  queueing pre-trigger samples...", NOW);
	reqStatPrintf(&pReq->sampleStat, "pre-trigger sampling");
	pReq->sampling = -1;
    }
    pReq->sample.halt = 0;
    if (pReq->step.samplePending) {
	pReq->step.samplePending = 0;
	pReq->step.sampleActive = 1;
    }

    sydInputResetKeepNewest(pSspec);
    sydSampleSetAlloc(pSspec, pReq->sample.bufNsamples);
    pSspec->refTs.secPastEpoch = 0;
    pSspec->refTs.nsec = 0;
    if (pReq->sample.startWhen != REQ_TYPE_BEFORE) {
	if (pReq->sample.forType == REQ_TYPE_TIMED)
	    TsAddStamps(&pReq->sample.forStop, &nowTs, &pReq->sample.forStamp);
    }
}

/*+/internal******************************************************************
* NAME	arReqSampleSnapStop - stop sampling in snapshot mode
*
*-*/
arReqSampleSnapStop(pReq, nowTs)
AR_REQ	*pReq;
TS_STAMP nowTs;
{
    long	stat;
    AR_CTX	*pCtx=pReq->pCtx;
    SYD_SPEC	*pSspec=pReq->pSspec;
    char	fromTxt[28], toTxt[28];
    chtype	type;
    AR_CHAN	*pRChan;
    void	*pData;

    if (pReq->sampling != 1)
	return;
    pReq->sampling = 0;
    if (pReq->sample.halt) {
	REQ_PRINTF "%s  snapshot aborted", NOW);
	reqStatPrintf(&pReq->sampleStat, "snapshot aborted");
	return;
    }
    reqTrig_setDone(pReq->inactTrig);
    if (pSspec->sampleCount <= 0) {
	REQ_PRINTF "%s  snapshot complete; no samples retrieved", NOW);
	reqStatPrintf(&pReq->sampleStat, "snap done--no samples");
	return;
    }
    tsStampToText(&pSspec->pTimeStamp[pSspec->firstData],
			    TS_TEXT_MMDDYY, fromTxt);
    tsStampToText(&pSspec->pTimeStamp[pSspec->lastData],
			    TS_TEXT_MMDDYY, toTxt);
    REQ_PRINTF "snapshot complete; %d samples from %s to %s",
			    pSspec->sampleCount, fromTxt, toTxt);
    reqStatPrintf(&pReq->sampleStat, "%.5s-%.5s n=%d",
			    &fromTxt[12], &toTxt[12], pSspec->sampleCount);
    if (pReq->sample.outputType == REQ_OUT_SSF) {
	FILE	*pFile;
	char	path[GUI_TDIM*2];

	sprintf(path, "%s/%s", pReq->sample.outDir, pReq->sample.outFile);
	if ((pFile = fopen(path, "a")) == NULL) {
	    guiNoticeFile(&pReq->pCtx->guiCtx, "couldn't open file", path);
	    REQ_PRINTF "snapshot aborted--couldn't open %s", path);
	    reqStatPrintf(&pReq->sampleStat, "error on file write");
	    return;
	}
	sydSampleSetWriteSSF(pReq->pSspec, pFile,
			"PROGRAM: AR", "operator snapshot");
	fclose(pFile);
    }
    else {
	int		samp, count;
	SYD_CHAN	*pSChan;

	for (pRChan=pReq->pChanHead; pRChan!=NULL; pRChan=pRChan->pNext) {
	    pSChan = pRChan->pSChan;
	    type = pSChan->dbfType;
	    count = pSChan->elCount;
	    samp = pSspec->firstData;
	    if (pSChan->pData == NULL)
		continue;	/* skip to end of for loop */
	    while (1) {
		if (pSChan->pFlags[samp].eof)
		    break;
		if (pSChan->pFlags[samp].missing)
		    ;	/* no action on missing data */
		else if (samp != pSspec->firstData &&
					pSChan->pDataCodeL[samp] == 'F') {
		    ;	/* no action on filled data, except first */

#if 0
jjj	/* need to handle snap_begin and _end for filled stuff */
	/* also need to consider what to do if start a snap write and
		it's interrupted by EOF or INACT */
#endif

		}
		else {
		    if (samp == pSspec->firstData) {
			if (pSspec->sampleCount > 1) {
			    arCFChanWriteByte(pRChan->pCfChan,
						AR_DAT_IC_SNAP_BEGIN);
			}
			else {
			    arCFChanWriteByte(pRChan->pCfChan,
						AR_DAT_IC_SNAP_SINGLE);
			}
		    }
		    else if (samp == pSspec->lastData)
			arCFChanWriteByte(pRChan->pCfChan, AR_DAT_IC_SNAP_END);
		    pData = (void *)(((char *)pSChan->pData) +
					samp * dbr_value_size[type]);
		    stat = arCFChanWrite_args(pRChan->pCfChan,
			    pSspec->pTimeStamp[samp], type, count, pData,
			    pSChan->pDataAlStat[samp],
			    pSChan->pDataAlSev[samp]);
		    assert(stat == OK);
		}
		if (samp == pSspec->lastData)
		    break;
		else if (++samp >= pSspec->dataDim)
		    samp = 0;
	    }
#if 0
	    jjj
#endif
	}
	stat = arCFSyncWrite(pReq->pCfDesc);
	assert(stat == OK);
    }
}

/*+/subr**********************************************************************
* NAME	arReqStatus - print status messages
*
*-*/
void
arReqStatus_mi(menu, item)
Menu	menu;
Menu_item item;
{   AR_REQ	*pReq=(AR_REQ *)xv_get(item, XV_KEY_DATA, ARREQ);
    AR_INF	type=(AR_INF)xv_get(item, XV_KEY_DATA, ARINFO);
    arReqStatus(pReq, type);
}

void
arReqStatus(pReq, type)
AR_REQ	*pReq;
AR_INF	type;
{
    int		nCon, nDis, nCtrlDis;
    char	lastTxt[28];
    SYD_SPEC	*pSspec=pReq->pSspec;

    arReqCheckConn(pReq, &nCon, &nDis, &nCtrlDis, 0);

    if (type == AR_INF_STATE) {
	if (pReq->status == REQ_ST_ENABLED) {
	    REQ_PRINTF "opened; %d connected, %d connections pending",
			nCon, nDis);
	}
	else if (pReq->status == REQ_ST_ACTIVE) {
	    char	*conn="", *samp="";
	    if (pReq->sampling) samp = "; sampling";
	    if (nDis > 0) conn = "; partly connected";
	    REQ_PRINTF "activated%s%s", conn, samp);
	}
	else if (pReq->status == REQ_ST_INACTIVE) {
REQ_PRINTF "inactivated; %d connected, %d connections pending", nCon, nDis);
	}
    }
    else if (type == AR_INF_SAMP) {
	if (pReq->sampling) {
	    if (pReq->sample.type == REQ_TYPE_SNAP) {
		if (pReq->sampling == 1) {
		    tsStampToText(&pSspec->pTimeStamp[pSspec->lastData],
			    TS_TEXT_MMDDYY, lastTxt);
		    REQ_PRINTF "snapshot in progress; %d samples, last at %s",
			    pSspec->sampleCount, lastTxt);
		}
		else {
		    REQ_PRINTF "pre-trigger sampling");
		}
	    }
	    else {
		tsStampToText(&pReq->lastTs, TS_TEXT_MMDDYY, lastTxt);
		REQ_PRINTF "monitoring in progress; %d values, last at %s",
			    pReq->monitorCount, lastTxt);
	    }
	}
	else
	    REQ_PRINTF "not presently sampling");
    }
    else if (type == AR_INF_CONN) {
	REQ_PRINTF "%d connected, %d connections pending", nCon, nDis);
    }
    else if (type == AR_INF_STEP) {
	char	*pState;
	if (pReq->step.active) {
	    if (pReq->step.getPending)		pState = "get pending";
	    else if (pReq->step.delayPending)	pState = "delay pending";
	    else if (pReq->step.samplePending)	pState = "sample pending";
	    REQ_PRINTF "stepping:%s; last val=%f", pState, pReq->step.present);
	}
	else
	    REQ_PRINTF "not stepping");
    }
}

/*+/subr**********************************************************************
* NAME	configAddCannedItem - add an item to the add-on list
*
* DESCRIPTION
*	Searches the keywords for the canned list of add-on items.
*	If the item is found, it is added to the add-on list.
*
*	No check is made for duplicate items.
*
*-*/
configAddCannedItem(pCtx, pField, place)
AR_CTX	*pCtx;
char	*pField;
char	*place;
{
    int		i;
    ADDON	*pAdd;

    for (i=0; i<ADD_CANNED_DIM; i++) {
	if (strcmp(pField, configAddItems[i].keyword) == 0) {
	    configAddItem(pCtx,
		&configAddItems[i], place, configAddItems[i].type, NULL);
	    return;
	}
    }
    printf("configAddCannedItem: couldn't find %s\n", pField);
}

/*+/subr**********************************************************************
* NAME	configAddonsClear - clear list of add-on panel items
*
*-*/
configAddonsClear(pCtx)
AR_CTX	*pCtx;
{
    ADDON	*pAdd, *pAdd1;
    for (pAdd=pCtx->pAddonHead; pAdd!=NULL; ) {
	pAdd1 = pAdd;
	pAdd = pAdd->pNext;
	DoubleListRemove(pAdd1, pCtx->pAddonHead, pCtx->pAddonTail);
	genFree((char *)pAdd1);
    }
}

/*+/subr**********************************************************************
* NAME	configAddItem - allocate and fill in an add-on item structure
*
*-*/
configAddItem(pCtx, pItem, place, type, pArg)
AR_CTX	*pCtx;
ADD_CANNED *pItem;
char	*place;
ADD_TY	type;
void	*pArg;
{
    ADDON	*pAdd;

    pAdd = (ADDON *)genMalloc(sizeof(ADDON));
    assertAlways(pAdd != NULL);
    DoubleListAppend(pAdd, pCtx->pAddonHead, pCtx->pAddonTail);
    pAdd->item = pItem;
    pAdd->place = place;
    pAdd->type = type;
    if (type == ADD_TY_X)
	pAdd->intVal = (int)pArg;
    else if (type == ADD_TY_STR) {
	assert(strlen((char *)pArg) < GUI_TDIM);
	strcpy(pAdd->strVal, (char *)pArg);
    }
    else if (type == ADD_TY_EDIT)
	strcpy(pAdd->strVal, (char *)pArg);
    else if (type == ADD_TY_REQ)
	strcpy(pAdd->strVal, (char *)pArg);
}

/*+/subr**********************************************************************
* NAME	configAddRealize - realize add-on panel items
*
*-*/
configAddRealize(pCtx, panel, pX,pY,pHt, place)
AR_CTX	*pCtx;
Panel	panel;
int	*pX,*pY,*pHt;
char	*place;		/* place whose items are being realized */
{
    ADDON	*pAdd;
    for (pAdd=pCtx->pAddonHead; pAdd!=NULL; pAdd=pAdd->pNext) {
	if (strcmp(pAdd->place, place) != 0)
	    ;	/* ignore items for other places */
	else if (pAdd->type == ADD_TY_BTN) {
	    ADD_CANNED	*pBtn=(ADD_CANNED *)pAdd->item;;
	    guiButton(pBtn->label, panel,
		pX,pY,pHt, pBtn->fn, NULL, ARCTX,pCtx,0,NULL);
	}
	else if (pAdd->type == ADD_TY_NEWLINE) {
	    *pY += *pHt + 10; *pX = 5; *pHt = 0;
	}
	else if (pAdd->type == ADD_TY_X) {
	    *pX = pAdd->intVal;
	}
	else if (pAdd->type == ADD_TY_STR) {
	    guiMessage(pAdd->strVal, panel, pX,pY,pHt);
	}
	else if (pAdd->type == ADD_TY_EDIT) {
	    edit(pCtx, NULL, pAdd->strVal, 1);
	}
	else if (pAdd->type == ADD_TY_REQ) {
	    arReqReadFile(pCtx, pAdd->strVal);
	}
    }
}

/*+/subr**********************************************************************
* NAME	configAddEdit - add an archive request item to the add-on list
*
* DESCRIPTION
*	No check is made for duplicate items.
*
*-*/
configAddEdit(pCtx, reqName, place)
AR_CTX	*pCtx;
char	*reqName;
char	*place;
{
    ADDON	*pAdd;

    pAdd = (ADDON *)genMalloc(sizeof(ADDON));
    assertAlways(pAdd != NULL);
    DoubleListAppend(pAdd, pCtx->pAddonHead, pCtx->pAddonTail);
    strcpy(pAdd->strVal, reqName);
    pAdd->type = ADD_TY_EDIT;
    pAdd->place = place;
}

/*+/subr**********************************************************************
* NAME	configAddReq - add an archive request item to the add-on list
*
* DESCRIPTION
*	No check is made for duplicate items.
*
*-*/
configAddReq(pCtx, reqName, place)
AR_CTX	*pCtx;
char	*reqName;
char	*place;
{
    ADDON	*pAdd;

    pAdd = (ADDON *)genMalloc(sizeof(ADDON));
    assertAlways(pAdd != NULL);
    DoubleListAppend(pAdd, pCtx->pAddonHead, pCtx->pAddonTail);
    strcpy(pAdd->strVal, reqName);
    pAdd->type = ADD_TY_REQ;
    pAdd->place = place;
}

/*+/subr**********************************************************************
* NAME	configParse - parse configuration file
*
*-*/
configParse(pFile, pCtx)
FILE	*pFile;
AR_CTX	*pCtx;
{
    char	buf[120];
    char	*pBuf, *pField, dlm, *dir, *file;
    int		l, l1;
    ADDON	*pAdd;


    configAddonsClear(pCtx);

    while (fgets(buf, 120, pFile) != NULL) {
	pBuf = buf;
	l = 9999;
	while (l > 1) {
	    if ((l = nextNonSpaceField(&pBuf, &pField, &dlm)) > 1) {
		if (*pField != '#') {
		    if (strcmp(pField, "MAIN") == 0) {
			configParseITEM(pCtx, &pBuf, "MAIN");
			l = 0;
		    }
		    else if (strcmp(pField, "EDIT") == 0) {
			configParseITEM(pCtx, &pBuf, "EDIT");
			l = 0;
		    }
		    else if (strcmp(pField, "GLOSSARY") == 0) {
			l=nextNonSpaceField(&pBuf, &dir, &dlm);
			l1=nextNonSpaceField(&pBuf, &file, &dlm);
			if (l <= 1 || l1 <= 1)
 printf("AR: reading config file; directory and file required for GLOSSARY\n");
			else {
			    strcpy(pCtx->glossDir, dir);
			    strcpy(pCtx->glossFile, file);
			}
			l = 0;
		    }
		    else if (strcmp(pField, "REQ") == 0) {
			configParseREQ(pCtx, &pBuf, NULL);
			l = 0;
		    }
		    else if (strcmp(pField, "Dret") == 0) {
			if (nextNonSpaceField(&pBuf, &dir, &dlm) <= 1)
 printf("AR: reading config file; directory required for Dret\n");
			else
			    strcpy(pCtx->Req_subdir, pField);
		    }
		}
		else
		    l = 0;
	    }
	}
    }
}

/*+/subr**********************************************************************
* NAME	configParseITEM - handle adding items to add-on list
*
*-*/
configParseITEM(pCtx, ppBuf, place)
AR_CTX	*pCtx;		/* I pointer to context */
char	**ppBuf;	/* I ptr to pointer to input buffer */
char	*place;		/* I text id for place; e.g. "MAIN" */
{
    char	*pField, dlm;
    int		i;
    ADDON	*pAdd;
    int		x;

    if (nextNonSpaceField(ppBuf, &pField, &dlm) <= 1) {
 printf("AR: reading config file; parameter required with %s\n", place);
	return;
    }
    if (strcmp(pField, "NEWLINE") == 0)
	configAddItem(pCtx, NULL, place, ADD_TY_NEWLINE, NULL);
    else if (strcmp(pField, "STRING") == 0)
	configAddItem(pCtx, NULL, place, ADD_TY_STR, *ppBuf);
    else if (strcmp(pField, "X") == 0) {
	if (nextIntFieldAsInt(ppBuf, &x, &dlm) <= 1) {
 printf("AR: reading config file; illegal X for %s\n", place);
	    return;
	}
	configAddItem(pCtx, NULL, place, ADD_TY_X, x);
    }
    else
	configAddCannedItem(pCtx, pField, place);
}

/*+/subr**********************************************************************
* NAME	configParseReq - handle a REQ item
*
*-*/
configParseREQ(pCtx, ppBuf, pArg)
AR_CTX	*pCtx;
char	**ppBuf;
void	*pArg;
{
    char	*pField, dlm;

    if (nextNonSpaceField(ppBuf, &pField, &dlm) <= 1) {
 printf("AR: reading config file; REQ name missing\n");
	return;
    }
    if (strlen(pField) >= GUI_TDIM) {
 printf("AR: reading config file; REQ name too long\n");
	return;
    }
    configAddItem(pCtx, NULL, "REQ", ADD_TY_REQ, pField);
}


/*+/subr**********************************************************************
* NAME	edit - bring up "edit archive requests" command frame
*
*-*/
static void
edit(pCtx, dir, file, show)
AR_CTX	*pCtx;
char	*dir, *file;
int	show;
{
    char	mydir[120];

    if (dir == NULL) {
	sprintf(mydir, "%s/%s", pCtx->guiCtx.pwd, pCtx->Req_subdir);
	dir = mydir;
    }
    if (file != NULL)
	guiEditorLoadFile(pCtx->pEditor, dir, file);

    if (show)
	guiEditorShowCF(pCtx->pEditor);
}

/*+/subr**********************************************************************
* NAME	edit_pb - handle edit button push
*
*-*/
static void edit_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    edit(pCtx, NULL, NULL, 1);
}

/*+/subr**********************************************************************
* NAME	editExtras - add extra buttons to "edit archive requests" panel
*
*-*/
static void editExtras(pGuiCtx, pEdit, panel, pX,pY,pHt, pCtx)
GUI_CTX	*pGuiCtx;
GUI_EDIT *pEdit;
Panel	panel;
int	*pX, *pY, *pHt;
AR_CTX	*pCtx;
{
    guiButton("Add to main panel", panel, pX,pY,pHt, editOpen_pb, NULL,
			ARCTX, pCtx, 0, NULL);
    guiButton("Validate...", panel, pX,pY,pHt, guiCFshowPin_xvo, NULL,
			GUI_PCTX, &pCtx->guiCtx, GUI_CF, pCtx->validate_CF);
    configAddRealize(pCtx, panel, pX,pY,pHt, "EDIT");
}

/*+/subr**********************************************************************
* NAME	editOpen_pb - open an archive request from a text subwindow
*
* DESCRIPTION
*	Open the archive request in the edit window.
*
*	A lock for "open" is placed on the archive request.  This is
*	simply a file with the name of the request file suffixed with
*	".open".
*
*	The request is appended to the list of opened archive requests.
*
*-*/
static void editOpen_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    GUI_EDIT	*pEdit=pCtx->pEditor;
    RAW_REQ	*pRaw;

    pRaw = editValGetReq(pCtx, pEdit);
    if (pRaw == NULL) {
	guiNotice(&pCtx->guiCtx, "couldn't open request");
	return;
    }
    arReqOpen(pCtx, pRaw, pEdit->path);
}

/*+/subr**********************************************************************
* NAME	editValPuts - write messages to validation text subwindow
*
*-*/
static void editValPuts(message, pCtx)
char	*message;
AR_CTX	*pCtx;
{
    guiTextswPuts(pCtx->validate_TSW, message, &pCtx->validatePos);
}

/*+/subr**********************************************************************
* NAME	editValGetReq - build a raw request struct from info in textsw
*
* DESCRIPTION
*	Reads the lines in a text subwindow, parsing them as an archive
*	request.  Error messages are sent to the validation text subwindow.
*
*	When complete, the archive request is in the "raw request" structure.
*	The ->totalErrors item in the structure indicates whether any errors
*	were encountered during the parse.
*
* RETURNS
*	RAW_REQ *
*
*-*/
RAW_REQ *
editValGetReq(pCtx, pEdit)
AR_CTX	*pCtx;
GUI_EDIT *pEdit;
{
    RAW_REQ	*pRaw;
    int		inPos=0;
    char	line[120];
    Textsw	textsw=pCtx->validate_TSW;

    if (textsw != NULL) {
	guiTextswReset(pCtx->validate_TSW);
	pCtx->validatePos = 0;	/* for editValPuts */
    }
    pRaw = rawReqAlloc();
    assertAlways(pRaw != NULL);
    strcpy(pRaw->file, (char *)xv_get(pEdit->file_PT, PANEL_VALUE));
    strcpy(pRaw->dir, (char *)xv_get(pEdit->dir_PT, PANEL_VALUE));
    while (guiTextswGets(pEdit->text_TSW, line, sizeof(line), &inPos) != NULL)
	rawReqParseLine(pRaw, line, editValPuts, pCtx);
    rawReqCheckSemantics(pRaw, editValPuts, pCtx);
    return pRaw;
}

/*+/subr**********************************************************************
* NAME	editValidateInitCF - show the validation command frame
*
* DESCRIPTION
*	Makes the validation command frame visible.  If the editor and/or
*	validation command frames haven't yet been created, this is done.
*
*-*/
static void editValidateInitCF(pCtx)
AR_CTX	*pCtx;
{

/*-----------------------------------------------------------------------------
* Syntax check   Connect check   Interval check
* seconds for:  interval ___  jitter ___  test time ___
*----------------------------------------------------------------------------*/
    Panel	panel;
    int	x=5, y=10, ht=0;

    pCtx->validate_CF = guiFrameCmd("AR: validate archive request",
			    pCtx->guiCtx.baseFrame, &panel, NULL);
    guiButton("Syntax check", panel, &x,&y,&ht, editValSyntax_pb, NULL,
		    ARCTX, pCtx, 0, NULL);
    guiButton("Connect check", panel, &x,&y,&ht, editValConn_pb, NULL,
		    ARCTX, pCtx, 0, NULL);
    guiButton("Interval check", panel, &x,&y,&ht, editValInterval_pb, NULL,
		    ARCTX, pCtx, 0, NULL);
    y += ht + 10; x = 5; ht = 0;
    guiMessage("seconds for:", panel, &x,&y,&ht);
    pCtx->interval_PT = guiTextField("interval", panel, &x,&y,&ht,
		    NULL, 5, 5, 0, NULL, 0, NULL);
    pCtx->jitter_PT = guiTextField("jitter", panel, &x,&y,&ht,
		    NULL, 5, 5, 0, NULL, 0, NULL);
    pCtx->testTime_PT = guiTextField("test time", panel, &x,&y,&ht,
		    NULL, 5, 5, 0, NULL, 0, NULL);
    xv_set(pCtx->interval_PT, PANEL_VALUE, ".2", NULL);
    xv_set(pCtx->jitter_PT, PANEL_VALUE, ".01", NULL);
    xv_set(pCtx->testTime_PT, PANEL_VALUE, "10.", NULL);
    y += ht + 10; x = 5; ht = 0;
    window_fit_height(panel);
    pCtx->validate_TSW = guiBrowser(pCtx->validate_CF, panel, 130, 20, NULL);
    pCtx->validatePos = 0;
    window_fit(pCtx->validate_CF);
}

/*+/subr**********************************************************************
* NAME	editValSyntax_pb - validate syntax for an archive request
*
*-*/
static void editValSyntax_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    GUI_EDIT	*pEdit=pCtx->pEditor;
    RAW_REQ	*pRaw;

    if (pEdit->loaded == 0) {
	guiNotice(&pCtx->guiCtx, "no file has been loaded");
	return;
    }
    pRaw = editValGetReq(pCtx, pEdit);	/* also checks syntax */
    guiTextswPrintf(pCtx->validate_TSW, 
				"%d errors encountered\n", pRaw->totalErrors);

    rawReqFree(pRaw);
}

/*+/subr**********************************************************************
* NAME	editValConn_pb - validate channel connections an archive request
*
*-*/
static void editValConn_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    GUI_EDIT	*pEdit=pCtx->pEditor;
    RAW_REQ	*pRaw;
    double	testTime;
    char	*testTimeText=(char *)xv_get(pCtx->testTime_PT, PANEL_VALUE);

    if (pCtx->pReqHead != NULL) {
	guiTextswPrintf(pCtx->validate_TSW, 
		"operation not allowed while any requests are opened");
	return;
    }
    if (pEdit->loaded == 0) {
	guiNotice(&pCtx->guiCtx, "no file has been loaded");
	return;
    }
    if (sscanf(testTimeText, "%lf", &testTime) != 1) {
	guiNotice(&pCtx->guiCtx, "illegal testTime; 10 will be used");
	testTime = 10.0;
    }

    pRaw = editValGetReq(pCtx, pEdit);
    rawReqCheckSemantics(pRaw, editValPuts, pCtx);
    if (pRaw->totalErrors > 0) {
	guiTextswPrintf(pCtx->validate_TSW, 
				"%d errors encountered", pRaw->totalErrors);
    }
    rawReqCheckConnect_init(pRaw, editValPuts, pCtx);
    connect_pRaw = pRaw;
    tsLocalTime(&connectTs);
    connectTestTime = testTime;
}

/*+/subr**********************************************************************
* NAME	editValInterval_pb - validate channel connections an archive request
*
*-*/
static void editValInterval_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    GUI_EDIT	*pEdit=pCtx->pEditor;
    RAW_REQ	*pRaw;
    double	interval, jitter, testTime;
    char	*intervalText=(char *)xv_get(pCtx->interval_PT, PANEL_VALUE);
    char	*jitterText=(char *)xv_get(pCtx->jitter_PT, PANEL_VALUE);
    char	*testTimeText=(char *)xv_get(pCtx->testTime_PT, PANEL_VALUE);

    if (pEdit->loaded == 0) {
	guiNotice(&pCtx->guiCtx, "no file has been loaded");
	return;
    }
    if (pCtx->pReqHead != NULL) {
	guiTextswPrintf(pCtx->validate_TSW, 
		"operation not allowed while any requests are opened");
	return;
    }
    if (sscanf(intervalText, "%lf", &interval) != 1) {
	guiNotice(&pCtx->guiCtx, "illegal interval; .2 will be used");
	interval = .2;
    }
    if (sscanf(jitterText, "%lf", &jitter) != 1) {
	guiNotice(&pCtx->guiCtx, "illegal jitter; .01 will be used");
	jitter = .01;
    }
    if (sscanf(testTimeText, "%lf", &testTime) != 1) {
	guiNotice(&pCtx->guiCtx, "illegal testTime; 10 will be used");
	testTime = 10.0;
    }

    pRaw = editValGetReq(pCtx, pEdit);
    rawReqCheckInterval_init(pRaw, editValPuts, pCtx, interval, jitter);
    interval_pRaw = pRaw;
    tsLocalTime(&intervalTs);
    intervalTestTime = testTime;
}

/*+/subr**********************************************************************
* NAME	glossary_pb - bring up Channel Glossary command frame
*
*-*/
static void glossary_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);

    if (pCtx->pGlossary == NULL) {
	pCtx->pGlossary = guiEditor(&pCtx->guiCtx, "AR: channel glossary",
		pCtx->glossDir, pCtx->glossFile, pCtx->pGuiPrt, NULL, NULL);
    }

    guiEditorShowCF(pCtx->pGlossary);
}

/*+/subr**********************************************************************
* NAME	quit_pb - perform quit operations
*
* BUGS
* o	doesn't wrap up editing operations
*
*-*/
static void quit_pb(item)
Panel_item item;
{   AR_CTX	*pCtx=(AR_CTX *)xv_get(item, XV_KEY_DATA, ARCTX);
    arQuit(pCtx, 1);
}
/*+/subr**********************************************************************
* NAME	arQuit
*
*-*/
void arQuit(pCtx, destroyFlag)
AR_CTX	*pCtx;
int	destroyFlag;
{
    if (destroyFlag) {
	if (pCtx->pEditor != NULL && pCtx->pEditor->updFlag) {
	    guiNotice(&pCtx->guiCtx, "can't quit; edit mode is set");
	    guiEditorShowCF(pCtx->pEditor);
	    return;
	}
	if (pCtx->pGlossary != NULL && pCtx->pGlossary->updFlag) {
	    guiNotice(&pCtx->guiCtx, "can't quit; edit mode is set");
	    guiEditorShowCF(pCtx->pGlossary);
	    return;
	}
    }
    else {
	if (pCtx->pEditor != NULL && pCtx->pEditor->updFlag)
	    guiLock(&pCtx->guiCtx, "", pCtx->pEditor->lockFile, 0);
	if (pCtx->pGlossary != NULL && pCtx->pGlossary->updFlag)
	    guiLock(&pCtx->guiCtx, "", pCtx->pGlossary->lockFile, 0);
    }
    while (pCtx->pReqHead != NULL)
	arReqClose(pCtx->pReqHead, destroyFlag);

    if (destroyFlag) {
	if (pCtx->validate_CF != NULL)
	    guiTextswReset(pCtx->validate_TSW);
	xv_destroy_safe(pCtx->guiCtx.baseFrame);
    }
}
