/*	@(#)AR_kbd.c	1.5 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS; changed to use EPICS_ENV
 *				variable to get port number
 * .03	10-02-91	rac	fix minor bugs
 * .04	11-27-91	rac	avoid infinite loop on abort
 * .05	12-04-91	rac	remove use of lightweight processes
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *	-DDEBUG		compile various debug code, including checks on
 *			malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	AR_kbd - EPICS archiver with keyboard interface
*
* DESCRIPTION
*
*
* BUGS
* o	not all signals are caught
*
*-***************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <genDefs.h>
#include <cmdDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>
#include <envDefs.h>


/*/subhead -------------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
void arCmdCrunch();
long arListenCheck();
void arListenSetReopen();
long arInitAtStartup();
void arSigHandler();

/*-----------------------------------------------------------------------------
* global definitions
*----------------------------------------------------------------------------*/
static char	*progName = "AR";
static AR_REQ	*pglCurReq;	/* pointer to current request */
static AR_REQ	*pglOldReq;	/* pointer to previous request */
static char	glArId[80];
static CX_CMD	glCmdCtx;
static CX_CMD	*pglCmdCtx=NULL;
static AR_CTX	glArCtx;
AR_CTX	*pglArCtx=NULL;
static int	glArIPPort;
static int	serverMode=0;  /* 1 says running as server */


/*+/subr**********************************************************************
* NAME	main - AR_kbd main routine
*
* DESCRIPTION
*       % AR              (runs as normal interactive program)
*       % AR server       (runs as a server)
*       % AR server&      (runs as a server)
*       % AR hostName     (runs as a client to AR on hostName)
*
* BUGS
* o	need a place to send output when running detached.  Maybe to AR log.
* o	need a graceful way to wind down operations at exit
* o	doesn't optimize channel access gets and monitors--if several
*	requests get data from the same channel, then several gets or
*	monitors will be done
* o	the use of ca_xxx calls and ECA_ error codes is inconsistent.
*	Ideally, all ca_xxx calls and ECA_ error codes would be hidden in
*	arXxx() routines.
*
*-*/
main(argc, argv)
int     argc;		/* number of command line args */
char   *argv[];		/* command line args */
{
    long	stat;		/* status return from calls */
    FILE	*f;
    int		sigNum;
    int		i;
    char	message[80];
    AR_REQ	*pArReq, *pArReqNext;
    int		syncCount;
    char	portText[10];

    pglArCtx = &glArCtx;
    pglCmdCtx = &glCmdCtx;

    stat = arInitAtStartup(pglArCtx, pglCmdCtx);
    assert(stat == OK);

    if (argc > 2) {
	printf("Usage: %s [server or hostName]\n", argv[0]);
	exit(1);
    }
    else if (argc == 2) {
	if (envGetConfigParam(&EPICS_AR_PORT, 10, portText) == NULL) {
	    printf("error getting %s\n", EPICS_AR_PORT.name);
	    return ERROR;
	}
	sscanf(portText, "%d", &glArIPPort);
	if (strcmp(argv[1], "server") != 0) {
	    execlp("cmdClient", "cmdClient", argv[1], portText, (char *)0);
	    (void)printf("couldn't exec to cmdClient\n");
	    exit(1);
	}
	serverMode = 1;
	pglArCtx->inUseName = argv[1];
/*-----------------------------------------------------------------------------
*       establish a socket to use to listen for potential clients to connect
*----------------------------------------------------------------------------*/
	if (ezsCreateListenSocket(&pglArCtx->listenSock, glArIPPort) != 0) {
	    perror("create listen socket");
	    exit(1);
	}
    }
    else
	pglArCtx->inUseName = "shell";

/*-----------------------------------------------------------------------------
* we're going to be a "real" archiver--either interactive or a server.  Make
* sure there will be only one of us in this directory.
*----------------------------------------------------------------------------*/
    if ((f = fopen("AR.lock", "r")) != NULL) {
	fclose(f);
	(void)printf(
"AR.lock exists.  Either another copy of AR is running or else the lock\n\
file has inadvertently been left (perhaps by an abort).  You will need\n\
to diagnose which.\n");
	exit(1);
    }
    if ((f = fopen("AR.lock", "a")) == NULL) {
	(void)printf("couldn't create AR.lock\n");
	exit(1);
    }
    fclose(f);

    pglCurReq = pglArCtx->pReqMainTail;
    pglOldReq = pglCurReq;

    (void)arLogMessage(progName, "starting", NULL);

    stat = ca_task_initialize();
    assert(stat == ECA_NORMAL);

    genSigInit(arSigHandler);
    if ((sigNum = setjmp(pglArCtx->sigEnv)) != 0) {
        (void)sprintf(message, "signal received: %d", sigNum);
	(void)arLogMessage(progName, message, NULL);
	goto arTaskWrapup;
    }

/*-----------------------------------------------------------------------------
*    in server mode read all enabled requests at startup time
*----------------------------------------------------------------------------*/
    if (serverMode) {
	stat = arReadEnabled();
	if (stat != AR_S_OK)
	    exit(1);
    }

/*----------------------------------------------------------------------------
*    "processing loop"
*---------------------------------------------------------------------------*/
    syncCount = 0;
    while (!pglArCtx->stop) {
	if (serverMode)
	    arListenCheck(&pglCmdCtx);
	if (!serverMode || pglArCtx->inUse) {
	    if (cmdRead(&pglCmdCtx) != NULL) {
		if (serverMode && pglCmdCtx->inputEOF) {
		    pglCmdCtx->pCommand = "close";
		    pglCmdCtx->inputEOF = 0;
		}
		arCmdCrunch(&pglCmdCtx, pglArCtx);
	    }
	}
	stat = ca_pend_event(0.001);
	assert(stat != ECA_EVDISALLOW);
	stat = arReadQCheck();		/* check requests being read */
	stat = arSampleSetCheck(&pglArCtx->stop);
	stat = arMethSetCheck();
	stat = arConnQCheck();		/* check requests being connected */
	stat = arMainQCheck();
	stat = arSampleCheck();
	if (++syncCount >= 100) {
	    syncCount = 0;
	    if (pglArCtx->pChanFileDesc != NULL)
		stat = arCFSyncWrite(pglArCtx->pChanFileDesc);
	}
        fflush(stdout);         	/* flushes needed for socket I/O */
        fflush(stderr);
        fflush(pglCmdCtx->dataOut);
	usleep(100000);		/* sleep .1 second */
    }

arTaskWrapup:
    pglArCtx->stop = 1;
    stat = arSampleSetCheck(&pglArCtx->stop);
    if (pglArCtx->reopenIO) {
	(void)arFreopenAllStd(pglArCtx,pglArCtx->reopenIOoption);
	pglArCtx->reopenIO = 0;
    }
    arListenQuit(&pglCmdCtx);

    pArReq = glArCtx.pReqMainHead;
    while (pArReq != NULL) {
	pArReqNext = pArReq->pNext;
	DoubleListRemove(pArReq, glArCtx.pReqMainHead,
						glArCtx.pReqMainTail);
	stat = arReqDelete(pArReq);
	pArReq = pArReqNext;
    }
    pArReq = glArCtx.pReqConnHead;
    while (pArReq != NULL) {
	pArReqNext = pArReq->pNext;
	DoubleListRemove(pArReq, glArCtx.pReqConnHead,
						glArCtx.pReqConnTail);
	stat = arReqDelete(pArReq);
	pArReq = pArReqNext;
    }
    pArReq = glArCtx.pReqReadHead;
    while (pArReq != NULL) {
	pArReqNext = pArReq->pNext;
	DoubleListRemove(pArReq, glArCtx.pReqReadHead,
						glArCtx.pReqReadTail);
	stat = arReqDelete(pArReq);
	pArReq = pArReqNext;
    }
    if (glArCtx.pChanFileDesc != NULL) {
	stat = arCFClose(&glArCtx.pChanFileDesc);
    }

    stat = ca_task_exit();
    if (stat != ECA_NORMAL)
	(void)printf("ar: ca_task_exit error: %s\n", ca_message(stat));

    if (pglCmdCtx->dataOutRedir)
	fclose(pglCmdCtx->dataOut);

    (void)arLogMessage(progName, "stopping", NULL);

    fflush(stdout);         /* flushes needed for sockets */
    fflush(stderr);

    unlink("AR.lock");

    exit(0);
}

/*+/subr**********************************************************************
* NAME  arSigHandler - signal handling
*
* DESCRIPTION
*       These routines set up for the signals to be caught for arTask
*       and handle the signals when they occur.
*
* RETURNS
*       void
*
* BUGS
* o     not all signals are caught
* o     it's not clear how useful it is to catch the signals which originate
*       from the keyboard
*
*-*/
static void
arSigHandler(signo)
int     signo;
{
    if (signo == SIGPIPE) {
#if 0
	if (arFreopenAllStd(pglArCtx, 0) != OK)
	    assertAlways(0);
	if (!pglArCtx->stop) {
	    (void)printf("client connection broken\n");
	    pglArCtx->arInTaskInfo.stop = 1;
	}
#endif
	return;
    }
    printf("entered arSigHandler for signal:%d\n", signo);
    signal(signo, SIG_DFL);
    longjmp(pglArCtx->sigEnv, signo);
}

/*+/subr**********************************************************************
* NAME	arCmdCrunch - process a command line
*
* DESCRIPTION
*
* RETURNS
*	void
*
* BUGS
* o	should allow "help command" to get more information
*
*-*/
void
arCmdCrunch(ppCmdCtx, pArCtx)
CX_CMD	**ppCmdCtx;	/* IO ptr to pointer to command context */
AR_CTX *pArCtx;	/* IO pointer to ar context */
{
    CX_CMD	*pCmdCtx=*ppCmdCtx;/* local copy of pointer */
    char	message[80];	/* temp for status message */
    int		stat;		/* status return */

    chid	pCh;		/* channel pointer */
    AR_REQ	*pReq;		/* request pointer */
    AR_REQ	*arReqFind();	/* find named request on list */
    char	reqName[AR_NAME_DIM];/* request name */

    if (strcmp(pCmdCtx->pCommand,			"assert0") == 0) {
/*-----------------------------------------------------------------------------
*	under SunOS, this generates SIGABRT
*----------------------------------------------------------------------------*/
	assertAlways(0);
    }
    else if (strcmp(pCmdCtx->pCommand,			"quit") == 0) {
	pArCtx->stop = 1;
	if (serverMode)
	    arListenClose(ppCmdCtx);
    }
    else if (strcmp(pCmdCtx->pCommand,			"trap") == 0) {
	int	*j;
	j = (int *)(-1);
	j = (int *)(*j);
    }
    else if (strcmp(pCmdCtx->pCommand,	"close") == 0) {
	if (*ppCmdCtx != pCmdCtx->pCxCmdRoot)
	    (void)printf("close illegal in source'd file\n");
	else if (strcmp(pglArCtx->inUseName, "shell") == 0)
	    (void)printf("close illegal--AR not running as server\n");
	else if (serverMode) {
	    arListenClose(ppCmdCtx);
	}
    }
    else if (strcmp(pCmdCtx->pCommand,	"dataOut") == 0) {
	if (pCmdCtx->dataOutRedir)
	    fclose (pCmdCtx->dataOut);
	if (nextNonSpaceField( &pCmdCtx->pLine, &pCmdCtx->pField,
					    &pCmdCtx->delim) < 1) {
	    pCmdCtx->dataOut = stdout;
	    pCmdCtx->dataOutRedir = 1;
	}
	else {
	    pCmdCtx->dataOut = fopen(pCmdCtx->pField, "a");
	    if (pCmdCtx->dataOut == NULL) {
		(void)printf("couldn't open %s\n", pCmdCtx->pField);
		pCmdCtx->dataOut = stdout;
		pCmdCtx->dataOutRedir = 0;
	    }
	}
    }
    else if (strcmp(pCmdCtx->pCommand,	"server") == 0) {
	(void)printf( "\"server\" legal only on command line\n");
	(void)printf( "use   help usage   for more information\n");
    }
    else if (strcmp(pCmdCtx->pCommand,	"source") == 0) {
	cmdSource(ppCmdCtx);
    }
    else if (strcmp(pCmdCtx->pCommand, "activate") == 0) {
/*----------------------------------------------------------------------------
* activate - activate a particular archive request
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    (void)printf("usage: activate reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->act.type != AR_METH_DEMAND) {
	    (void)printf("activation type not 'demand'\n");
	    return;
	}
	pReq->act.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand, "create") == 0) {
/*----------------------------------------------------------------------------
* create - create a `by channel' data file with specified size in mBytes
*---------------------------------------------------------------------------*/
	int	mBytes;

	if (pArCtx->pChanFileDesc != NULL)
	    (void)printf("channel file already open\n");
	else {
	    if (nextIntFieldAsInt(&pCmdCtx->pLine, &mBytes, &pCmdCtx->delim) <= 1)
		stat = AR_S_SYNERR;
	    else if (mBytes <= 0 || mBytes > 100)
		stat = AR_S_SYNERR;
	    else {
		pArCtx->pChanFileDesc =
			arCFCreate(AR_FILE_CHAN_PATH, mBytes*1000000);
		if (pArCtx->pChanFileDesc == NULL)
		    (void)printf("couldn't create file\n");
	    }
	}
	if (stat == AR_S_SYNERR)
	    (void)printf("usage: create mBytes\n");
    }
    else if (strcmp(pCmdCtx->pCommand, "current") == 0) {
/*----------------------------------------------------------------------------
* current - change the current request pointer
*	current		reports the name of the current request and the
*			names of the open requests
*	current last	changes to the previously current request
*	current reqName changes to the named request
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    (void)printf("current request is: %s\n", pglCurReq->name);
	    (void)printf("open requests are:\n");
	    pReq = pglArCtx->pReqMainHead;
	    while (pReq != NULL) {
		(void)printf("\t%s\n", pReq->name);
		pReq = pReq->pNext;
	    }
	}    
	else if (strcmp(pCmdCtx->pField, "last") == 0) {
	    pReq = pglCurReq;
	    pglCurReq = pglOldReq;
	    pglOldReq = pReq;
	}
	else {
	    if ((pReq = arReqFind(pCmdCtx->pField)) == NULL) {
		(void)printf("couldn't find: %s\n", pCmdCtx->pField);
		(void)printf("open requests are:\n");
		pReq = pglArCtx->pReqMainHead;
		while (pReq != NULL) {
		    (void)printf("\t%s\n", pReq->name);
		    pReq = pReq->pNext;
		}
	    }    
	    else {
		pglOldReq = pglCurReq;
		pglCurReq = pReq;
	    }
	}    
    }
    else if (strcmp(pCmdCtx->pCommand, "disable") == 0) {
/*----------------------------------------------------------------------------
* disable - disable a particular archive request or all archive requests
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim)<=1) {
	    (void)printf("usage: disable reqName  or  all\n");
	}
	else if (strcmp(pCmdCtx->pField, "all") == 0) {
	    pReq = pglArCtx->pReqMainHead;
	    while (pReq != NULL) {
		stat = arReqDisable(pReq);
		pReq = pReq->pNext;
	    }
	    pReq = pglArCtx->pReqReadHead;
	    while (pReq != NULL) {
		stat = arReqDisable(pReq);
		pReq = pReq->pNext;
	    }
	    pReq = pglArCtx->pReqConnHead;
	    while (pReq != NULL) {
		stat = arReqDisable(pReq);
		pReq = pReq->pNext;
	    }
	}
	else {
	    if ((pReq = arReqFind(pCmdCtx->pField)) == NULL) {
		(void)printf("couldn't find %s\n", pCmdCtx->pField);
		return;
	    }
	    stat = arReqDisable(pReq);
	}
    }
    else if (strcmp(pCmdCtx->pCommand, "enable") == 0) {
/*----------------------------------------------------------------------------
* enable - read an archive request from file in ./arReq/name.req or all
*	the requests in ./arReq/enabled.dat
*	A new request is created.  If everything is successful, then
*	the requests activation method is queued.
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    (void)printf("usage: enable reqName  or  all\n");
	}
	else if (strcmp(pCmdCtx->pField, "all") == 0)
	    stat = arReadEnabled();
	else
	    stat = arReadQAddReq(pCmdCtx->pField);
    }
    else if (strcmp(pCmdCtx->pCommand, "inactivate") == 0) {
/*----------------------------------------------------------------------------
* inactivate - inactivate a particular archive request
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    (void)printf("usage: inactivate reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->inact.type != AR_METH_DEMAND) {
	    (void)printf("inactivation type not 'demand'\n");
	    return;
	}
	pReq->inact.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand, "print") == 0) {
/*----------------------------------------------------------------------------
* print - print information about the current request, or about all
*	requests and the method queues
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    if (pglCurReq == NULL)
		(void)printf("use `current' to select a request to print\n");
	    else
		(void)arReqPrintInfo(pglCurReq, 0);
	}
	else {
	    (void)arMainQPrintInfo(0);
	    (void)arReadQPrintInfo(0);
	    (void)arConnQPrintInfo(0);
	    (void)arMethSetPrintInfo();
	}
    }
    else if (strcmp(pCmdCtx->pCommand, "sample") == 0) {
/*----------------------------------------------------------------------------
* sample - sample a particular archive request
*---------------------------------------------------------------------------*/
	if (nextNonSpaceField(
		    &pCmdCtx->pLine, &pCmdCtx->pField, &pCmdCtx->delim) <= 1) {
	    (void)printf("usage: sample reqName\n");
	    return;
	}
	if ((pReq = arReqFind(pCmdCtx->pField)) == NULL) {
	    (void)printf("couldn't find %s\n", pCmdCtx->pField);
	    return;
	}
	if (pReq->samp.type != AR_METH_DEMAND) {
	    (void)printf("sample type not 'demand'\n");
	    return;
	}
	pReq->samp.demand = 1;
    }
    else if (strcmp(pCmdCtx->pCommand, "status") == 0) {
/*----------------------------------------------------------------------------
* status - print archiver status
*---------------------------------------------------------------------------*/
	(void)arMainQPrintInfo(1);
	(void)arReadQPrintInfo(1);
	(void)arConnQPrintInfo(1);
    }
    else {
/*----------------------------------------------------------------------------
* help (or illegal command)
*----------------------------------------------------------------------------*/
	(void)nextNonSpaceField(&pCmdCtx->pLine, &pCmdCtx->pField,
							&pCmdCtx->delim);
	helpIllegalCommand(stdout, &pCmdCtx->helpList, pCmdCtx->pCommand,
							pCmdCtx->pField);
    }

    return;
}

/*+/subr**********************************************************************
* NAME  arListenCheck - listen for client connections
*
* DESCRIPTION
*
* RETURNS
*
* BUGS
* o     cleanup of listen socket isn't dealt with.  Under SunOS, the system
*       seems to clean up OK.
*
*       In fact, under SunOS, shutdown and/or close seem to inhibit proper
*       cleanup by the system.
*
* SEE ALSO
*
* EXAMPLE
*
*-*/
long
arListenCheck(ppCmdCtx)
CX_CMD  **ppCmdCtx;      /* IO ptr to pointer to command context */
{
    long        stat;
    int         sigNum;
    char        clientName[30];
    static char sockPrompt[80];
    char	message[80];

/*-----------------------------------------------------------------------------
*       check to see if a connect attempt has been made.  If the server is
*       already in use by a client, connect attempts will be rejected.
*----------------------------------------------------------------------------*/
    if (ezsListenExclusiveUse(&pglArCtx->clientSock,
                            &pglArCtx->listenSock, &pglArCtx->inUse,
                            clientName, glArId) < 0) {
	perror("check for connect attempt");
	exit(1);
    }

    if (pglArCtx->inUse && pglArCtx->inUseInit == 0) {
/*-----------------------------------------------------------------------------
*       When a connect attempt succeeds, change stdin and stdout to use
*       the socket, so that server I/O (i.e., I/O for this program) will
*       automatically be re-directed to the client.
*
*	The prompt is changed to a special prompt which ends with \n.  This
*	is done to avoid buffer flushing problems on some operating systems.
*	A special prefix of "#p#r#" is added to the normal prompt so that
*	the client knows that some massaging needs to be done before
*	displaying the prompt to the user.
*----------------------------------------------------------------------------*/
	(void)sprintf(message, "%s connected", clientName);
	(void)arLogMessage(progName, message, NULL);
	if (arFreopenAllStd(pglArCtx, 1) != OK)
	    assertAlways(0);
	pglArCtx->oldPrompt = (*ppCmdCtx)->prompt;
	assert(strlen(pglArCtx->oldPrompt) < 72);
	(void)sprintf(sockPrompt, "#p#r#%s\n", pglArCtx->oldPrompt);
	(*ppCmdCtx)->prompt = sockPrompt;
	arListenSetReopen(2);	/* reopen all other tasks to socket */
	pglArCtx->inUseInit = 1;
	(*ppCmdCtx)->promptFlag = 1;
    }
}

/*+/subr**********************************************************************
* NAME	arListenClose - close a connection to a client
*
*-*/
long
arListenClose(ppCmdCtx)
CX_CMD  **ppCmdCtx;      /* IO ptr to pointer to command context */
{
    char	message[80];

    close(pglArCtx->clientSock);
    pglArCtx->clientSock = -1;
    pglArCtx->inUse = 0;
    pglArCtx->inUseInit = 0;
    if (arFreopenAllStd(pglArCtx, 0) != OK)
	assertAlways(0);
    (void)sprintf(message, "%s disconnected", pglArCtx->inUseName);
    (void)arLogMessage(progName, message, NULL);
    (*ppCmdCtx)->prompt = pglArCtx->oldPrompt;
    arListenSetReopen(0);	/* reopen fd's to std... */
    return OK;
}

/*+/subr**********************************************************************
* NAME	arListenQuit - quit listening for client connect attempts
*
*-*/
long
arListenQuit(ppCmdCtx)
CX_CMD  **ppCmdCtx;      /* IO ptr to pointer to command context */
{
    if (pglArCtx->clientSock != -1) {
        shutdown(pglArCtx->clientSock, 2);
        close(pglArCtx->clientSock);
    }

    close(pglArCtx->listenSock);
    return OK;
}

/*+/subr**********************************************************************
* NAME	arListenSetReopen - set flag to reopen std... fd's
*
*-*/
void
arListenSetReopen(option)
int	option;		/* I option for reopen for use with arFreopenAllStd */
{
    pglArCtx->reopenIOoption = option;
    pglArCtx->reopenIO = 1;
}

/*+/subr**********************************************************************
* NAME  arFreopenAllStd - reopen stdin, stdout, and stderr
*
* DESCRIPTION
*
* RETURNS
*       OK, or
*       ERROR   (an error message has been printed)
*
*-*/
static int
arFreopenAllStd(pCtx, option)
AR_CTX *pCtx;		/* IO pointer to AR context */
int     option;		/* I activity selector:
			0  set back to use old fd's
			1  use client socket, save old fd's
			2  use client socket, don't save old fd's */
{
    if (option == 0) {
        if (ezsFreopenToFd(stdin, &pCtx->oldStdinFd, NULL) < 0) {
            perror("restore stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->oldStdoutFd, NULL) < 0) {
            perror("restore stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->oldStderrFd, NULL) < 0) {
            perror("restore stderr");
            return ERROR;
        }
    }    
    else if (option == 1) {
        if (ezsFreopenToFd(stdin, &pCtx->clientSock, &pCtx->oldStdinFd) < 0) {
            perror("reopen stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->clientSock, &pCtx->oldStdoutFd) < 0) {            perror("reopen stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->clientSock, &pCtx->oldStderrFd) < 0) {            perror("reopen stderr");
            return ERROR;
        }
    }
    else if (option == 2) {
        if (ezsFreopenToFd(stdin, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stdin");
            return ERROR;
        }
        if (ezsFreopenToFd(stdout, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stdout");
            return ERROR;
        }
        if (ezsFreopenToFd(stderr, &pCtx->clientSock, NULL) < 0) {
            perror("reopen stderr");
            return ERROR;
        }
    }    
    return OK;
}

/*+/subr**********************************************************************
* NAME	arInitAtStartup - initialization for AR
*
* DESCRIPTION
*	Perform several initialization duties:
*	o   initialize the AR context
*	o   initialize the command context block
*	o   initialize the help information
*
* RETURNS
*	OK, or
*	ERROR
*
*-*/
long
arInitAtStartup(pArCtx, pCmdCtx)
AR_CTX	*pArCtx;
CX_CMD	*pCmdCtx;
{
    long	retStat=OK;
    long	stat;
    AR_REQ	*pArReq;

    strcpy(glArId, progName);
    pArCtx->progName = progName;
    pArCtx->pReqMainHead = NULL;
    pArCtx->pReqMainTail = NULL;
    pArCtx->pReqConnHead = NULL;
    pArCtx->pReqConnTail = NULL;
    pArCtx->pReqReadHead = NULL;
    pArCtx->pReqReadTail = NULL;
    pArCtx->pMethClkHead = NULL;
    pArCtx->pMethClkTail = NULL;
    pArCtx->pMethEvHead = NULL;
    pArCtx->pMethEvTail = NULL;
    pArCtx->pMethMiscHead = NULL;
    pArCtx->pMethMiscTail = NULL;
    pArCtx->pChanFileDesc = NULL;

    pArCtx->stop = 0;
    pArCtx->inUse = 0;
    pArCtx->inUseInit = 0;
    pArCtx->listenSock = -1;
    pArCtx->clientSock = -1;
    pArCtx->oldStdinFd = -1;
    pArCtx->oldStdoutFd = -1;
    pArCtx->oldStderrFd = -1;

    cmdInitContext(pCmdCtx, "  AR:  ");

/*-----------------------------------------------------------------------------
* help information initialization
*----------------------------------------------------------------------------*/
    helpInit(&pCmdCtx->helpList);
/*-----------------------------------------------------------------------------
* help info--legal generic commands
*----------------------------------------------------------------------------*/
	helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpCmds, "commands", "\n\
Generic commands are:\n\
   close                     (use    help usage   for more info)\n\
   dataOut    [filePath]     (default is to normal output)\n\
   help       [topic]\n\
   quit (or ^D)              (use    help usage   for more info)\n\
   source     filePath\n\
");
/*-----------------------------------------------------------------------------
* help info--bg command
*----------------------------------------------------------------------------*/
    helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpBg, "bg", "\n\
Under SunOS, no bg command is directly available.  Instead, if AR\n\
is being run from the C shell (csh), it can be put in the background\n\
using several steps from the keyboard (with the first % on each line being\n\
the prompt from csh):\n\
        type ^Z, then type\n\
        % bg %AR\n\
\n\
To move AR to the foreground, type\n\
        % fg %AR\n\
");
/*-----------------------------------------------------------------------------
* help info--legal AR-specific commands
*----------------------------------------------------------------------------*/
	helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpCmdsSpec, "commands", "\n\
AR-specific commands are:\n\
    activate    reqName;\n\
    clear\n\
    create      mBytes\n\
    current     [reqName]\n\
    current     last\n\
    disable     reqName  or  all\n\
    enable      reqName  or  all\n\
    help        [command]\n\
    inactivate  reqName;\n\
    print       [all]\n\
    sample      reqName\n\
    status\n\
\n\
Output from commands flagged with * can be routed to a file by using the\n\
\"dataOut filePath\" command.  The present contents of the file are\n\
preserved, with new output being written at the end.\n\
");
/*-----------------------------------------------------------------------------
* help info--AR usage information
*----------------------------------------------------------------------------*/
	helpTopicAdd(&pCmdCtx->helpList, &pCmdCtx->helpUsage, "usage", "\n\
AR (the data archiver) acquires data from channel access and saves the\n\
data to disk.  (The companion program ARR (archive retriever) can be used\n\
to retrieve the data from disk for printing, plotting, etc.)\n\
\n\
For information on moving AR into the background, use the \"help bg\"\n\
command.\n\
\n\
AR can be run as a server by:\n\
    % AR_kbd server\n\
\n\
AR can be run as a client connecting to an AR server on \"hostName\" by:\n\
    % AR_kbd hostName\n\
\n\
    When running AR as a client, two commands are of special interest:\n\
\n\
    close         shuts down the client but leaves the server running\n\
    quit (or ^D)  shuts down both the client and the server\n\
");

    return retStat;
}
