/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef INC_arChanIO_h
#define INC_arChanIO_h
/*	arChanIO.h,v 1.2 2002/08/02 15:37:53 jba Exp
 *	Author:	Roger A. Cole
 *	Date:	03-09-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 * make options
 *	-DvxWorks	makes a version for VxWorks
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arChanIO.h - AR channel data file definitions
*
* DESCRIPTION
*	Provide `private' definitions related to AR channel data files.
*	The definitions in this file aren't intended for general use.
*	
* SEE ALSO
*	bfDefs.h	for information about block0 and the block file
*			routines.
*	arAccessLib.h	for some channel I/O definitions which are included
*			there (rather than here) because they are needed by
*			programs which use the AR access library.
*	
*-***************************************************************************/

#ifndef INC_genDefs_h
#   include <genDefs.h>
#endif
#ifndef INC_tsDefs_h
#   include <tsDefs.h>
#endif
#ifndef INC_bfDefs_h
#   include <bfDefs.h>
#endif
#ifndef INC_arCS_h
#   include <arCS.h>
#endif
#ifndef INCLdb_accessh
#   include <db_access.h>
#endif


/*----------------------------------------------------------------------------
* channel data file--organization
*
*    The AR channel data file supports storing and retrieving data for
*    GTACS channels.  Within the file, each channel exists as a sub-file
*    which can be opened, closed, read, written, etc.
*
*    The objective for each sub-file is to store enough information to
*    properly respond to requests for historical data with request types
*    of DBR_xxx, DBR_STS_xxx, DBR_TIME_xxx, and DBR_GR_xxx.
*
*    The AR channel data file is based on the block files implemented by
*    the bfXxx() subroutines.  The channel data file is organized as a
*    tree with blocks at 4 levels.  Within the context of a block file,
*    pointers are simply block numbers.  Block0 is the root of the tree; it
*    contains pointers to the head and tail of the linked list of master
*    index blocks.  Each master index block contains several chanHdr items;
*    a chanHdr points to the head and tail of the linked list of index
*    blocks for a channel.  Each index block for a channel contains several
*    datInfo items; a datInfo points to a block containing actual data for
*    a channel.  The data blocks for a channel contain a stream of data
*    items intermixed with date and time items.
*
*    One detail which must be kept in mind is that all blocks in a block
*    file begin with 3 items.  In defining the structures for the AR channel
*    data file blocks, these 3 items are kept in a bfInfo sub-structure:
*    bfInfo.flink is the forward link to the next block in a linked list of
*    blocks; bfInfo.firstByte is the byte number (relative to byte 0 of the
*    block) of the first byte of data; and bfInfo.lastByte is the relative
*    byte number of the last byte of data.  The discussion for each type of
*    block explains the actual usage of these fields for that type of block.
*
*    The various types of blocks are discussed in some detail below, and
*    in greater detail in the sections which actually define the structure
*    for the block.
*
*    o  MI blocks (master index) allow finding channels.  A master
*	index block contains several 'chanHdr' structures.  (chanHdr
*	structures aren't allowed to overlap MI block boundaries.  This
*	allows using a chanHdr 'in-situ' in a buffer; if it could overlap
*	block boundaries, then it would have to be copied into an actual
*	structure for use.)  Each channel which has data stored in the file
*	has a corresponding chanHdr structure in the master index; the
*	chanHdr for a channel tells the time span of the available data
*	for a channel and has pointers which allow finding the actual data.
*
*	block0 has pointers to the head and tail of the linked list of MI
*	blocks.  The bfInfo.flink is used to traverse the list of MI blocks.
*
*    o  index blocks allow finding data blocks for a channel.  An index block
*	contains several 'datInfo' structures.  (datInfo structures aren't
*	allowed to overlap index block boundaries.  See the comment above
*	for the reason for this restriction.)  Each data block for the
*	channel has a datInfo structure in the index; the datInfo for a block
*	tells the date and time of the first data item in the block and has
*	a pointer to the data block.
*
*	The chanHdr (in an MI block) for a channel has pointers to the head
*	and tail of the linked list of index blocks for the channel.  The
*	bfInfo.flink is used to traverse the list of index blocks for a
*	channel.
*
*    o  data blocks contain actual data.  There are some restrictions on
*	crossing block boundaries with data (the restrictions are to allow
*	releasing blocks of old data without losing the integrity of the
*	date and time in the new 'first block' of data):
*	-  time stamps can't cross block boundaries
*	Aside from these restrictions, and in conjunction with firstByte and
*	lastByte, these blocks implement a stream of data related items.
*	Items are recognized by 'item codes' which occupy part or all of the
*	first byte of an item.
*
*	The datInfo (in an index block) for a data block has a pointer to the
*	data block.  Traversal of the data blocks for a channel is done by
*	traversing the index blocks for the channel and processing the datInfo
*	structures one by one; datInfo.flink in data blocks not used.
*
* blocks in memory
*
*    Except for block0, memory buffers for blocks are managed by arCFxxx
*    routines.  The memory buffers contain several fields, referred to as
*    a preamble, followed by the actual block of data.  Each block type has
*    its own typedef.
*
*---------------------------------------------------------------------------*/

/*/subhead block0-------------------------------------------------------------
* block0, user area definitions:
*
*    [0]  block number of first master index block
*    [1]  block number of last master index block
*    [2]  number of blocks in master index 'chain'
*    [3]  block number of the master index block containing the first
*	unused 'chanHdr' item
*    [4]  chanHdr subscript in MI block for first unused chanHdr item
*    [5]  revision level of file format
*---------------------------------------------------------------------------*/
#define ArB0MIHead(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[0]
#define ArB0MITail(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[1]
#define ArB0MINblk(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[2]
#define ArB0MIFree_blkNum(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[3]
#define ArB0MIFree_hdrNum(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[4]
#define ArB0FormatRev(pArCfDesc)	pArCfDesc->pBfDesc->pBlock0->user[5]


/*/subhead misc---------------------------------------------------------------
* AR_CF_BLKSIZE
*
*    This is the size to use for the blocks.  All blocks will be this size.
*    arCFChanOpen() insists that any file it opens have this block size.
*    Various parts of the code depend on block size, and changing it should
*    be done with care--current files must be thrown away if block size
*    changes.
*
* AR_CF_FORMAT_REV
*
*    This is a revision number for file format.  The `hundreds' digits
*    indicate a major revision change, with no compatibility.  The
*    low order digits indicate minor revisions, with partial compatibility.
*
*    For partial compatibility, new code can still read old format files,
*    (of the same major revision), but old code can't read the new format.
*---------------------------------------------------------------------------*/
#define AR_CF_BLKSIZE 1032
#define AR_CF_FORMAT_REV 400

#define PREAMBLE \
    void	*pWriteNext;	/* next block in "write list" */  \
    void	*pWritePrev;	/* previous block in "write list" */  \
    BF_BLOCK	*pBlock;	/* pointer to data block */  \
    BF_BLKNUM	blkNum;		/* block number */  \
    char	modified;	/* 1 if block modified, else 0 */  \
    char	padP0, padP1, padP2;  \
    long	padP3


/*/subhead AR_GR_BUF-----------------------------------------------------------
*AR_GR_BUF - graphics information buffer
*
*	A graphics information buffer in memory consists of two parts: a
*	'preamble' with run time information; and a 'body' which contains
*	the actual disk block.  The preamble is padded so that the body
*	begins on an 8 byte boundary.
*
*	Presently, a graphics information block is used only for DBF_ENUM
*	channels to hold the state string information.  Although, depending
*	on block size, there would be room for the graphics information for
*	several DBF_ENUM channels, a block is used for a single channel,
*	thus avoiding a bunch of complication.
*----------------------------------------------------------------------------*/
#define AR_GR_BUF_PAD \
    (AR_CF_BLKSIZE-BF_BLOCK_DATA-(int)sizeof(short)-db_state_dim*db_state_text_dim)

typedef struct arGrBuf {
    PREAMBLE;
    BF_BLOCK	bfInfo;		/* actual block start: block file info */
    short	numStrings;	/* number of state strings */
    char	states[db_state_dim][db_state_text_dim];/* states */
    char	pad[AR_GR_BUF_PAD];
} AR_GR_BUF;

/*/subhead AR_CHAN_HDR--------------------------------------------------------
*    some structures for use in the chanHdr structure
*---------------------------------------------------------------------------*/
struct arChFltGr {
    char	units[AR_UNITS_DIM];	/* units (must be first!!) */
    short	precision;		/* precision (must be second!!) */
    short	pad;
    float	upDispLim;		/* upper display limit */
    float	lowDispLim;		/* lower display limit */
    float	upAlmLim;		/* upper alarm limit */
    float	upWarnLim;		/* upper warning limit */
    float	lowWarnLim;		/* lower warning limit */
    float	lowAlmLim;		/* lower alarm limit */
};
struct arChShortGr {
    char	units[AR_UNITS_DIM];	/* units (must be first!!) */
    short	upDispLim;		/* upper display limit */
    short	lowDispLim;		/* lower display limit */
    short	upAlmLim;		/* upper alarm limit */
    short	upWarnLim;		/* upper warning limit */
    short	lowWarnLim;		/* lower warning limit */
    short	lowAlmLim;		/* lower alarm limit */
};
struct arChEnumGr {
    BF_BLKNUM	block;			/* block with graphics info */
    AR_GR_BUF	*pGRBuf;		/* pointer to buffer */
};
struct arChCharGr {
    char	units[AR_UNITS_DIM];	/* units (must be first!!) */
    char	upDispLim;		/* upper display limit */
    char	lowDispLim;		/* lower display limit */
    char	upAlmLim;		/* upper alarm limit */
    char	upWarnLim;		/* upper warning limit */
    char	lowWarnLim;		/* lower warning limit */
    char	lowAlmLim;		/* lower alarm limit */
    short	pad;
};
struct arChLongGr {
    char	units[AR_UNITS_DIM];	/* units (must be first!!) */
    long	upDispLim;		/* upper display limit */
    long	lowDispLim;		/* lower display limit */
    long	upAlmLim;		/* upper alarm limit */
    long	upWarnLim;		/* upper warning limit */
    long	lowWarnLim;		/* lower warning limit */
    long	lowAlmLim;		/* lower alarm limit */
};
struct arChDoubleGr {
    char	units[AR_UNITS_DIM];	/* units (must be first!!) */
    short	precision;		/* precision (must be second!!) */
    short	pad0;
    long	pad1;
    double	upDispLim;		/* upper display limit */
    double	lowDispLim;		/* lower display limit */
    double	upAlmLim;		/* upper alarm limit */
    double	upWarnLim;		/* upper warning limit */
    double	lowWarnLim;		/* lower warning limit */
    double	lowAlmLim;		/* lower alarm limit */
};
struct arChMIFree {
    BF_BLKNUM	blkNum;			/* blk # next free chanHdr */
    BF_BLKNUM	hdrNum;			/* chanHdr # of next free */
};

/*----------------------------------------------------------------------------
* AR_CHAN_HDR - channel header in AR channel data file
*
*    The chanHdr for a channel contains some general information about the
*    channel and some information about how to find the channel's data in the
*    file.
*
*    It is possible for a chanHdr to be 'empty'.  This might occur, for
*    example, when a file is purged of 'old' data.  To avoid 'losing' space
*    in the master index chain, these empty chanHdr's are linked together
*    as a list of 'free chanHdr's', with the pointer to the head of this
*    list being stored in the user information in block0.  A pointer in this
*    list consists of a block number and a byte number within the block.  In
*    memory, the is list handled as the address of the first chanHdr available;
*    when this chanHdr is allocated, then (sometimes) a search must be made
*    of the MI blocks in memory to obtain the address of the next free
*    chanHdr.  (Since adding a new channel to a file is a relatively rare
*    occurrence, and since it occurs not in real time, the penalty of doing a
*    search seems preferable to the complexity necessary to avoid the search.)
*
*    An 'empty' chanHdr has '\0' as the first byte of name.
*
* MACROS
*    ArCDChanHdr(pChanDesc)	selects the specific chanHdr in use by
*				a channel descriptor.  Example usage:
*				    i = ArCDChanHdr(pChanDesc).elCount;
*
*---------------------------------------------------------------------------*/

typedef struct {
    char	name[AR_NAME_DIM];	/* name of channel; can be either
					recordName or recordName.fieldName .
					If name[0]=='\0', chanHdr is free */
    short	chanType;		/* type of channel, one of DBF_xxx */
    unsigned short elCount;		/* element count; 1 for scalar
					channels, > 1 for 'array channels' */
    union {
	struct arChMIFree nextMIFree;	/* info for free chanHdr */
        struct arChFltGr floatGr;	/* graphics info for 'float' fields */
        struct arChShortGr shortGr;	/* graphics info for 'short' fields */
        struct arChCharGr charGr;	/* graphics info for 'char' fields */
        struct arChEnumGr enumGr;	/* graphics info for 'enum' fields */
        struct arChLongGr longGr;	/* graphics info for 'long' fields */
        struct arChDoubleGr doubleGr;	/* graphics info for 'double' fields */
    } graphics;
    BF_BLKNUM	indexHead;		/* block # of first index block */
    BF_BLKNUM	indexTail;		/* block # of last index block */
    ULONG	oldestSecPastEpoch;	/* `stamp' for oldest data */
    ULONG	newestSecPastEpoch;	/* `stamp' for newest data */
    ULONG	newestNsec;		/* rest of `stamp' for newest data */
    USHORT	flags;			/* flags--combination of AR_CHDR_xxx */
#define		    AR_CHDR_WRITE 1	/* chanHdr has a writer active to it */
    short	count;			/* # of chanDesc open to channel */
} AR_CHAN_HDR;


/*/subhead AR_MI_BUF----------------------------------------------------------
* AR_MI_BUF - master index buffer
*
*    A master index buffer in memory consists of two parts: a 'preamble' with
*    run time information; and a 'body' which contains the actual disk block.
*
*    A master index block will ALWAYS have as many chanHdr as will fit in
*    it--unused chanHdr don't quit using space, but are just linked into the
*    free chanHdr list
*
*---------------------------------------------------------------------------*/

#define AR_CF_NHDR	((AR_CF_BLKSIZE-BF_BLOCK_DATA)/(int)sizeof(AR_CHAN_HDR))
#define AR_MI_BUF_PAD	((AR_CF_BLKSIZE-BF_BLOCK_DATA)%(int)sizeof(AR_CHAN_HDR))

typedef struct arMiBuf {
    PREAMBLE;			/* common runtime info; bfInfo must follow */
    BF_BLOCK	bfInfo;		/* actual block start: block file info */
    AR_CHAN_HDR chanHdr[AR_CF_NHDR];/* channel headers */
    char	pad[AR_MI_BUF_PAD?AR_MI_BUF_PAD:4];
    struct arMiBuf *pNextMI;	/* pointer to next MI buffer in memory */
    long	pad3;
} AR_MI_BUF;

/*/subhead AR_INDEX_BUF-------------------------------------------------------
* AR_INDEX_BUF - index buffer
*
*    An index buffer in memory consists of two parts: a 'preamble' with
*    run time information; and a 'body' which contains the actual disk block.
*
*    an index block will have a variable number of datInfo structures in
*    use.  The ones which are actually meaningful can be determined from
*    bfInfo.firstByte and bfInfo.lastbyte .
*
* MACROS
*    ArCDDatInfo(pChanDesc)	selects the datInfo currently in use by the
*				channel descriptor.  Example usage:
*				    block = ArCDDatInfo(pChanDesc).dataBlock;
*---------------------------------------------------------------------------*/
typedef struct {
    BF_BLKNUM	dataBlock;	/* block number for data */
    TS_STAMP	stamp;		/* time stamp for first data in block */
} AR_DAT_INFO;

#define AR_CF_NDATINFO ((AR_CF_BLKSIZE-BF_BLOCK_DATA)/(int)sizeof(AR_DAT_INFO))
#define AR_INDEX_BUF_PAD ((AR_CF_BLKSIZE-BF_BLOCK_DATA)%(int)sizeof(AR_DAT_INFO))

typedef struct {
    PREAMBLE;			/* common runtime info; bfInfo must follow */
    BF_BLOCK	bfInfo;		/* actual block start: block file info */
    AR_DAT_INFO datInfo[AR_CF_NDATINFO];/* datInfo structures */
    char	pad[AR_INDEX_BUF_PAD?AR_INDEX_BUF_PAD:4];
} AR_INDEX_BUF;

/*/subhead AR_DATA_BUF--------------------------------------------------------
* AR_DATA_BUF - data buffer
*
*    A data buffer in memory consists of two parts: a 'preamble' with
*    run time information; and a 'body' which contains the actual disk block.
*
*    Data buffers for a channel are used to store a stream of items for
*    the channel.  The items include `time information' items and `value'
*    items.  An item is identified by an item code in the first byte of
*    the item.
*
*    bfInfo.lastByte indicates the last used byte in the disk block, relative
*    to bfInfo.  If bfInfo.lastByte < BF_BLOCK_DATA, then the block is empty.
*
*    bfInfo.firstByte indicates the first byte, relative to bfInfo, of the
*    first data item which begins in the block.  If the block doesn't
*    contain the beginning of a data item, then bfInfo.firstByte will be
*    BF_BLOCK_DATA-1  .
*
*    The time stamp for the first data item which begins in a block
*    uses the time in the datInfo structure for a reference.
*
*---------------------------------------------------------------------------*/
#define AR_DAT_IC_VAL_XXX	0x80
#define AR_DAT_IC_VAL_STAT	0x80	/* value with status */
#define AR_DAT_IC_VAL		0x81	/* value without status */
#define AR_DAT_IC_TIME_XXX	0x90
#define AR_DAT_IC_TIME_STAMP	0x90	/* time stamp item.  5 bytes.  This
					sets time to whole seconds, as
					seconds past epoch, GMT.  The epoch
					is defined in tsDefs.h, and isn't
					necessarily same as system epoch */
#define AR_DAT_IC_TIME_USEC	0x91	/* delta time (usec) item.  4 bytes
					This adds the delta time to the
					present time stamp.  This can be as
					large as 15000000; if delta time is
					larger, a new time stamp will be
					issued. */
#define AR_DAT_IC_TIME_NSEC	0x92	/* delta time (nsec) item.  5 bytes
					This adds the delta time to the
					present time stamp.  This can be as
					large as 4000000000; if delta time is
					larger, a new time stamp will be
					issued. */
#define AR_DAT_IC_TIME_MSEC	0	/* delta time (msec) item.  2 bytes
					This is a `fake' item code--this
					item is identified by byte&0x80 == 0.
					This adds the delta time to the
					present time stamp.  This can be as
					large as 30000; if delta time is
					larger, a new time stamp will be
					issued. */
#define AR_DAT_IC_MON_XXX	0xa0
#define AR_DAT_IC_MON_BEGIN	0xa0
#define AR_DAT_IC_MON_INACT	0xa1
#define AR_DAT_IC_MON_DISCON	0xa2
#define AR_DAT_IC_MON_HIATUS	0xa3
#define AR_DAT_IC_SNAP_BEGIN	0xa4
#define AR_DAT_IC_SNAP_SINGLE	0xa5
#define AR_DAT_IC_SNAP_END	0xa6

typedef struct {
    PREAMBLE;			/* common runtime info; bfInfo must follow */
    BF_BLOCK	bfInfo;		/* actual block start: block file info */
    char	data[AR_CF_BLKSIZE-BF_BLOCK_DATA];/* data bytes */
} AR_DATA_BUF;

/*----------------------------------------------------------------------------
* macros for use by arCFChanWrite() and friends
*
*    int ArCFStore(pDesc, c)    insert character into current write buffer; if
*                               there isn't room (or if there isn't a current
*                               buffer) flush current buffer and insert
*                               character into new buffer
*
*    int ArCFFlushAndStore(pDesc, c)	flush current write buffer (if any)
*                               and insert character into new buffer
*    void ArCFModifyInit(pBuf)		initialize the `write-if-modified'
*				logic in the buffer
*    void ArCFModifySet(pCFDesc, pBuf)	if the buffer hasn't already been
*				flagged as modified, set the `modified' flag
*				and add the buffer to the `write list' for
*				the channel file descriptor
*    int ArCFModifyTestAndReset(pCFDesc, pBuf) if the buffer's `modified'
*				flag is set, remove the buffer from the
*				`write list' for the channel file descriptor
*				and set the `modified' flag to 0.  The
*				original setting of the `modified' flag is
*				returned.
*---------------------------------------------------------------------------*/
#define ArCFFlushAndStore(c, pDesc) \
          arCF_DFlushWbufAndPutc((c), pDesc)

#define ArCFStore(c, pDesc) \
        (  (--pDesc->remainCount >= 0) \
                    ? (  *(++(pDesc->pData)) = (c), OK) \
                    :   ArCFFlushAndStore((c), pDesc) \
              )

#define ArCFModifyInit(pBuf) \
    (pBuf)->modified = 0, \
    (pBuf)->pWriteNext = NULL, \
    (pBuf)->pWritePrev = NULL

#define ArCFModifySet(pCFDesc, pBuf) \
    assert(pBuf->bfInfo.lastByte > 0 && pBuf->bfInfo.lastByte < 2000); \
    if (!(pBuf)->modified) { \
        (pBuf)->modified = 1; \
	(pBuf)->pWriteNext = NULL;\
	(pBuf)->pWritePrev = pCFDesc->pWriteTail;\
	if (pCFDesc->pWriteTail != NULL)\
	    ((AR_MI_BUF *)pCFDesc->pWriteTail)->pWriteNext = pBuf;\
	pCFDesc->pWriteTail = pBuf;\
	if (pCFDesc->pWriteHead == NULL)\
	    pCFDesc->pWriteHead = pBuf;\
    }

#define ArCFModifyTestAndReset(pCFDesc, pBuf) \
  ( (pBuf)->modified \
    ? ( (pBuf)->modified = 0, \
	assert(pBuf->bfInfo.lastByte > 0 && pBuf->bfInfo.lastByte < 2000), \
	((pBuf)->pWritePrev != NULL \
	    ?  (((AR_MI_BUF *)((pBuf)->pWritePrev))->pWriteNext = \
							(pBuf)->pWriteNext) \
	    :  (pCFDesc->pWriteHead = (pBuf)->pWriteNext) \
	), \
	((pBuf)->pWriteNext != NULL \
	    ?  (((AR_MI_BUF *)((pBuf)->pWriteNext))->pWritePrev = \
							(pBuf)->pWritePrev) \
	    :  (pCFDesc->pWriteTail = (pBuf)->pWritePrev) \
	), \
      (pBuf)->pWriteNext = NULL, \
      (pBuf)->pWritePrev = NULL, \
      1) \
    : 0 \
  )



/*/subhead prototypes----------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
long  arCFAsserts();
long  arCF_DFillRbufAndGetc();
long  arCF_DFlushAndFree();
long  arCF_DFlushWbufAndPutc();
long  arCF_DNextBlock();
long  arCF_DRead();

long  arCF_GRGetBlock();
long  arCF_GRReadBlock();

long  arCF_IFlushAndFree();
long  arCF_INextDatInfo();
long  arCF_IRead();

long  arCF_MIAddBlock();
long  arCF_MIAddChan();
long  arCF_MIDelChan();
long  arCF_MIFindChan();
long  arCF_MIFlush();

void arFree();
void *arMalloc();
void arMallocDebug();
void arMallocStats();

AR_MI_BUF *arCF_mallocMI_BUF();
AR_INDEX_BUF *arCF_mallocINDEX_BUF();
AR_DATA_BUF *arCF_mallocDATA_BUF();
AR_GR_BUF *arCF_mallocGR_BUF();

void arCF_freeMI_BUF();
void arCF_freeINDEX_BUF();
void arCF_freeDATA_BUF();
void arCF_freeGR_BUF();


#endif
