/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef INC_bfDefs_h
#define INC_bfDefs_h
/*	bfDefs.h,v 1.2 2002/08/02 15:37:54 jba Exp
 *	Author:	Roger A. Cole
 *	Date:	03-05-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 * make options
 *	-DvxWorks	makes a version for VxWorks
 */
/*+/mod***********************************************************************
* TITLE	bfDefs.h - 'block file' definitions
*
* DESCRIPTION
*	This file contains various definitions used by the bfXxx code and
*	by routines which use the bfXxx() routines.
*
*   TYPEDEFS
*
*	BF_BLKNUM  is the declaration for a block number
*	BF_BLOCK0  is a template for block0 of a block file
*	BF_BLOCK   is a template for the generic initial part of a block
*			in a block file--this doesn't include any space
*			for data
*	BF_DESC    is a file descriptor for use with block file routines
*
* SEE ALSO:
* o	bfTest.c	implements a 'shell' for testing the block file
*			routines.  This may serve as a starting point
*			for learning how to use the routines.
* o	bfSubr.c	implements the actual block file routines
*	
*-***************************************************************************/

#ifndef INC_genDefs_h
#   include <genDefs.h>
#endif

typedef ULONG BF_BLKNUM;	/* block number within file */

/*----------------------------------------------------------------------------
*    includes and definitions for VxWorks compile
*---------------------------------------------------------------------------*/
#ifdef vxWorks
#   include <vxWorks.h>
#else
/*----------------------------------------------------------------------------
*    includes and definitions for Sun compile
*---------------------------------------------------------------------------*/

#endif

/*----------------------------------------------------------------------------
* general - general use definitions related to block files
*---------------------------------------------------------------------------*/


#define BF_MAX_BLOCK_SIZE 4095	/* max block size, in bytes--12 bits */
#define BF_MIN_BLOCK_SIZE 200	/* min block size, in bytes */
#define BF_MAX_BLOCKS 648575	/* max file size, in blocks--20 bits */
#define BF_MIN_BLOCKS 1		/* min file size, in blocks */


/*----------------------------------------------------------------------------
* BF_BLOCK0 - definitions for block 0
*---------------------------------------------------------------------------*/
struct fileInfo {
    long	blockSize;	/* size of file's blocks, in bytes */
    ULONG	flags;		/* file flags, from BF_FLAG_XXX */
    BF_BLKNUM	maxSize;	/* maximum size of file, in blocks */
    BF_BLKNUM	currentSize;	/* present size of file, in blocks */
    BF_BLKNUM	currentFree;	/* present number of free blocks */
    long	fileType;	/* type of file--one of BF_TYPE_xxx */
};

#define BF_FLAG_GROW 1		/* file size can increase dynamically */
#define BF_FLAG_BIN 2		/* file contains binary information */

#define BF_USRINF_DIM 16

#define BfTypeToText(type) \
	(assert(type>=0 && type<(int)BF_TYPE_PAST), glBfTypeKw[type])

#ifndef BF_TEXT_GLBLSOURCE
    extern char *glBfTypeKw[];
#else
    char *glBfTypeKw[] = {
	/* BF_TYPE_UNDEF    */      "undefined",
	/* BF_TYPE_RING     */      "ringFile",
	/* BF_TYPE_CHAN_DATA*/      "ARChannelData",
	/* BF_TYPE_USER     */      "userDefined",
	/* BF_TYPE_PAST     */      "illegal"
    };
#endif

enum bfType {
    BF_TYPE_UNDEF,		/* MUST BE ZERO */
    BF_TYPE_RING,
    BF_TYPE_CHAN_DATA,
    BF_TYPE_USER,

    BF_TYPE_PAST
};

#define BfB0Flink(pDesc)	(pDesc->pBlock0->flink)
#define BfB0FirstByte(pDesc)	(pDesc->pBlock0->firstByte)
#define BfB0LastByte(pDesc)	(pDesc->pBlock0->lastByte)
#define BfB0BlockSize(pDesc)	(pDesc->pBlock0->fileInfo.blockSize)
#define BfB0Flags(pDesc)	(pDesc->pBlock0->fileInfo.flags)
#define BfB0MaxSize(pDesc)	(pDesc->pBlock0->fileInfo.maxSize)
#define BfB0CurrentSize(pDesc)	(pDesc->pBlock0->fileInfo.currentSize)
#define BfB0CurrentFree(pDesc)	(pDesc->pBlock0->fileInfo.currentFree)
#define BfB0FileType(pDesc)	(pDesc->pBlock0->fileInfo.fileType)

#define BfB0RingHead(pDesc)	(pDesc->pBlock0->user[0])
#define BfB0RingTail(pDesc)	(pDesc->pBlock0->user[1])

typedef struct {
    BF_BLKNUM	flink;		/* forward link to next block */
    short	firstByte;	/* byte number of first user byte */
    short	lastByte;	/* byte number of last user byte */
    struct fileInfo fileInfo;
    long	user[BF_USRINF_DIM];
    char	pad[BF_MAX_BLOCK_SIZE-sizeof(struct fileInfo)-
				sizeof(long)*(2+BF_USRINF_DIM)];
} BF_BLOCK0;

/*----------------------------------------------------------------------------
* BF_BLOCK - definitions for other blocks
*
*    The first 3 fields in blocks other than block0 have slightly different
*    meanings than for block0:
*
*	o  flink points to the next block in a chain of user blocks
*
*	o  firstByte is the byte number of the beginning of the first
*	   complete user 'message' in the block.  User information which
*	   has overflowed from the previous block into the current block
*	   occupies the byte positions from BF_BLOCK_DATA up to firstByte.
*
*	   If the user information is not organized as messages, then
*	   firstByte will always be zero.
*
*	   If a block has only overflow information from the previous block,
*	   but no complete messages which begin in the block, then firstByte
*	   will be BF_BLOCK_DATA-1.
*
*	o  lastByte is the byte number of the last byte position occupied by
*	   user information.  If lastByte is < BF_BLOCK_DATA, then the
*	   block is empty.
*
*	   lastByte is subject to some indirect handling.  The file descriptor
*	   provides a pointer for storing bytes into a block; using this
*	   pointer leaves lastByte inaccurate.  For the most part, the
*	   application software is responsible for getting lastByte back
*	   in sync; bfClose() handles this job if it writes a block.
*
*---------------------------------------------------------------------------*/
typedef struct {
    BF_BLKNUM	flink;		/* forward link to next block */
    short	firstByte;	/* byte number of first user byte */
    short	lastByte;	/* byte number of last user byte */
} BF_BLOCK;
#define BF_BLOCK_DATA 8		/* first byte for user information */

/*----------------------------------------------------------------------------
* BF_DESC - 'file descriptor' for block files
*
*	Much of the information in the file descriptor is included for the
*	convenience of 'applications' which use the basic block file
*	routines.  Among the basic block file routines, only bfOpen() and
*	bfClose() interact with those fields.
*
*	bfOpen() initializes readCount and writeCount to -10 to indicate
*	that those fields haven't been initialized.  The other read and
*	write fields are initialized to the values indicated in the comment
*	below, as 'or ...'.
*
*	bfClose() will write a presumably altered block if writeBlock is
*	non-zero.  bfClose() will also free a presumably malloc'ed buffer
*	for each of pReadBuf and pWriteBuf if either or both are non-NULL.
*
*---------------------------------------------------------------------------*/
 
#define BfWriteable(pDesc)	((pDesc->flags) & BF_DESC_WRITE)

typedef struct {
    long	fd;		/* file descriptor */
    long	flags;		/* descriptor flags, from BF_DESC_XXX */
#define BF_DESC_WRITE 1
    short	readCount;	/* number of char left to read in block */
    short	writeCount;	/* # of char pos left to write in block */
    char       *pRead;		/* pointer to last char read, or NULL */
    BF_BLKNUM	readBlock;	/* block number for reading, or 0 */
    BF_BLOCK   *pReadBuf;	/* pointer to read buffer, or NULL */
    char       *pWrite;		/* pointer to last char written, or NULL */
    BF_BLKNUM	writeBlock;	/* block number for writing, or 0 */
    BF_BLOCK   *pWriteBuf;	/* pointer to write buffer, or NULL */
    BF_BLOCK0  *pBlock0;	/* pointer to block0 */
    char	name[GEN_FNAME_DIM];	/* file name */
} BF_DESC;

#define BfRingGetc(pDesc) \
	(--(pDesc)->readCount >= 0 \
	? *(++(pDesc)->pRead) : bfRingFillRbufAndGetc(pDesc))
#define BfRingPutc(c, pDesc) \
	(--(pDesc)->writeCount >= 0 \
	? *(++(pDesc)->pWrite)=c : bfRingFlushWbufAndPutc(c, pDesc))

BF_BLKNUM  bfAcquire();
      int  bfClose();
  BF_DESC *bfCreate();
      int  bfDumpChar();
      int  bfDumpHex();
      int  bfInfo();
  BF_DESC *bfOpen();
      int  bfRead();
      int  bfRelease();
      int  bfWrite();

     char *bfRingFgets();
      int  bfRingFputs();
      int  bfRingGetc();
      int  bfRingPutc();
#endif
