/*	@(#)arLibCONN.c	1.2 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03	10-21-91	rac	split into smaller files; handle REQ_STATUS:
 *				and SAMP_STATUS: commands
 * .04	12-05-91	rac	don't use LWP
 *
 * make options
 *	-DNDEBUG	don't compile assert checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arLibCONN - channel connection routines for AR code
*
* DESCRIPTION
*
*-***************************************************************************/

#include <stdio.h>
#include <sys/file.h>	/* O_RDWR definition */

#include <tsDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>

/*+/subr**********************************************************************
* NAME	arConnBreak - break the request's Channel Access connections
*
* DESCRIPTION
*	Issues ca_clear_channel requests for all the channels in the
*	request, including those in the elements and in the methods.
*
*	In addition to issuing the ca_clear_channel's, the following actions
*	are taken:
*	o  request status is set to AR_REQ_CONN_NONE
*	o  item event pointer and channel pointer set to null for all items
*
* RETURNS
*	AR_S_OK, or
*	AR_CAERR if a channel access error is encountered
*
* SEE ALSO
*	arConnMake and arConnHandler
*
*-*/
int
arConnBreak(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    AR_EL	*pArEl;		/* pointer to archive element */
    int		stat;		/* status from calls */
    int		retStat=AR_S_OK;/* return status to caller */

    assert(pArReq != NULL);

/*----------------------------------------------------------------------------
*    set request status and initialize the counter
*---------------------------------------------------------------------------*/
    pArReq->status = AR_REQ_CONN_NONE;
    pArReq->connPendCount = 0;

/*----------------------------------------------------------------------------
*    queue the disconnects for all the elements
*---------------------------------------------------------------------------*/
    pArEl = pArReq->pElListHead;
    while (pArEl != NULL) {
	if (pArEl->pCh != NULL) {
	    if ((stat = ca_clear_channel(pArEl->pCh)) == ECA_NORMAL) {
		pArEl->pCh = NULL;
		pArEl->pEv = NULL;
	    }
	    else {
		(void)printf("arConnBreak: error queueing disconnect %s\n",
								pArEl->name);
		retStat = AR_S_CAERR;
	    }
	}
	pArEl = pArEl->pNext;
    }

/*----------------------------------------------------------------------------
*    queue the disconnects for any methods which have channels
*---------------------------------------------------------------------------*/
    if (pArReq->act.type == AR_METH_EVENT) {
	if (pArReq->act.ctrl.event.pCh != NULL) {
	    stat = ca_clear_channel(pArReq->act.ctrl.event.pCh);
	    if (stat == ECA_NORMAL) {
		assert(pArReq->act.ctrl.event.pCh != NULL);
		pArReq->act.ctrl.event.pCh = NULL;
		pArReq->act.ctrl.event.pEv = NULL;
	    }
	    else {
		(void)printf("arConnBreak: error queueing disconnect %s\n",
					pArReq->act.ctrl.event.name);
		retStat = AR_S_CAERR;
	    }
	}
    }
    if (pArReq->inact.type == AR_METH_EVENT) {
	if (pArReq->inact.ctrl.event.pCh != NULL) {
	    stat = ca_clear_channel(pArReq->inact.ctrl.event.pCh);
	    if (stat == ECA_NORMAL) {
		assert(pArReq->inact.ctrl.event.pCh != NULL);
		pArReq->inact.ctrl.event.pCh = NULL;
		pArReq->inact.ctrl.event.pEv = NULL;
	    }
	    else {
		(void)printf("arConnBreak: error queueing disconnect %s\n",
					pArReq->inact.ctrl.event.name);
		retStat = AR_S_CAERR;
	    }
	}
    }
    if (pArReq->samp.type == AR_METH_EVENT) {
	if (pArReq->samp.ctrl.event.pCh != NULL) {
	    stat = ca_clear_channel(pArReq->samp.ctrl.event.pCh);
	    if (stat == ECA_NORMAL) {
		assert(pArReq->samp.ctrl.event.pCh != NULL);
		pArReq->samp.ctrl.event.pCh = NULL;
		pArReq->samp.ctrl.event.pEv = NULL;
	    }
	    else {
		(void)printf("arConnBreak: error queueing disconnect %s\n",
					pArReq->samp.ctrl.event.name);
		retStat = AR_S_CAERR;
	    }
	}
    }
/*----------------------------------------------------------------------------
*    for the status channels, close them
*---------------------------------------------------------------------------*/
    if (pArReq->reqStChan.pCh != NULL) {
	if ((stat = ca_clear_channel(pArReq->reqStChan.pCh)) == ECA_NORMAL)
	    pArReq->reqStChan.pCh = NULL;
	else {
	    (void)printf("arConnBreak: error queueing disconnect %s\n",
					pArReq->reqStChan.name);
	    retStat = AR_S_CAERR;
	}
    }
    if (pArReq->sampStChan.pCh != NULL) {
	if ((stat = ca_clear_channel(pArReq->sampStChan.pCh)) == ECA_NORMAL)
	    pArReq->sampStChan.pCh = NULL;
	else {
	    (void)printf("arConnBreak: error queueing disconnect %s\n",
					pArReq->sampStChan.name);
	    retStat = AR_S_CAERR;
	}
    }

    stat = ca_flush_io();
    if (stat != ECA_NORMAL) {
	(void)printf("arConnBreak: error on ca_flush_io\n");
	retStat = AR_S_CAERR;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arConnHandler - handle connect/disconnect events
*
* DESCRIPTION
*	Handles connect and disconnect events for all AR channels, for all
*	AR tasks.
*
*	THIS ROUTINE ASSUMES THAT THE pCh HAS A STRUCTURE POINTER IN THE
*	puser FIELD, and that the structure pointed to has the first few
*	items the same as an AR_EL structure.  In particular, the first items
*	must be:
*		enum arItem itemType;
*		struct arReq *pReq;
*
*	When a connect event is received for a request, the .connPendCount
*	is decremented.  If the count reaches 0, then the request status is
*	changed to AR_REQ_CONN_CA_OK.  For archive elements, the element status
*	is set to AR_EL_CONNOK.  (For the status channels, .connPendCount
*	isn't affected.)
*
*	When a disconnect event is received for a request, the .connPendCount
*	is incremented.  The request status is set to AR_REQ_CONN_CA_BAD.  For
*	archive elements, the element status is set to AR_EL_CONNPEND.  An
*	outside agent must move the request from its current queue to the
*	connect queue.  (For the status channels, .connPendCount
*	isn't affected.)
*
* RETURNS
*	none
*
* BUGS
* o	doesn't check for multiple connect or disconnect events for a
*	channel
*
* SEE ALSO
*	arConnBreak and arConnMake
*	arConnQCheck, arMainQCheck, arReadQCheck
*
* NOTES
* 1.	This routine doesn't do a ca_clear_event for a disconnect; this
*	is done for all channels in a request when the request is moved to
*	the connect queue.
* 2.	Since, in principle, a request can have its status changed by this
*	routine at any time, other routines which change the status must do
*	so under the protection of ArLock.  In addition, status mustn't be
*	changed unless it is presently either AR_REQ_ACTIVE or AR_REQ_INACTIVE.
*	(arReqDelete and the various arReqConnXxx routines are
*	exceptions.)  This means that, for status changes, code similar
*	to the following fragment should be used:
*
*	    ArLock;
*	    if (pArReq->status == AR_REQ_ACTIVE)
*		pArReq->status = AR_REQ_INACTIVE;
*	    ArUnlock;
*
*-*/
void
arConnHandler(arg)
struct connection_handler_args arg;	/* IO structure from Channel Access */
{
    AR_REQ	*pArReq;		/* pointer to request */
    void	*pArItem;		/* pointer to item */

    pArItem = (void *)ca_puser(arg.chid);
    pArReq = ((struct arEl*)pArItem)->pReq;
    assert(pArItem != NULL);
    assert(pArReq != NULL);

    if (arg.op == CA_OP_CONN_UP) {
/*----------------------------------------------------------------------------
*    handle connects
*---------------------------------------------------------------------------*/
	switch (((struct arEl*)pArItem)->itemType) {
	    case AR_ITEM_EL:
		((struct arEl *)pArItem)->status = AR_EL_CONNOK;
		break;
	    case AR_ITEM_METH_ACT:
	    case AR_ITEM_METH_INACT:
	    case AR_ITEM_METH_SAMPLE:
		assert(((struct arMe *)pArItem)->type == AR_METH_EVENT);
		break;
	    case AR_ITEM_STAT_REQ:
		if (pArReq->status == AR_REQ_ACTIVE)
		    ca_put(DBR_STRING, pArReq->reqStChan.pCh, "active");
		else if (pArReq->connPendCount > 0)
		    ca_put(DBR_STRING, pArReq->reqStChan.pCh, "conn_pend");
		else
		    ca_put(DBR_STRING, pArReq->reqStChan.pCh, "inactive");
		ca_flush_io();
		return;
	    case AR_ITEM_STAT_SAMP:
		return;
	    default:
		assertAlways(0);
	}
	if (--pArReq->connPendCount == 0) {
	    pArReq->status = AR_REQ_CONN_CA_OK;
	    if (pArReq->reqStChan.pCh != NULL &&
			ca_field_type(pArReq->reqStChan.pCh) != TYPENOTCONN) {
		ca_put(DBR_STRING, pArReq->reqStChan.pCh, "inactive");
		ca_flush_io();
	    }
	}
    }
    else {
/*----------------------------------------------------------------------------
*    handle disconnects; either arMainQCheck or arReadQCheck will
*    move the request to the connect queue
*---------------------------------------------------------------------------*/
	switch (((struct arEl*)pArItem)->itemType) {
	    case AR_ITEM_EL:
		((struct arEl *)pArItem)->status = AR_EL_CONNPEND;
		break;
	    case AR_ITEM_METH_ACT:
	    case AR_ITEM_METH_INACT:
	    case AR_ITEM_METH_SAMPLE:
		assert(((struct arMe *)pArItem)->type == AR_METH_EVENT);
		assert(((struct arMe *)pArItem)->ctrl.event.pCh != NULL);
		break;
	    case AR_ITEM_STAT_REQ:
	    case AR_ITEM_STAT_SAMP:
		return;
	    default:
		assertAlways(0);
	}
	if (pArReq->reqStChan.pCh != NULL &&
			ca_field_type(pArReq->reqStChan.pCh) != TYPENOTCONN) {
	    ca_put(DBR_STRING, pArReq->reqStChan.pCh, "conn_pend");
	    ca_flush_io();
	}
	++pArReq->connPendCount;
	pArReq->status = AR_REQ_CONN_CA_BAD;
    }

    return;
}

/*+/subr**********************************************************************
* NAME	arConnMake - make the Channel Access connections
*
* DESCRIPTION
*	Queues up the ca_search requests for all the channels in the
*	request, including those in the elements and in the methods.  (If
*	a channel already has a ca_search queued, it is skipped.)
*
*	In addition to queueing up the ca_search's, the following actions
*	are taken:
*	o  request status is set to AR_REQ_CONN_CA_BAD
*	o  the .connPendCount field in the request is incremented (from zero)
*	   once for each connection which is being attempted.  (For the status
	    channels, .connPendCount isn't affected.)
*	o  for non-connected element items, the status is set to AR_EL_CONNPEND
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CAERR if a channel access error is encountered
*	AR_S_CFOPENERR if an error occured trying to open channel file
*
* BUGS
* o	doesn't verify that graphics info is in header for 'old' channels
* o	doesn't handle user archive
*
* SEE ALSO
*	arConnBreak and arConnHandler
*
* NOTES
* 1.	If an error occurs while queueing a ca_search, then the
*	connPendCount in the request will never be able to go to zero, so
*	the request can never achieve AR_REQ_CONN_CA_OK status.  This is a
*	deliberate feature.
*
*-*/
int
arConnMake(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    char	message[80];	/* temp for log messages */
    AR_EL	*pArEl;		/* pointer to archive element */
    chid	pCh;		/* pointer to channel table */
    int		stat;		/* status from calls */
    int		retStat=AR_S_OK;/* return status to caller */
    void	arConnHandler();/* connection handler */

    assert(pArReq != NULL);

/*----------------------------------------------------------------------------
*    set request status and initialize the counter
*---------------------------------------------------------------------------*/
    pArReq->status = AR_REQ_CONN_CA_BAD;
    pArReq->connPendCount = 0;

/*----------------------------------------------------------------------------
*    queue the searches for all the elements
*---------------------------------------------------------------------------*/
    pArEl = pArReq->pElListHead;
    while (pArEl != NULL) {
	if (pArEl->pCh == NULL) {
/*----------------------------------------------------------------------------
*	no ca_search queued up yet for this element
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	    pArEl->status = AR_EL_CONNPEND;
	    stat = ca_build_and_connect(pArEl->name, TYPENOTCONN, 0, &pCh,
					(void *)NULL, arConnHandler, pArEl);
	    if (stat == ECA_NORMAL)
		pArEl->pCh = pCh;
	    else {
		pArEl->status = AR_EL_NOTCONN;
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
								pArEl->name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
	else {
/*----------------------------------------------------------------------------
*	there is a ca_search queued up; if it isn't complete, bump the count
*---------------------------------------------------------------------------*/
	    if (pArEl->status == AR_EL_CONNPEND ||
	        pArEl->status == AR_EL_NOTCONN)
		pArReq->connPendCount++;
	}
	pArEl = pArEl->pNext;
    }

/*----------------------------------------------------------------------------
*    queue the searches for any methods which have channels.  There is no
*    connection-related status for methods.  Instead, the field type in
*    a method's chid is used to determine its connection status.
*---------------------------------------------------------------------------*/
    if (pArReq->act.type == AR_METH_EVENT) {
	if (pArReq->act.ctrl.event.pCh == NULL) {
/*----------------------------------------------------------------------------
*	no ca_search queued up yet for activate method
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	    stat = ca_build_and_connect(pArReq->act.ctrl.event.name,
				    TYPENOTCONN, 0, &pCh, (void *)NULL,
				    arConnHandler, &pArReq->act);
	    if (stat == ECA_NORMAL)
		pArReq->act.ctrl.event.pCh = pCh;
	    else {
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
						pArReq->act.ctrl.event.name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
	else if (ca_field_type(pArReq->act.ctrl.event.pCh) == TYPENOTCONN) {
/*----------------------------------------------------------------------------
*	there is a ca_search queued up; if it isn't complete, bump the count
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	}
    }
    if (pArReq->inact.type == AR_METH_EVENT) {
	if (pArReq->inact.ctrl.event.pCh == NULL) {
/*----------------------------------------------------------------------------
*	no ca_search queued up yet for inactivate method
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	    stat = ca_build_and_connect(pArReq->inact.ctrl.event.name,
				    TYPENOTCONN, 0, &pCh, (void *)NULL,
				    arConnHandler, &pArReq->inact);
	    if (stat == ECA_NORMAL)
		pArReq->inact.ctrl.event.pCh = pCh;
	    else {
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
						pArReq->inact.ctrl.event.name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
	else if (ca_field_type(pArReq->inact.ctrl.event.pCh) == TYPENOTCONN) {
/*----------------------------------------------------------------------------
*	there is a ca_search queued up; if it isn't complete, bump the count
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	}
    }
    if (pArReq->samp.type == AR_METH_EVENT) {
	if (pArReq->samp.ctrl.event.pCh == NULL) {
/*----------------------------------------------------------------------------
*	no ca_search queued up yet for sample method
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	    stat = ca_build_and_connect(pArReq->samp.ctrl.event.name,
				    TYPENOTCONN, 0, &pCh, (void *)NULL,
				    arConnHandler, &pArReq->samp);
	    if (stat == ECA_NORMAL)
		pArReq->samp.ctrl.event.pCh = pCh;
	    else {
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
						pArReq->samp.ctrl.event.name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
	else if (ca_field_type(pArReq->samp.ctrl.event.pCh) == TYPENOTCONN) {
/*----------------------------------------------------------------------------
*	there is a ca_search queued up; if it isn't complete, bump the count
*---------------------------------------------------------------------------*/
	    pArReq->connPendCount++;
	}
    }

/*----------------------------------------------------------------------------
*   queue the searches for status channels, if any
*---------------------------------------------------------------------------*/
    if (pArReq->reqStChan.name[0] != '\0') {
	if (pArReq->reqStChan.pCh == NULL) {
	    stat = ca_build_and_connect(pArReq->reqStChan.name,
				TYPENOTCONN, 0, &pCh, (void *)NULL,
				arConnHandler, &pArReq->reqStChan);
	    if (stat == ECA_NORMAL)
		pArReq->reqStChan.pCh = pCh;
	    else {
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
						pArReq->reqStChan.name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
    }
    if (pArReq->sampStChan.name[0] != '\0') {
	if (pArReq->sampStChan.pCh == NULL) {
	    stat = ca_build_and_connect(pArReq->sampStChan.name,
				TYPENOTCONN, 0, &pCh, (void *)NULL,
				arConnHandler, &pArReq->sampStChan);
	    if (stat == ECA_NORMAL)
		pArReq->sampStChan.pCh = pCh;
	    else {
		(void)sprintf(message,
			"arConnMake: error queueing open %s\n",
						pArReq->sampStChan.name);
		(void)arCaStatusMessage(stat, message, NULL);
		retStat = AR_S_CAERR;
	    }
	}
    }

    stat = ca_flush_io();
    if (stat != ECA_NORMAL) {
	(void)arCaStatusMessage(stat,
			"arConnMake: error on ca_flush_io\n", NULL);
	retStat = AR_S_CAERR;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arConnOpenCF - open channel files & channel descriptors for request
*
* DESCRIPTION
*       Opens the channel file(s) needed for the request, as necessary, for
*       requests with 'channel data' type of output.  For each element in
*       the request, open a channel descriptor into the channel file, as
*       necessary.  If an open has been done for a file or channel descriptor,
*       then the open is skipped.  If the element wasn't previously in the
*       file, DBR_GR_xxx information is fetched and stored into the file.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CFOPENERR if an error occurs opening a channel file
*	AR_S_CFDESCERR if an error occurs opening a channel descriptor
*
* BUGS
* o	does a synchronous get for graphics info for new channels
* o	handles only AR chanFile, but not usr chanFile
*
* SEE ALSO
*
* NOTES
* 1.	This routine must be called only for requests with `by channel'
*	output type.
*
*-*/
int
arConnOpenCF(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    char	message[80];	/* temp for log messages */
    AR_EL	*pArEl;		/* pointer to archive element */
    chid	pCh;		/* pointer to channel table */
    int		stat;		/* status from calls */
    int		retStat=AR_S_OK;/* return status to caller */
    struct event_handler_args arg;/* temp for graphics call */
    union db_access_val grBuf;	/* temp for graphics call */

    assert(pArReq != NULL);
    assert(pArReq->output.ARType == AR_OUT_CHAN_BIN);

/*----------------------------------------------------------------------------
*    link to the necessary channel file descriptors if output type is `by
*    channel'.  Channel files are opened, if necessary.
*----------------------------------------------------------------------------*/
    if (pArReq->output.pARCFDesc == NULL) {
	if (pglArCtx->pChanFileDesc == NULL) {
	    pglArCtx->pChanFileDesc = arCFOpen(AR_FILE_CHAN_PATH, O_RDWR);
	    if (pglArCtx->pChanFileDesc == NULL) {
		(void)arLogMessage(" ",
			"arConnOpenCF: couldn't open AR chanData file", NULL);
		retStat = AR_S_CFOPENERR;
	    }
	}
	pArReq->output.pARCFDesc = pglArCtx->pChanFileDesc;
    }
    if (retStat == AR_S_OK) {
	pArEl = pArReq->pElListHead;
	while (pArEl != NULL) {
/*----------------------------------------------------------------------------
*    open chanDesc for all the elements.  If this results in a new channel
*    for the file, request the graphics information for the channel to store
*    into the file.  If an error occurs in trying to get or store the graphics
*    information, the new channel will be deleted from the file.
*---------------------------------------------------------------------------*/
	    if (pArEl->status == AR_EL_CONNOK && pArEl->pARChanDesc == NULL) {
		pArEl->pARChanDesc = arCFChanOpen(pArReq->output.pARCFDesc,
							pArEl->name, O_RDWR);
		if (pArEl->pARChanDesc == NULL) {
		    (void)sprintf(message,
			    "arConnOpenCF: error opening chanDesc for %s\n",
								pArEl->name);
		    (void)arLogMessage(" ", message, NULL);
		    retStat = AR_S_CFDESCERR;
		}
		else if ((pArEl->pARChanDesc->flags & AR_CDESC_NEW) != 0) {
		    arg.dbr = &grBuf;
		    arg.count = ca_element_count(pArEl->pCh);
		    arg.chid = pArEl->pCh;
		    arg.type = dbf_type_to_DBR_GR(ca_field_type(pArEl->pCh));
		    stat = ca_get(arg.type, pArEl->pCh, &grBuf);
assert(stat == ECA_NORMAL);
stat = ca_pend_io(1.);
		    if (stat != ECA_NORMAL) {
			(void)sprintf(message,
			    "arConnOpenCF: error on get for GR info %s\n",
								pArEl->name);
			(void)arCaStatusMessage(stat, message, NULL);
			(void)arCFChanDel(pArEl->pARChanDesc);
			retStat = AR_S_CAERR;
		    }
		    else {
			stat = arCFChanWriteGR(pArEl->pARChanDesc, &arg);
			if (stat != OK) {
			    (void)sprintf(message,
				"arConnOpenCF: error storing GR info %s\n",
								pArEl->name);
			    (void)arLogMessage(" ", message, NULL);
			    (void)arCFChanDel(pArEl->pARChanDesc);
			    retStat = stat;
			}
		    }
		}
	    }
	    pArEl = pArEl->pNext;
	}
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arConnQCheck - check the connect list for needed attention
*
* DESCRIPTION
*	Check the requests on the pReqConnHead list.  The action taken
*	for a request depends on its status:
*
*	AR_REQ_CONN_NONE results in queueing the ca_search's for the
*	channels in the request.  If the queueing of the ca_search's
*	results in errors, request is deleted (with appropriate clean up).
*	(Any channels for which ca_search has already been done are
*	skipped.  This allows using a status of AR_REQ_CONN_NONE to trigger
*	connections for channels which have been added to a request.)
*
*	AR_REQ_CONN_CA_OK results in further processing for requests with
*	`by channel' sampling.  For these requests, channel descriptors to
*	the channel file are opened as needed.  If all elements have
*	valid channel descriptors, the status is set to AR_REQ_CONN_CF_OK;
*	otherwise to AR_REQ_CONN_CF_BAD.  Requests which don't have `by
*	channel' sampling are moved directly from AR_REQ_CONN_CA_OK to
*	AR_REQ_CONN_OK.
*
*	AR_REQ_CONN_OK or AR_REQ_CONN_CF_OK results in removing the request
*	from the connect list and appending it to either the main archive
*	list or the read list, depending on whether the .mainQ or .readQ
*	flag is set, respectively.
*
*	If the .disable flag is set in the request, the request is removed
*	from the list and deleted.
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	no notification is printed if connect never succeeds
*
*-*/
int
arConnQCheck()
{
    char	message[80];	/* buffer for error messages */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    AR_REQ	*pArReq;	/* request pointer */
    AR_REQ	*pArReqNext;	/* request pointer to next request */

    pArReq = pglArCtx->pReqConnHead;
    while (pArReq != NULL) {
	pArReqNext = pArReq->pNext;
	if (pArReq->disable)
	    pArReq->status = AR_REQ_DISABLE;
	if (pArReq->status == AR_REQ_CONN_NONE) {
/*----------------------------------------------------------------------------
*    handle Channel Access connections
*----------------------------------------------------------------------------*/
	    if ((stat = arConnMake(pArReq)) != AR_S_OK) {
		(void)sprintf(message,
			"errors while connecting request: %s", pArReq->name);
		arLogMessage(" ", message, stat);
		arLogMessage(" ", "request being deleted", NULL);
		pArReq->status = AR_REQ_DISABLE;
	    }
	}
	if (pArReq->status == AR_REQ_CONN_CA_OK) {
/*----------------------------------------------------------------------------
*    handle channel file connection--all Channel Access connections are now OK
*----------------------------------------------------------------------------*/
	    if (pArReq->output.ARType == AR_OUT_CHAN_BIN ||
	        		pArReq->output.usrType == AR_OUT_CHAN_BIN) {
		if ((stat = arConnOpenCF(pArReq)) != AR_S_OK) {
		    (void)sprintf(message,
				"errors opening bin file for request: %s",
				pArReq->name);
		    arLogMessage(" ", message, stat);
		    arLogMessage(" ", "request being deleted", NULL);
		    pArReq->status = AR_REQ_DISABLE;
		}
		else
		    pArReq->status = AR_REQ_CONN_OK;
	    }
	    else
		pArReq->status = AR_REQ_CONN_OK;
	}
	if (pArReq->status == AR_REQ_CONN_OK) {
/*----------------------------------------------------------------------------
*    everything is OK.  The request can be moved to the main list or back to
*    the read list.
*----------------------------------------------------------------------------*/
	    DoubleListRemove(pArReq, pglArCtx->pReqConnHead,
						    pglArCtx->pReqConnTail);
	    if (pArReq->mainQ) {
		DoubleListAppend(pArReq, pglArCtx->pReqMainHead,
						    pglArCtx->pReqMainTail);
		pArReq->status = AR_REQ_INACTIVE;
		if ((stat = arMethAdd(&pArReq->act, AR_METH_AC_ACT)) !=
								    AR_S_OK) {
		    (void)sprintf(message,
			    "error queueing for activation: %s", pArReq->name);
		    arLogMessage(" ", message, stat);
		    arLogMessage(" ", "request being deleted", NULL);
		    pArReq->status = AR_REQ_DISABLE;
		    DoubleListAppend(pArReq, pglArCtx->pReqConnHead,
						    pglArCtx->pReqConnTail);
		}
	    }
	    else {
		DoubleListAppend(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
	    }
	}
	if (pArReq->status == AR_REQ_DISABLE) {
	    DoubleListRemove(pArReq, pglArCtx->pReqConnHead,
						    pglArCtx->pReqConnTail);
	    if ((stat = arReqDelete(pArReq)) != AR_S_OK) {
		(void)sprintf(message,
			"errors while disabling request: %s", pArReq->name);
		arLogMessage(" ", message, stat);
	    }
	}
	pArReq = pArReqNext;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arConnQPrintInfo - print information for requests on conn list
*
* DESCRIPTION
*	Prints information for each archive request in the archive request
*	connection list.
*
* RETURNS
*	AR_S_OK, or
*	status code from arReqPrintInfo
*
*-*/
int
arConnQPrintInfo(abbrev)
int	abbrev;		/* I 0,1 for full,short print */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    AR_REQ	*pArReq;	/* pointer to next request */

    if (!abbrev)
	(void)arLogMessage(" ", "**** CONN LIST ****", NULL);

    pArReq = pglArCtx->pReqConnHead;
    while (pArReq != NULL) {
	stat = arReqPrintInfo(pArReq, abbrev);
	if (stat != AR_S_OK)
	    retStat = stat;
	pArReq = pArReq->pNext;
    }

    return retStat;
}
