/*	@(#)arLibEL.c	1.2 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03	10-21-91	rac	split into smaller files; handle REQ_STATUS:
 *				and SAMP_STATUS: commands
 * .04	12-05-91	rac	don't use LWP
 *
 * make options
 *	-DNDEBUG	don't compile assert checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arLibEL - routines for handling `archive elements'
*
* DESCRIPTION
*
*-***************************************************************************/

#include <stdio.h>
#include <sys/file.h>	/* O_RDWR definition */

#include <tsDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>

/*+/subr***********************************************************************
* NAME	arElCreateAndAppend - create an AR_EL and append to archive request
*
* DESCRIPTION
*	Creates an archive element, stores the channel name in it, and
*	adds it to the end of the archive request.  The archive request
*	status is changed to AR_REQ_CONN_CA_BAD.  (No ca_search is done.
*	This will be done by arConnMake as a result of the new
*	request status which indicates that the request contains a
*	disconnected channel.)
*
*	The previous end of list element (if any) is linked to the
*	new element.
*
*	If this is the first element for a request, pElListHead is
*	properly established.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_NOMEM	if no memory is available to hold new element
* BUGS
* o	a conflict can be exist between this routine (adding an element
*	to the end of the list) and any other routine which is traversing
*	the element list.  A lock in the request to be used for these types
*	of operations would solve the problem.
*
* SEE ALSO
*	arElCreateWrapup
*
*-*/
int
arElCreateAndAppend(pArReq, chanName)
AR_REQ	*pArReq;	/* IO pointer to archive request */
char	*chanName;	/* I channel name */
{
    AR_EL	*pNewEl;	/* pointer to new element */
    int		retStat=AR_S_OK;/* return status to caller */

    assert(pArReq != NULL);
    assert(chanName != NULL);
    assert(strlen(chanName) < AR_NAME_DIM);

    if ((pNewEl = (AR_EL *)malloc(sizeof(AR_EL))) == NULL) {
	retStat = AR_S_NOMEM;
    }
    else {
	pNewEl->itemType = AR_ITEM_EL;		/* initialize the element */
	pNewEl->pReq = pArReq;
	pNewEl->pNext = NULL;
	pNewEl->pCh = NULL;
	pNewEl->pEv = NULL;
	(void)strcpy(pNewEl->name, chanName);
	pNewEl->status = AR_EL_NOTCONN;
	pNewEl->pARChanDesc = NULL;
	pNewEl->pUsrChanDesc = NULL;
	pNewEl->sampleCount = 0;
	pNewEl->dbrType = TYPENOTCONN;

	DoubleListAppend(pNewEl, pArReq->pElListHead, pArReq->pElListTail);
	pArReq->connPendCount++;
	if (  pArReq->status == AR_REQ_ACTIVE ||
	      pArReq->status == AR_REQ_INACTIVE  )
	    pArReq->status = AR_REQ_CONN_CA_BAD;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arElCreateWrapup - wrapup element create processing
*
* DESCRIPTION
*	Wrapup activities for an element following connection to Channel
*	Access:
*	o  for array channels, expand the element size to accomodate a
*	   larger value buffer; in this case, the caller's element
*	   pointer will be altered to point to the expanded element.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_NOMEM, if no memory is available for expanding an array element
*
* BUGS
* o	doesn't optimize buffer size
*
*-*/
int
arElCreateWrapup(ppArEl)
AR_EL	**ppArEl;		/* IO pointer to element pointer */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    chtype	dbfType;	/* dbf type from channel table */
    int		count;		/* native count for channel */
    AR_REQ	*pArReq;	/* pointer to request containing element */
    AR_EL	*pArEl;		/* pointer to element */
    AR_EL	*pTempEl;	/* pointer to expanded element */

    assert(ppArEl != NULL);
    pArEl = *ppArEl;
    assert(pArEl != NULL);
    assert(pArEl->status == AR_EL_CONNOK);
    pArReq = pArEl->pReq;
    assert(pArReq != NULL);
    assert(pArEl->pCh != NULL);
    dbfType = ca_field_type(pArEl->pCh);
    assert(dbf_type_is_valid(dbfType));

    if ((count = ca_element_count(pArEl->pCh)) > 1) {
	pTempEl = (AR_EL *)malloc(sizeof(AR_EL) + count*AR_STRVAL_DIM);
	if (pTempEl == NULL)
	    retStat = AR_S_NOMEM;
	else {
	    *pTempEl = *pArEl;
	    assert(pTempEl->pReq == pArReq);
	    if (pTempEl->pPrev != NULL)
		pTempEl->pPrev->pNext = pTempEl;
	    if (pTempEl->pNext != NULL)
		pTempEl->pNext->pPrev = pTempEl;
	    if (pArReq->pElListHead == pArEl)
	        pArReq->pElListHead = pTempEl;
	    if (pArReq->pElListTail == pArEl)
	        pArReq->pElListTail = pTempEl;
	    free((char *)pArEl);
	    *ppArEl = pTempEl;
	}
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arElEventHandler - handle monitor event for an element
*
* DESCRIPTION
*	Handles a channel access monitor 'event'.  The new value for the
*	element is stored, along with the status, severity, and time stamp.
*	(Only DBR_TIME_xxx monitors are handled by this routine.)  The
*	.dbrType field is set indicating the type of value in the buffer.
*
*	If the status is AR_EL_GETPEND, the value is stored in the element,
*	its status is changed to AR_EL_GETDONE, and the `elPendCount' for the
*	request is decremented.
*
*	If the status of the archive request isn't AR_REQ_ACTIVE, or if the
*	element status is other than indicated above, the value is discarded.
*
* RETURNS
*	void
*
* BUGS
* o	handles only STRING, SHORT, and FLOAT
* o	doesn't do any validity checking on element
*
*-*/
void
arElEventHandler(arg)
struct event_handler_args arg;	/* I handler arg */
{
    AR_EL	*pArEl;		/* I element pointer */
    chid	pCh;		/* I channel pointer */
    chtype	bufType;	/* I buffer type */
    int		bufCount;	/* I count */
    void	*pBuf;		/* IO pointer to value buffer */
    char	message[80];	/* temp buffer for error messages */
    int		i;		/* temp for loop index */
    pArEl = arg.usr;
    assert(pArEl != NULL);
    pCh = arg.chid;
    bufType = arg.type;
    assert(dbr_type_is_TIME(bufType));
    bufCount = arg.count;
    pBuf = arg.dbr;

    if (pArEl->status != AR_EL_GETPEND ||
    					pArEl->pReq->status != AR_REQ_ACTIVE) {
#if 0
	(void)sprintf(message, "error in arElEventHandler for %s", pArEl->name);
	(void)arLogMessage(" ", message, NULL);
	(void)sprintf(message, "in request: %s", pArEl->pReq->name);
	(void)arLogMessage(" ", message, NULL);
	if (pArEl->pReq->status != AR_REQ_ACTIVE) {
	    (void)sprintf(message, "bad request status: %s",
					ArReqStatusToText(pArEl->pReq->status));
	    (void)arLogMessage(" ", message, NULL);
	}
	else {
	    (void)sprintf(message, "bad element status: %s",
					ArElStatusToText(pArEl->status));
	    (void)arLogMessage(" ", message, NULL);
	}
	(void)arLogMessage(" ", "discarding value", NULL);
#endif
	return;
    }

/*----------------------------------------------------------------------------
* status, severity, and time stamp.  (If stamp from Channel Access is zero,
*    assume it's missing and fill it in with local system time.)
*---------------------------------------------------------------------------*/
    pArEl->dbrVal.tstrval.status = ((struct dbr_time_string *)pBuf)->status;
    pArEl->dbrVal.tstrval.severity = ((struct dbr_time_string *)pBuf)->severity;
    if (((struct dbr_time_string *)pBuf)->stamp.secPastEpoch == 0)
	(void)tsLocalTime(&pArEl->dbrVal.tstrval.stamp);
    else
	pArEl->dbrVal.tstrval.stamp = ((struct dbr_time_string *)pBuf)->stamp;

    if (bufType == DBR_TIME_STRING) {
	char	*pStr, *pStrDest;
	pStr = ((struct dbr_time_string *)pBuf)->value;
	pStrDest = pArEl->dbrVal.tstrval.value;
	for (i=0; i<bufCount; i++) {
	    strcpy(pStrDest, pStr);
	    pStr += db_strval_dim;
	    pStrDest += db_strval_dim;
	}
    }
    else if (bufType == DBR_TIME_SHORT) {
	short	*pShort, *pShortDest;
	pShort = &((struct dbr_time_short *)pBuf)->value;
	pShortDest = &pArEl->dbrVal.tshrtval.value;
	for (i=0; i<bufCount; i++)
	    *pShortDest++ = *pShort++;
    }
    else if (bufType == DBR_TIME_FLOAT) {
	float	*pFloat, *pFloatDest;
	pFloat = &((struct dbr_time_float *)pBuf)->value;
	pFloatDest = &pArEl->dbrVal.tfltval.value;
	for (i=0; i<bufCount; i++)
	    *pFloatDest++ = *pFloat++;
    }
    else {
	printf("arElEventHandler: illegal buffer type: %s\n",
						dbr_type_to_text(bufType));
    }
    if (pArEl->status == AR_EL_GETPEND) {
	pArEl->status = AR_EL_GETDONE;
	pArEl->pReq->elPendCount--;
    }
    else if (pArEl->status == AR_EL_MONPEND) {
	pArEl->status = AR_EL_MONITOR;
    }

    pArEl->dbrType = bufType;

    return;
}

/*+/subr***********************************************************************
* NAME	arElFind - search for element in archive request
*
* DESCRIPTION
*	The elements on the specified archive request are searched,
*	looking for one whose channel name matches the specified name.
*
* RETURNS
*	NULL, or
*	a pointer (AR_EL *) to the desired element.
*
*-*/
AR_EL *
arElFind(pArReq, name)
AR_REQ	*pArReq;	/* I pointer to archive request */
char	*name;		/* I channel name, null terminated string */
{
    AR_EL	*pEl;	/* element pointer */

    pEl = pArReq->pElListHead;
    while (pEl != NULL) {
	if (strcmp(name, pEl->name) == 0)
	    break;		/* found element in list! */
	else
	    pEl = pEl->pNext;
    }
    return pEl;
}

/*+/subr***********************************************************************
* NAME	arElGetValue - do a ca_array_get_callback
*
* DESCRIPTION
*	Queues up a request to get the value for the channel associated
*	with an archive element.  The `elPendCount' for the request is
*	incremented and element status is changed to AR_EL_GETPEND.
*
*	New values won't be posted to the element until ca_pend_event
*	has been called.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CAERR   	an error was encountered when calling ca_get
*			(an error message has been printed)
*	AR_S_GETILL	attempt to ca_get for element which isn't OPEN
*			or whose status is otherwise invalid
*
* BUGS
* o	this routine always uses DBR_TIME_STRING; there should probably be
*	two of these routines, arElGetStringValue being the other one.
* o	if an error occurs, the 'elPendCount' stays non-zero
* o	there is no protection against several calls for same element
*
*-*/
int
arElGetValue(pArEl)
AR_EL	*pArEl;		/* I pointer to archive element */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArEl != NULL);
    assert(pArEl->pReq != NULL);

    if (pArEl->status == AR_EL_CONNOK || pArEl->status == AR_EL_GETDONE) {
	pArEl->status = AR_EL_GETPEND;
	pArEl->pReq->elPendCount++;
	stat = ca_array_get_callback(DBR_TIME_STRING,
			ca_element_count(pArEl->pCh), pArEl->pCh,
			arElEventHandler, pArEl);
	if (stat != ECA_NORMAL) {
	    pArEl->status = AR_EL_GETFAIL;
	    (void)arCaStatusMessage(stat,
				"error queueing get request", pArEl->pCh);
	    retStat = AR_S_CAERR;
	}
    }
    else
	retStat = AR_S_GETILL;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arElGetNativeValue - do a ca_array_get_callback using native type
*
* DESCRIPTION
*	Queues up a request to get the value for the channel associated
*	with an archive element.  The `elPendCount' for the request is
*	incremented and element status is changed to AR_EL_GETPEND.
*
*	New values won't be posted to the element until ca_pend_event
*	has been called.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CAERR   	an error was encountered when calling ca_get
*			(an error message has been printed)
*	AR_S_GETILL	attempt to ca_get for element which isn't OPEN
*			or whose status is otherwise invalid
*
* BUGS
* o	if an error occurs, the 'elPendCount' stays non-zero
* o	there is no protection against several calls for same element
*
*-*/
int
arElGetNativeValue(pArEl)
AR_EL	*pArEl;		/* I pointer to archive element */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArEl != NULL);
    assert(pArEl->pReq != NULL);

    if (pArEl->status == AR_EL_CONNOK || pArEl->status == AR_EL_GETDONE) {
	pArEl->status = AR_EL_GETPEND;
	pArEl->pReq->elPendCount++;
	stat = ca_array_get_callback(
			dbf_type_to_DBR_TIME(ca_field_type(pArEl->pCh)),
			ca_element_count(pArEl->pCh), pArEl->pCh,
			arElEventHandler, pArEl);
	if (stat != ECA_NORMAL) {
	    pArEl->status = AR_EL_GETFAIL;
	    (void)arCaStatusMessage(stat,
				"error queueing get request", pArEl->pCh);
	    retStat = AR_S_CAERR;
	}
    }
    else
	retStat = AR_S_GETILL;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arElLogStatus - print on stderr message for status code
*
* DESCRIPTION
*	The channel name for an element, and the text for an AR_S_xxx
*	status code are printed on stderr, in the format used by
*	arLogMessage.
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	need to make sure status code is a valid AR status code
*
*-*/
int
arElLogStatus(pArEl, status)
AR_EL	*pArEl;		/* I pointer to archive element */
int	status;		/* I archive status code */
{
    char	message[80];	/* temp for status message */

    if (pArEl->pCh != NULL)
	(void)sprintf(message, "%s: %s", ca_name(pArEl->pCh),
	    					ArStatusToText(status));
    else
	(void)sprintf(message, "%s: %s, no pCh", pArEl->name,
	    					ArStatusToText(status));
    arLogMessage(" ", message, NULL);

    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arElPrintInfo
*
* DESCRIPTION
*	Status information for an archive element is printed.
*
* RETURNS
*	AR_S_OK
*
*-*/
int
arElPrintInfo(pArEl)
AR_EL	*pArEl;		/* I pointer to archive element, */
{
    char	message[80];	/* temp for message */

    assert(pArEl != NULL);

    (void)sprintf(message, "%s %s", pArEl->name,
	    			dbf_type_to_text(ca_field_type(pArEl->pCh)));
    (void)arLogMessage(" ", message, NULL);

    if (pArEl->pCh != NULL) {
	(void)sprintf(message, "	%s; %s; # el = %d",
				ArElStatusToText(pArEl->status),
	    			dbr_type_to_text(pArEl->dbrType),
				ca_element_count(pArEl->pCh));
	(void)arLogMessage(" ", message, NULL);
	(void)arElPrintValue(pArEl);
    }
    else {
	(void)sprintf(message, "	%s; no pCh",
					ArElStatusToText(pArEl->status));
	(void)arLogMessage(" ", message, NULL);
    }

    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arElPrintValue - print the value for an element
*
* DESCRIPTION
*	Prints the value(s) for an element, along with the element's
*	channel name and time stamp.
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	handles only STRING, SHORT, and FLOAT
* o	doesn't print channel status and severity
*
*-*/
int
arElPrintValue(pArEl)
AR_EL	*pArEl;		/* I pointer to archive element */
{
    int		retStat=AR_S_OK;/* return status to caller */
    char	message[80];	/* temp for messages */
    char	stampText[32];	/* text for time stamp */
    char	*pChar;		/* printf character pointer */
    short	*pShort;	/* printf short pointer */
    float	*pFloat;	/* printf float pointer */
    int		i;		/* counter for loops */
    int		count;		/* number of elements left to print */

    assert(pArEl != NULL);

    if (pArEl->pCh == NULL) {
	(void)arLogMessage(" ", "	no pCh", NULL);
    }
    else if (pArEl->status != AR_EL_GETDONE && pArEl->status != AR_EL_MONITOR){
	(void)arLogMessage(" ", "	no value in buffer", NULL);
    }
    else {
	(void)tsStampToText(&pArEl->dbrVal.tstrval.stamp, TS_TEXT_MMDDYY,
								stampText);
	(void)sprintf(message, "	%s", stampText);
	(void)arLogMessage(" ", message, NULL);
	if (pArEl->dbrType == DBR_TIME_STRING) {
	    pChar = pArEl->dbrVal.tstrval.value;
	    for (i=0; i<ca_element_count(pArEl->pCh); i++) {
		(void)sprintf(message, "	%d: %s", i, pChar);
		(void)arLogMessage(" ", message, NULL);
		pChar += db_strval_dim;
	    }
	}
	else if (pArEl->dbrType == DBR_TIME_SHORT) {
	    pShort = &pArEl->dbrVal.tshrtval.value;
	    count = ca_element_count(pArEl->pCh);
	    for (i=0; i<count;) {
		if (i+6 < count) {
		    (void)sprintf(message, "  %5d:%6d%6d%6d%6d%6d%6d%6d", i,
			*pShort++, *pShort++, *pShort++, *pShort++,
			*pShort++, *pShort++, *pShort++);
		    i += 7;
		}
		else
		    (void)sprintf(message, "  %5d:%6d", i++, *pShort++);
		(void)arLogMessage(" ", message, NULL);
	    }
	}
	else if (pArEl->dbrType == DBR_TIME_FLOAT) {
	    pFloat = &pArEl->dbrVal.tfltval.value;
	    for (i=0; i<ca_element_count(pArEl->pCh); i++) {
		(void)sprintf(message, "	%d: %d", i, *pFloat++);
		(void)arLogMessage(" ", message, NULL);
	    }
	}
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arElPutStringValue - write string value to element's channel
*
* DESCRIPTION
*	A new value is queued for writing to the channel for the element.
*
*	A flush must be done before the value will actually be sent.
*	The value in the element is not changed by this routine.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CAERR	error encountered on ca_put
*	AR_S_PUTILL	element status must be OPEN or MONITOR
*
* BUGS
* o	no direct notification is available when errors are encountered
*	in converting the string value to the native field type.
* o	doesn't handle array values
*
*-*/
int
arElPutStringValue(pArEl, pValue)
AR_EL	*pArEl;		/* I pointer to element */
char	*pValue;	/* I pointer to string value */
{
    char	message[80];	/* temp for status message */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArEl != NULL);
    assert(pArEl->pCh != NULL);
    assert(pValue != NULL);

    if (pArEl->status == AR_EL_CONNOK || pArEl->status == AR_EL_GETDONE) {
	stat = ca_put(DBR_STRING, pArEl->pCh, pValue);
	if (stat != ECA_NORMAL) {
	    (void)arCaStatusMessage(stat,
			"error queueing put request", pArEl->pCh);
	    retStat = AR_S_CAERR;
	}
    }
    else {
	(void)sprintf(message,
	    "invalid status to do put: %s; must be CONNOK or GETDONE",
	    ArElStatusToText(pArEl->status));
	(void)arLogMessage(" ", message, NULL);
	retStat = AR_S_PUTILL;
    }
    return retStat;
}
