/*	@(#)arLibMETH.c	1.2 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03	10-21-91	rac	split into smaller files; handle REQ_STATUS:
 *				and SAMP_STATUS: commands
 * .04	12-05-91	rac	don't use LWP
 *
 * make options
 *	-DNDEBUG	don't compile assert checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arLibMETH - routines for handling archive `methods'
*
* DESCRIPTION
*
*-***************************************************************************/

#include <stdio.h>
#include <sys/file.h>	/* O_RDWR definition */

#include <tsDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>

/*+/subr***********************************************************************
* NAME	arMethActionAct - activation action routine for requests
*
* DESCRIPTION
*	Handles activating a request.  If the request status isn't INACTIVE,
*	nothing is done.
*
* RETURNS
*	AR_S_OK, or
*	other status codes
*
* BUGS
* o	no history is written
* o	doesn't handle user `by channel' files
* o	special handling for inactivation method isn't done
* o	does no error/status handling
*
* EXAMPLE
*
*-*/
int
arMethActionAct(pArMeth)
AR_METH	*pArMeth;	/* IO pointer to method */
{
    char	message[80];	/* temp for messages */
    AR_EL	*pArEl;
    AR_REQ	*pArReq;

    assert(pArMeth != NULL);
    pArReq = pArMeth->pReq;
    assert(pArReq != NULL);

/*----------------------------------------------------------------------------
*	actions needed for a request activation event
*	o  for `by channel' files, write a monitor sample begin item for each
*	    channel in the request
*	o  change the request status to ACTIVE
*	o  set `elPendCount' to zero
*	o  set `sampleCount' to zero
*	o  if activation method is not 'periodic', remove it from queue
*	o  add the inactivation method to a method queue
*	o  add the sample method to a method queue
*	o  write a message to the request history file
*---------------------------------------------------------------------------*/
    if (pArReq->status != AR_REQ_ACTIVE) {
	(void)sprintf(message, "%s activated", pArReq->name);
	if (pArReq->output.ARType == AR_OUT_CHAN_BIN) {
	    pArEl = pArReq->pElListHead;
	    while (pArEl != NULL) {
		(void)arCFChanWriteByte(pArEl->pARChanDesc,AR_DAT_IC_MON_BEGIN);
		pArEl = pArEl->pNext;
	    }
	}
	(void)arLogMessage(" ", message, NULL);
	if (pArReq->reqStChan.pCh != NULL &&
			ca_field_type(pArReq->reqStChan.pCh) != TYPENOTCONN) {
	    ca_put(DBR_STRING, pArReq->reqStChan.pCh, "active");
	    ca_flush_io();
	}
	pArReq->status = AR_REQ_ACTIVE;
	pArReq->elPendCount = 0;
	pArReq->sampleCount = 0;
	pArReq->output.status = AR_OUTS_IDLE;
	if (pArMeth->type != AR_METH_PERIODIC)
	    arMethDel(pArMeth);
	arMethAdd(&pArReq->inact, AR_METH_AC_INACT);
	arMethAdd(&pArReq->samp, AR_METH_AC_SAMPLE);
    }

    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arMethActionInact - inactivation action routine for requests
*
* DESCRIPTION
*
* RETURNS
*	AR_S_OK, or
*	other status codes
*
* BUGS
* o	no history is written
* o	doesn't handle user `by channel' files
* o	sample data aren't flushed
* o	does no error/status handling
*
* EXAMPLE
*
*-*/
int
arMethActionInact(pArMeth)
AR_METH	*pArMeth;	/* IO pointer to method */
{
    char	message[80];	/* temp for messages */
    AR_EL	*pArEl;
    AR_REQ	*pArReq;

    assert(pArMeth != NULL);
    pArReq = pArMeth->pReq;
    assert(pArReq != NULL);

/*----------------------------------------------------------------------------
*	actions needed for a request inactivation event
*	o  change the request status to INACTIVE
*	o  for `by channel' files, write a monitor sample inact item for each
*	    channel in the request
*	o  remove inactivation method from its method queue
*	o  if activation method is not 'periodic', add it to a method queue
*	o  flush sample data for the method's request
*	o  remove sample method from its method queue
*	o  write a message to the request history file
*---------------------------------------------------------------------------*/
    (void)sprintf(message, "%s inactivated", pArReq->name);
    if (pArReq->status == AR_REQ_ACTIVE) {
	if (pArReq->output.ARType == AR_OUT_CHAN_BIN) {
	    pArEl = pArReq->pElListHead;
	    while (pArEl != NULL) {
		(void)arCFChanWriteByte(pArEl->pARChanDesc,AR_DAT_IC_MON_INACT);
		pArEl = pArEl->pNext;
	    }
	}
    }
    (void)arLogMessage(" ", message, NULL);
    if (pArReq->reqStChan.pCh != NULL &&
			ca_field_type(pArReq->reqStChan.pCh) != TYPENOTCONN) {
	ca_put(DBR_STRING, pArReq->reqStChan.pCh, "inactive");
	ca_flush_io();
    }
    pArReq->status = AR_REQ_INACTIVE;
    arMethDel(pArMeth);
    arMethDel(&pArReq->samp);
    if (pArReq->act.type != AR_METH_PERIODIC)
	arMethAdd(&pArReq->act, AR_METH_AC_ACT);
    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arMethActionSample - sampling action routine for requests
*
* DESCRIPTION
*
* RETURNS
*	AR_S_OK, or
*	other status codes
*
* BUGS
* o	does no error/status handling
*
*-*/
int
arMethActionSample(pArMeth)
AR_METH	*pArMeth;	/* IO pointer to method */
{
    AR_REQ	*pArReq;	/* pointer to request containing method */
    int		stat;		/* status return from calls */

    assert(pArMeth != NULL);
    pArReq = pArMeth->pReq;
    assert(pArReq != NULL);

    if (pArReq->output.status != AR_OUTS_IDLE)
	return AR_S_NOTIDLE;

/*----------------------------------------------------------------------------
*	actions needed for a request sampling event
*	o  take a sample
*	o  set up for next sample
*	o  update sample count in request (???)
*	o  see if request should be inactivated (???)
*---------------------------------------------------------------------------*/
    if (pArReq->output.ARType == AR_OUT_SET_TXT  ||
	pArReq->output.usrType == AR_OUT_SET_TXT)  {
	pArReq->output.status = AR_OUTS_SAMPLE;
    }
    else if (pArReq->output.ARType == AR_OUT_CHAN_BIN  ||
				pArReq->output.usrType == AR_OUT_CHAN_BIN)  {
	if ((stat = arSampleChanAddMon(pArReq)) != AR_S_OK) {
	    pArReq->status = AR_REQ_DISABLE;
	}
	else
	    pArReq->output.status = AR_OUTS_MONITOR;
    }

    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arMethAdd - add a method to a method queue
*
* DESCRIPTION
*	Adds a method to the queue appropriate for the method type.  For more
*	details, see the discussions for the various method types, below.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_CAERR, for error on Channel Access call, or
*	assertFail is called for some other errors
*
* EXAMPLE
*	AR_REQ	*pReq;		pointer to archive request
*
*	arMethAdd(&pReq->act, AR_METH_AC_ACT);
*
* NOTES
* 1.	For event methods, ca_add_event is called, but ca_flush_io
*	isn't called--flushing the Channel Access buffer must be done
*	elsewhere.
*
*-*/
int
arMethAdd(pMeth, actionType)
AR_METH	*pMeth;		/* IO method pointer */
int	actionType;	/* I action type, one of the AR_METH_AC_xxx */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    void	arMethEventHandlerCa();	/* handler for channel changes */
    char	message[80];	/* error message buffer */
    chtype	type;		/* buffer type, for EVENT method */

    assert(pMeth != NULL);
    assert(pMeth->actionType == AR_METH_AC_UNDEF);
    assert(pMeth->pReq != NULL);
    assert(pMeth->pReq->status != AR_REQ_UNDEF);

    if (pMeth->type == AR_METH_PERIODIC) {
/*----------------------------------------------------------------------------
* PERIODIC
*	append method to end of 'periodic' list
*	store type of action into method
*       for sample methods, set timer to present time, for initial action
*               immediately
*	set .state to ARMED
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethClkHead,
						pglArCtx->pMethClkTail);
	pMeth->actionType = actionType;
	if (actionType == AR_METH_AC_SAMPLE)
	    (void)tsLocalTime(&pMeth->ctrl.per.nextTime);
	pMeth->state = AR_METH_ST_ARMED;
    }
    else if (pMeth->type == AR_METH_TIMED) {
/*----------------------------------------------------------------------------
* TIMED
*	append method to end of 'periodic' list
*	store type of action into method
*	set timer to present time + period, for delayed action
*	set .state to ARMED
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethClkHead,
						pglArCtx->pMethClkTail);
	pMeth->actionType = actionType;
	(void)tsLocalTime(&pMeth->ctrl.per.nextTime);
	TsAddStamps(&pMeth->ctrl.per.nextTime, &pMeth->ctrl.per.nextTime,
					    &pMeth->ctrl.per.period);
	pMeth->state = AR_METH_ST_ARMED;
    }
    else if (pMeth->type == AR_METH_EVENT) {
/*----------------------------------------------------------------------------
* EVENT
*	append method to end of 'event' list
*	store type of action into method
*	set .state to AR_METH_ST_NEW so that arMethEventHandlerCa will be
*		able to recognize the first value coming in from the channel
*		access monitor.
*	add channel access monitor for channel; if channel is DBF_ENUM, then
*		monitor buffer will be DBR_STRING
*	initialize counter to zero
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethEvHead, pglArCtx->pMethEvTail);
	pMeth->actionType = actionType;
	type = dbf_type_to_DBR(ca_field_type(pMeth->ctrl.event.pCh));
	if (type == DBR_ENUM)
	    type = DBR_STRING;
	pMeth->state = AR_METH_ST_NEW;
	pMeth->ctrl.event.counter = pMeth->ctrl.event.count;
	stat = ca_add_event(type, pMeth->ctrl.event.pCh,
		    arMethEventHandlerCa, pMeth, &pMeth->ctrl.event.pEv);
	if (stat != ECA_NORMAL) {
	    (void)sprintf(message,
		    "error adding monitor for event; request: %s",
		    pMeth->pReq->name);
	    (void)arCaStatusMessage(stat, message, pMeth->ctrl.event.pCh);
	}
    }
    else if (pMeth->type == AR_METH_MONITOR) {
/*----------------------------------------------------------------------------
* MONITOR
*	(This assumes that all the elements have had an arCFChanOpen.)
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethMiscHead,
						    pglArCtx->pMethMiscTail);
	pMeth->actionType = actionType;
	pMeth->state = AR_METH_ST_ARMED;
    }
    else if (pMeth->type == AR_METH_AFTER) {
/*----------------------------------------------------------------------------
* AFTER
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethMiscHead,
						    pglArCtx->pMethMiscTail);
	pMeth->actionType = actionType;
	pMeth->state = AR_METH_ST_UNARMED;
    }
    else if (pMeth->type == AR_METH_DEMAND) {
/*----------------------------------------------------------------------------
* DEMAND
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethMiscHead,
						    pglArCtx->pMethMiscTail);
	pMeth->actionType = actionType;
	pMeth->state = AR_METH_ST_UNARMED;
    }
    else if (pMeth->type == AR_METH_ALWAYS) {
/*----------------------------------------------------------------------------
* ALWAYS
*---------------------------------------------------------------------------*/
	DoubleListAppend(pMeth, pglArCtx->pMethMiscHead,
						    pglArCtx->pMethMiscTail);
	pMeth->actionType = actionType;
	pMeth->state = AR_METH_ST_ARMED;
    }
    else if (pMeth->type == AR_METH_NEVER) {
/*----------------------------------------------------------------------------
* NEVER
*---------------------------------------------------------------------------*/
	pMeth->actionType = actionType;
	pMeth->state = AR_METH_ST_UNARMED;
    }

    pMeth->demand = 0;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arMethBuild - set up an archive request method
*
* DESCRIPTION
*	Parse and check a method specification line from an archive request
*	file.  Set up the method structure according to the specification.
*	(It is expected that the caller has already looked at the specification
*	line and determined 1) that the line specifies a method and 2) which
*	method structure (act, inact, or sample) to use.)
*
*	For EVENT methods, a Channel Access open is queued for the channel.
*	The validation which must be done based on the characteristics of
*	the channel are NOT done in this routine, and MUST BE DONE ELSEWHERE!
*
*	At exit, the following fields in the method are set as indicated:
*	  .type       one of the AR_METH_xxx, such as AR_METH_EVENT
*	  .state      AR_METH_ST_UNDEF
*	  .actionType AR_METH_AC_UNDEF
*
*	Other fields in the method are set as appropriate for the type of
*	method.
*
*	The method specification line has the following form:
*	  actionType methodSpec
*
*	actionType can be one of ACT:  INACT:  or  SAMPLE:
*	methodSpec can have one of several forms, depending on method type
*	  after n samples
*	  always
*	  demand
*	  event chanName cond value [every n times]
*	  monitor
*	  never
*	  periodic hh:mm:ss
*	  timed hh:mm:ss
*
* RETURNS
*	AR_S_OK, or
*	AR_S_SYNERR if a syntax error is encountered in the record (an error
*		message--not including the record-- has been printed)
*
* SEE ALSO
*	arMethBuildCheck, arReadReq, and arReadReqCheck
*
* NOTES
* 1.	The comparison value for EVENT methods isn't validated, but is
*	stored into the method as a string.  Once the channel for the method
*	is opened, the comparison value can be properly scanned and validated.
*	(This is done by arMethBuildCheck.)
* 2.	Trailing 'noise' words sometimes aren't validated.
* 3.	This routine does limited validation of rules about what methods
*	can have what types, and how the different methods interact.
*
* EXAMPLE
*	stat = arMethBuild(&pArReq->act, "ACT: demand");
*
*-*/
int
arMethBuild(pArMeth, string)
AR_METH	*pArMeth;	/* IO pointer to method structure in request */
char	*string;	/* I string containing method information (from
				request file) */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    char	typeC[13];	/* method type from record */
    int		type;		/* method type */
    int		hh;		/* hours */
    int		mm;		/* minutes */
    int		ss;		/* seconds */
    char	chanC[AR_NAME_DIM];/* channel name */
    chid	pCh;		/* channel pointer */
    char	evTypeC[10];	/* event type name */
    int		evType;		/* event type code */
    char	evValC[10];	/* temp to check event if comp value present */
    char	everyC[10];	/* 'every' keyword */
    int		every;		/* 'every' count */
    char	timesC[6];	/* 'times' keyword */
    long	afterCount;	/* 'after' count */
    char	afterC[10];	/* 'after' keyword */

    assert(sizeof(typeC) == 13);
    if (sscanf(string, "%*s %12s", typeC) != 1) {
	(void)printf("no type field\n");
	retStat = AR_S_SYNERR;
    }
    else {
	if (strcmp(typeC, glArMethKw[AR_METH_PERIODIC]) == 0  ||
	         strcmp(typeC, glArMethKw[AR_METH_TIMED]) == 0) {
/*----------------------------------------------------------------------------
* PERIODIC
*	XXX: periodic hh:mm:ss
* TIMED
*	XXX: timed hh:mm:ss
*---------------------------------------------------------------------------*/
	    if (strcmp(typeC, glArMethKw[AR_METH_PERIODIC]) == 0)
		type = AR_METH_PERIODIC;
	    else
		type = AR_METH_TIMED;
	    if (sscanf(string, "%*s %*s %d:%d:%d", &hh, &mm, &ss) !=3) {
		(void)printf("improper interval specification\n");
		retStat = AR_S_SYNERR;
	    }
	    else {
		if (  hh >= 0 && hh < 24 &&
		      mm >= 0 && mm < 60 &&
		      ss >= 0 && ss < 60    ) {
		    pArMeth->ctrl.per.period.nsec = 0;
		    pArMeth->ctrl.per.period.secPastEpoch =
						 	(hh*60 + mm)*60 + ss;
		    if (  pArMeth->ctrl.per.period.secPastEpoch < 10 ||
		          pArMeth->ctrl.per.period.secPastEpoch > 86400) {
			(void)printf("improper interval specification\n");
			retStat = AR_S_SYNERR;
		    }
		    else
			(void)tsLocalTime(&pArMeth->ctrl.per.nextTime);
		}
		else {
		    (void)printf("improper interval specification\n");
		    retStat = AR_S_SYNERR;
		}
	    }
	}
	else if (strcmp(typeC, glArMethKw[AR_METH_EVENT]) == 0) {
/*----------------------------------------------------------------------------
* EVENT
*	XXX: event chanName cond value every count times
*
*    `cond' can be one of:  change  ==  !=  >  >=  <  <=
*    `value' must not appear for `cond' == `change'
*    `every count' is optional
*    if `every count' is specified, `times' is required
*---------------------------------------------------------------------------*/
	    type = AR_METH_EVENT;
	    pArMeth->ctrl.event.pCh = NULL;
	    evType = AR_EV_UNDEF;
	    assert(AR_NAME_DIM == 36);
	    if (sscanf(string, "%*s %*s %35s", chanC) !=1) {
		(void)printf("no channelName field\n");
		retStat = AR_S_SYNERR;
	    }
	    else
		(void)strcpy(pArMeth->ctrl.event.name, chanC);
	    assert(sizeof(evTypeC) == 10);
	    if (sscanf(string, "%*s %*s %*s %9s", evTypeC) != 1) {
		(void)printf("no condition field\n");
		retStat = AR_S_SYNERR;
	    }
	    else {
/*----------------------------------------------------------------------------
*        check the condition for validity, and get the corresponding code.
*	 (Validating the condition against the channel type must be done
*	 when the open is complete.)
*---------------------------------------------------------------------------*/
		for (evType=AR_EV_PAST-1; evType>0; evType--) {
		    if (strcmp(glArEvKw[evType], evTypeC) == 0)
			break;
		}
		if (evType == 0) {
		    assert(AR_EV_UNDEF == 0);
		    (void)printf("illegal condCode specifier\n");
		    retStat = AR_S_SYNERR;
		}
	    }
	    if (evType != AR_EV_UNDEF && evType != AR_EV_CHANGE) {
/*----------------------------------------------------------------------------
*        get the comparison value (except for AR_EV_CHANGE), storing it into
*        the event structure in the method structure.  (It must be checked
*        for validity after the open is complete for the channel.)
*---------------------------------------------------------------------------*/
		assert(sizeof(pArMeth->ctrl.event.compVal.strval) == 40);
		if (sscanf(string, "%*s %*s %*s %*s %39s",
				   pArMeth->ctrl.event.compVal.strval) != 1) {
		    (void)printf("no value field\n");
		    evType = AR_EV_UNDEF;
		    retStat = AR_S_SYNERR;
		}
	    }
/*----------------------------------------------------------------------------
*        if no errors have occurred on the condition code, get and check the
*        event counter keyword and count.  If no keyword, use count of 1.
*---------------------------------------------------------------------------*/
	    everyC[0] = '\0';	/* len to 0, to make checking easier */
	    every = -1;		/* invalid value, to ease checking */
	    timesC[0] = '\0';	/* len to 0, to make checking easier */
	    switch (evType) {
		case AR_EV_UNDEF:
		    break;
		case AR_EV_CHANGE:
		    assert(sizeof(everyC) == 10);
		    if (sscanf(string, "%*s %*s %*s %*s %9s", everyC) == 1) {
			assert(sizeof(timesC) == 6);
			if (sscanf(string, "%*s %*s %*s %*s %*s %d %5s",
							&every, timesC) != 2)
			    retStat = AR_S_SYNERR;
		    }
		    break;
		default:
		    assert(sizeof(everyC) == 10);
		    if (sscanf(string, "%*s %*s %*s %*s %*s %9s",everyC) == 1){
			assert(sizeof(timesC) == 6);
			if (sscanf(string, "%*s %*s %*s %*s %*s %*s %d %5s",
							&every, timesC) != 2)
			    retStat = AR_S_SYNERR;
		    }
		    break;
	    }
	    if (strlen(everyC) > 0) {
		if (strcmp(everyC, glArKwEvery) != 0) {
		    (void)printf("%s keyword expected\n", glArKwEvery);
		    evType = AR_EV_UNDEF;
		    retStat = AR_S_SYNERR;
		}
		else if (every < 0 || every > 30000) {
		    (void)printf("illegal %s count\n", glArKwEvery);
		    evType = AR_EV_UNDEF;
		    retStat = AR_S_SYNERR;
		}
		else if (strcmp(timesC, glArKwTimes) != 0) {
		    (void)printf("%s keyword expected\n", glArKwTimes);
		    evType = AR_EV_UNDEF;
		    retStat = AR_S_SYNERR;
		}
	    }
	    else		/* 'every' not specified */
		every = 1;

/*----------------------------------------------------------------------------
*        finally, store the event type into the method structure.  If an error
*        has been detected, the event type will be AR_EV_UNDEF.
*
*        also store the 'every' count now
*---------------------------------------------------------------------------*/
	    pArMeth->ctrl.event.type = evType;
	    pArMeth->ctrl.event.count = every;

/*----------------------------------------------------------------------------
*    ERROR on event method parsing will result in method type of AR_METH_UNDEF
*---------------------------------------------------------------------------*/
	    if (retStat != AR_S_OK)
		type = AR_METH_UNDEF;
	}
	else if (strcmp(typeC, glArMethKw[AR_METH_AFTER]) == 0) {
/*----------------------------------------------------------------------------
* AFTER
*	XXX: after n samples
*---------------------------------------------------------------------------*/
	    type = AR_METH_AFTER;
	    afterC[0] = '\0';
	    if (sscanf(string, "%*s %*s %ld %6s", &afterCount, afterC) < 1) {
		(void)printf("improper count for %s\n", typeC);
		retStat = AR_S_SYNERR;
	    }
	    else {
		if (afterCount <= 0 || afterCount > 999999999) {
		    (void)printf("improper count for %s\n", typeC);
		    retStat = AR_S_SYNERR;
		}
		else if (strncmp(afterC, "sample", 6) != 0) {
		    (void)printf("\"samples\" keyword missing\n");
		    retStat = AR_S_SYNERR;
		}
		else
		    pArMeth->ctrl.afterCount = afterCount;
	    }
	}
	else if (strcmp(typeC, glArMethKw[AR_METH_MONITOR]) == 0)
/*----------------------------------------------------------------------------
* MONITOR
*	XXX: monitor
*---------------------------------------------------------------------------*/
	    type = AR_METH_MONITOR;
	else if (strcmp(typeC, glArMethKw[AR_METH_DEMAND]) == 0)
/*----------------------------------------------------------------------------
* DEMAND
*	XXX: demand
*---------------------------------------------------------------------------*/
	    type = AR_METH_DEMAND;
	else if (strcmp(typeC, glArMethKw[AR_METH_ALWAYS]) == 0)
/*----------------------------------------------------------------------------
* ALWAYS
*	XXX: always
*---------------------------------------------------------------------------*/
	    type = AR_METH_ALWAYS;
	else if (strcmp(typeC, glArMethKw[AR_METH_NEVER]) == 0)
/*----------------------------------------------------------------------------
* NEVER
*	XXX: never
*---------------------------------------------------------------------------*/
	    type = AR_METH_NEVER;
	else {
/*----------------------------------------------------------------------------
* ILLEGAL
*---------------------------------------------------------------------------*/
	    type = AR_METH_UNDEF;
	    (void)printf("illegal type field\n");
	    retStat = AR_S_SYNERR;
	}
    }

    pArMeth->type = type;
    pArMeth->state = AR_METH_ST_UNDEF;
    pArMeth->pNext = NULL;
    pArMeth->pPrev = NULL;
    pArMeth->actionType = AR_METH_AC_UNDEF;
    pArMeth->demand = 0;

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arMethBuildCheck - do some `post ca_search' checks
*
* DESCRIPTION
*	Do final validation for a method to wrapup the `method build'
*	process which was begun by arReadReq and arMethBuild.  This
*	final validation follows the completion of the connections to
*	Channel Access for EVENT methods.  (Many of the validation
*	activities are done by the arReadReq and arMethBuild ).
*
* RETURNS
*	AR_S_OK, or
*	AR_S_SYNERR if a problem is found
*
*
* SEE ALSO
*	arMethBuild, arReadReq, and arReadReqCheck
*
*-*/
int
arMethBuildCheck(pArMeth)
AR_METH	*pArMeth;	/* I pointer to method */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    char	message[80];	/* temp for error messages */
    chid	pCh;		/* temp for channel pointer */
    char	string[AR_STRVAL_DIM];/* temp for comparison value text */

    assert(pArMeth != NULL);

/*----------------------------------------------------------------------------
* EVENT
*---------------------------------------------------------------------------*/
    if (pArMeth->type == AR_METH_EVENT) {
	pCh = pArMeth->ctrl.event.pCh;
	if (ca_element_count(pCh) != 1) {
	    (void)sprintf(message, "channel isn't scalar: %s", ca_name(pCh));
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	if (pArMeth->ctrl.event.type > AR_EV_OK_NON_NUM &&
		ca_field_type(pCh) != DBF_SHORT &&
		ca_field_type(pCh) != DBF_LONG &&
		ca_field_type(pCh) != DBF_DOUBLE &&
		ca_field_type(pCh) != DBF_FLOAT) {
	    (void)sprintf(message, "channel isn't numeric: %s", ca_name(pCh));
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	if (pArMeth->ctrl.event.type != AR_EV_CHANGE) {
/*----------------------------------------------------------------------------
*    check the comparison value (except for AR_EV_CHANGE), and
*    store into event structure.  The type of value must match the
*    native field type.
*---------------------------------------------------------------------------*/
	    (void)strcpy(string, pArMeth->ctrl.event.compVal.strval);
	    switch (ca_field_type(pCh)) {
		case DBF_STRING:
		case DBF_ENUM:
		    assert(AR_STRVAL_DIM == 40);
		    if (sscanf(string, "%39s",
                                   pArMeth->ctrl.event.compVal.strval) != 1) {
			(void)sprintf(message,
				"comparison value illegal: %s", string);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		    break;
		case DBF_SHORT:
		    if (sscanf(string, "%h",
				   &pArMeth->ctrl.event.compVal.shrtval) != 1) {
			(void)sprintf(message,
				"comparison value illegal: %s", string);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		    break;
		case DBF_FLOAT:
		    if (sscanf(string, "%f",
                                   &pArMeth->ctrl.event.compVal.fltval) != 1) {
			(void)sprintf(message,
				"comparison value illegal: %s", string);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		    break;
		default:
		    (void)sprintf(message, "channel type illegal: %s",
							        ca_name(pCh));
		    (void)arLogMessage(" ", message, NULL);
		    retStat = AR_S_SYNERR;
		    break;
	    }
	}    
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arMethDel - remove a method from a method queue
*
* DESCRIPTION
*	Removes a method from a method queue, maintaining the integrity of
*	the links for the methods left in the queue.  Method state is set
*	to AR_METH_ST_UNDEF; method action type is set to AR_METH_AC_UNDEF.
*
*	The Channel Access monitor is canceled for event method types.
*
*	For monitor method types (which means only SAMPLE methods), a
*	ca_clear_event is done for each element in the method's request.
*	If the element status was AR_EL_MONITOR, it is set to AR_EL_CONNOK.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_REQ_UNDEF if method's request has UNDEF as its status
*
* BUGS
* o	for event methods, there is a possible race between ca_clear_event
*	and removing the method from the list.  (Channel Access may handle
*	this OK--need to check it.)
* o	this probably isn't the `proper' place to clean up monitors
*	for request elements for monitor methods for sampling.
*
* EXAMPLE
*	AR_REQ	*pReq;		pointer to archive request
*
*	arMethDel(&pReq->act);
*
*-*/
int
arMethDel(pMeth)
AR_METH	*pMeth;		/* IO pointer to method */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    AR_EL	*pEl;		/* temp for winding down MONITOR mode */

    assert(pMeth != NULL);
    assert(pMeth->pReq != NULL);
    assert(pMeth->pReq->status != AR_REQ_UNDEF);

    if (pMeth->actionType == AR_METH_AC_UNDEF) {
	assert(pMeth->pNext == NULL);
	assert(pMeth->pPrev == NULL);
	;	/* no action */
    }
    else if (pMeth->type == AR_METH_PERIODIC || pMeth->type == AR_METH_TIMED) {
/*----------------------------------------------------------------------------
* PERIODIC
* TIMED
*---------------------------------------------------------------------------*/
	DoubleListRemove(pMeth, pglArCtx->pMethClkHead,
						pglArCtx->pMethClkTail);
    }
    else if (pMeth->type == AR_METH_EVENT) {
/*----------------------------------------------------------------------------
* EVENT
*---------------------------------------------------------------------------*/
	if (pMeth->ctrl.event.pEv != NULL) {
	    if ((stat = ca_clear_event(pMeth->ctrl.event.pEv)) != ECA_NORMAL) {
		(void)arCaStatusMessage(stat, "clearing method monitor",
							pMeth->ctrl.event.pCh);
	    }
	    pMeth->ctrl.event.pEv = NULL;
	}
	DoubleListRemove(pMeth, pglArCtx->pMethEvHead, pglArCtx->pMethEvTail);
    }
    else if (pMeth->type == AR_METH_MONITOR) {
/*----------------------------------------------------------------------------
* MONITOR
*----------------------------------------------------------------------------*/
	pEl = pMeth->pReq->pElListHead;
	while (pEl != NULL) {
	    if (pEl->status == AR_EL_MONITOR)
		pEl->status = AR_EL_CONNOK;
	    if (pEl->pEv != NULL) {
		if ((stat = ca_clear_event(pEl->pEv)) != ECA_NORMAL) {
		    (void)arCaStatusMessage(stat, "clearing method monitor",
							pMeth->ctrl.event.pCh);
		}
		pEl->pEv = NULL;
	    }
	    pEl = pEl->pNext;
	}
	DoubleListRemove(pMeth, pglArCtx->pMethMiscHead,
					    	pglArCtx->pMethMiscTail);
    }
    else if (pMeth->type == AR_METH_DEMAND  ||
	     pMeth->type == AR_METH_AFTER   ||
	     pMeth->type == AR_METH_ALWAYS  ) {
/*----------------------------------------------------------------------------
* DEMAND
* AFTER
* ALWAYS
*---------------------------------------------------------------------------*/
	DoubleListRemove(pMeth, pglArCtx->pMethMiscHead,
					    	pglArCtx->pMethMiscTail);
    }

    pMeth->state = AR_METH_ST_UNDEF;
    pMeth->actionType = AR_METH_AC_UNDEF;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arMethEventHandlerCa - change handler for method channels
*
* DESCRIPTION
*
* RETURNS
*
* BUGS
* o	This routine acts on values without regard to their status--invalid
*	values (e.g., hardware error or network error) shouldn't be used.
* o	The time stamp for the channel value is generated locally.  The
*	monitor for the channel should be a DBR_TIME_xxx.
*
* SEE ALSO
*
* EXAMPLE
*
* NOTES
* 1.	This routine assumes that DBF_ENUM channels were accessed as DBR_STRING
*	when the ca_add_event was done.
*
*-*/
void
arMethEventHandlerCa(arg)
struct event_handler_args arg;
{
    AR_METH	*pMeth;		/* IO method pointer */
    chid	pCh;		/* I channel pointer */
    chtype	type;		/* I receive buffer type */
    long	count;		/* I element count */
    void	*pVal;		/* I pointer to value */

    int		comp;		/* result of comparison: <0, 0, >0 for
				value is <, ==, > comparison value */
    int		call=0;		/* non-zero says "call the action routine"
				(subject to counter = 0) */
    char	message[80];	/* temp for error messages */

    pMeth = (AR_METH *)arg.usr;
    pCh = arg.chid;
    type = arg.type;
    count = arg.count;
    pVal = arg.dbr;

    assert(pMeth != NULL);
    assert(pCh != NULL);
    assert(pVal != NULL);

    if (pMeth->state == AR_METH_ST_SUCCESS) {
	(void)sprintf(message, "request: %s %s method",
		pMeth->pReq->name, ArMethActionToText(pMeth->actionType));
	(void)arLogMessage(" ", message, NULL);
	(void)arLogMessage(" ", "event lost; processing prior event not done",
									NULL);
	return;
    }

/*----------------------------------------------------------------------------
*    store the new value, with time stamp, into the method structure.  The
*    previous value is stored in '.oldVal'.  (If state is NEW, then this is
*    the first value from the Channel Access monitor; set '.oldVal' to this
*    new value, instead of to the previous value, which doesn't exist.)
*---------------------------------------------------------------------------*/
    (void)tsLocalTime(&pMeth->ctrl.event.timeStamp);
    switch (type) {
	case DBF_SHORT:
	    if (pMeth->state == AR_METH_ST_NEW) {
		pMeth->ctrl.event.oldVal.shrtval = *(short *)pVal;
	        pMeth->state = AR_METH_ST_UNARMED;
	    }
	    else
		pMeth->ctrl.event.oldVal.shrtval =
						pMeth->ctrl.event.val.shrtval;
	    pMeth->ctrl.event.val.shrtval = *(short *)pVal;
	    break;
	case DBF_FLOAT:
	    if (pMeth->state == AR_METH_ST_NEW) {
		pMeth->ctrl.event.oldVal.fltval = *(float *)pVal;
	        pMeth->state = AR_METH_ST_UNARMED;
	    }
	    else
		pMeth->ctrl.event.oldVal.fltval = pMeth->ctrl.event.val.fltval;
	    pMeth->ctrl.event.val.fltval = *(float *)pVal;
	    break;
	default:
	    if (pMeth->state == AR_METH_ST_NEW) {
		strcpy(pMeth->ctrl.event.oldVal.strval, (char *)pVal);
	        pMeth->state = AR_METH_ST_UNARMED;
	    }
	    else
		strcpy(pMeth->ctrl.event.oldVal.strval,
						pMeth->ctrl.event.val.strval);
	    strcpy(pMeth->ctrl.event.val.strval, (char *)pVal);
	    break;
    }

/*----------------------------------------------------------------------------
*    check the condition.  If the condition is satisfied, 'call' will be
*    one.
*---------------------------------------------------------------------------*/
    if (pMeth->ctrl.event.type == AR_EV_CHANGE) {
/*----------------------------------------------------------------------------
*	for CHANGE condition, the first 'change' is ignored, and all
*	succeeding changes satisfy the condition.  The first 'change' is just
*	the channel access monitor feeding in the initial value for the
*	channel.
*---------------------------------------------------------------------------*/
	if (pMeth->state == AR_METH_ST_ARMED)
	    call = 1;
	else
	    pMeth->state = AR_METH_ST_ARMED;
    }
    else {
/*----------------------------------------------------------------------------
*	for other conditions besides CHANGE, a little more logic is needed.
*	First the value is compared with the comparison value.  The result
*	of the comparison and the condition type are used together to decide
*	if the condition has been satisfied.
*
*	Whether to request a call (subject to counter) depends on the present
*	and the previous state of 'condition satisfied'.  If there is no
*	change (i.e., the condition was previously satisfied and is also
*	presently satisfied, or similarly for unsatisfied), then 'call' will
*	be set to zero.  If there is a transition to 'satisfied', 'call' will
*	be set to one.
*
*	When the condition is 'satisfied', state will be set to ARMED;
*	otherwise state will be UNARMED.  State is used to detect transitions
*	from 'unsatisfied' to 'satisfied'.
*---------------------------------------------------------------------------*/
	switch (type) {
	    case DBF_SHORT:
		if (*(short *)pVal == pMeth->ctrl.event.compVal.shrtval)
		    comp = 0;
		else if (*(short *)pVal < pMeth->ctrl.event.compVal.shrtval)
		    comp = -1;
		else
		    comp = 1;
		break;
	    case DBF_FLOAT:
		if (*(float *)pVal == pMeth->ctrl.event.compVal.fltval)
		    comp = 0;
		else if (*(float *)pVal < pMeth->ctrl.event.compVal.fltval)
		    comp = -1;
		else
		    comp = 1;
		break;
	    default:
		comp = strcmp((char *)pVal, pMeth->ctrl.event.compVal.strval);
		break;
	}
	if (comp == 0) {		/* value == compValue */
	    if (  pMeth->ctrl.event.type == AR_EV_EQ ||
	          pMeth->ctrl.event.type == AR_EV_LE ||
	          pMeth->ctrl.event.type == AR_EV_GE  )
		call = 1;
	}
	else if (comp < 0) {		/* value < compValue */
	    if (  pMeth->ctrl.event.type == AR_EV_NEQ ||
	          pMeth->ctrl.event.type == AR_EV_LT ||
	          pMeth->ctrl.event.type == AR_EV_LE  )
		call = 1;
	}
	else if (comp > 0) {		/* value > compValue */
	    if (  pMeth->ctrl.event.type == AR_EV_NEQ ||
	          pMeth->ctrl.event.type == AR_EV_GT ||
	          pMeth->ctrl.event.type == AR_EV_GE  )
		call = 1;
	}
	if (call) {
	    if (pMeth->state == AR_METH_ST_UNARMED)
	        pMeth->state = AR_METH_ST_ARMED; /* newly satis.; call */
	    else
		call = 0;	/* not transition to satisfied; don't call */
	}
	else
	    pMeth->state = AR_METH_ST_UNARMED;	/* not satisfied */
    }

/*----------------------------------------------------------------------------
*    check the counter.  If 'call' is one, decrement the counter.  If the
*    counter is zero, set the method state to AR_METH_ST_SUCCESS.
*    The method's date stamp and time stamp are set to the present time.
*
*    Since this routine is operating as a Channel Access handler, the
*    action routines aren't actually called from here.  The routine which
*    calls the action routines must re-start the counter.
*---------------------------------------------------------------------------*/
    if (call) {
	if (--pMeth->ctrl.event.counter <= 0) {
	    (void)tsLocalTime(&pMeth->timeStamp);
	    pMeth->state = AR_METH_ST_SUCCESS;
	}
    }

    return;
}

/*+/subr***********************************************************************
* NAME	arMethGetInfoAsText - get information about a satisfied method
*
* DESCRIPTION
*	Prepare a status line for a method whose condition has just been
*	satisfied.  This status line answers the question "Why is action
*	being taken?".
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	sprintf is used, and could overrun local char arrays
*
* EXAMPLE
*	stat = arMethGetInfoAsText("ACT:", &pArReq->act,
*						textBuf, sizeof textBuf);
*
*-*/
int
arMethGetInfoAsText(desc, pArMeth, buf, bufSize)
char	*desc;		/* I descriptor to print at beginning of line */
AR_METH	*pArMeth;	/* I pointer to method structure in request */
char	*buf;		/* O buffer for text information for method */
int	bufSize;	/* I size of buffer */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    char	message[300];	/* temp for information buffer */
    int		type;		/* method type */
    char	t1C[28];	/* time stamp, as text */
    char	t2C[28];	/* time stamp, as text */
    char	line1[80];		/* first line of info text */
    char	line2[200];		/* second line of info text */
    char	val[AR_STRVAL_DIM];	/* value as text */
    char	compVal[AR_STRVAL_DIM];	/* comp value as text */
    char	oldVal[AR_STRVAL_DIM];	/* old value as text */

    assert(pArMeth != NULL);
    assert(buf != NULL);
    assert(bufSize > 1);

    type = pArMeth->type;
    if (type == AR_METH_EVENT) {
/*----------------------------------------------------------------------------
* EVENT
*	'desc' at mm/dd/yy hh:mm:ss due to EVENT at hh:mm:ss.mmm after n times
*	'desc' chanName = value [cond compVal] changed from: oldVal
*---------------------------------------------------------------------------*/
	switch (ca_field_type(pArMeth->ctrl.event.pCh)) {
	    case DBF_SHORT:
		(void)sprintf(val, "%d", pArMeth->ctrl.event.val.shrtval);
		(void)sprintf(compVal, "%d",
					pArMeth->ctrl.event.compVal.shrtval);
		(void)sprintf(oldVal, "%d", pArMeth->ctrl.event.oldVal.shrtval);
		break;
	    case DBF_FLOAT:
		(void)sprintf(val, "%f", pArMeth->ctrl.event.val.fltval);
		(void)sprintf(compVal, "%f",
					pArMeth->ctrl.event.compVal.fltval);
		(void)sprintf(oldVal, "%f", pArMeth->ctrl.event.oldVal.fltval);
		break;
	    default:
		(void)sprintf(val, "%s", pArMeth->ctrl.event.val.strval);
		(void)sprintf(compVal, "%s",
					pArMeth->ctrl.event.compVal.strval);
		(void)sprintf(oldVal, "%s", pArMeth->ctrl.event.oldVal.strval);
		break;
	}
	(void)sprintf(line1, "%s at %s due to EVENT at %s after %d times",
		desc, tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C),
			tsStampToText(&pArMeth->ctrl.event.timeStamp,
							TS_TEXT_MMDDYY, t2C),
			pArMeth->ctrl.event.count);
	if (pArMeth->ctrl.event.type == AR_EV_CHANGE   ) {
	    /* for 'change', don't print [cond compVal] */
	    (void)sprintf(line2, "%s (%s = %s) changed from: %s", desc,
		ca_name(pArMeth->ctrl.event.pCh), val, oldVal);
        }
	else {
	    (void)sprintf(line2, "%s (%s = %s) %s %s changed from: %s", desc,
		ca_name(pArMeth->ctrl.event.pCh), val,
		ArEvTypeToText(pArMeth->ctrl.event.type), compVal, oldVal);
        }
	(void)sprintf(message, "%s\n%s\n", line1, line2);
    }
    else if (type == AR_METH_PERIODIC) {
/*----------------------------------------------------------------------------
* PERIODIC
*	'desc' at mm/dd/yy hh:mm:ss due to PERIOD of hh:mm:ss.mmm expiring
*---------------------------------------------------------------------------*/
	if (pArMeth->actionType == AR_METH_AC_SAMPLE &&
					pArMeth->pReq->sampleCount == 0) {
	    /* for 'initial sample', treat special */
	    (void)sprintf(message, "%s at %s--initial sample after activate\n",
		desc, tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C));
	}
	else {
	    (void)sprintf(message, "%s at %s due to PERIOD of %s expiring\n",
		desc, tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C),
			tsStampToText(&pArMeth->ctrl.per.period,
							TS_TEXT_MMDDYY, t2C));
	}
    }
    else if (type == AR_METH_TIMED) {
/*----------------------------------------------------------------------------
* TIMED
*	'desc' at mm/dd/yy hh:mm:ss due to TIMER of hh:mm:ss.mmm expiring
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s at %s due to TIMER of %s expiring\n",
		desc, tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C),
			tsStampToText(&pArMeth->ctrl.per.period,
							TS_TEXT_MMDDYY, t2C));
    }
    else if (type == AR_METH_MONITOR) {
/*----------------------------------------------------------------------------
* MONITOR
*	'desc' monitoring; state is xxx
*----------------------------------------------------------------------------*/
	(void)sprintf(message, "%s monitoring; state is %s", desc,
					ArMethStateToText(pArMeth->state));
    }
    else if (type == AR_METH_AFTER) {
/*----------------------------------------------------------------------------
* AFTER
*	'desc' at mm/dd/yy hh:mm:ss due to sampleCount of nnn
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s at %s due to sampleCount of %d >= %d\n",
		desc, tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C),
		pArMeth->pReq->sampleCount, pArMeth->ctrl.afterCount);
    }
    else if (type == AR_METH_DEMAND) {
/*----------------------------------------------------------------------------
* DEMAND
*	'desc' at mm/dd/yy hh:mm:ss due to DEMAND
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s at %s due to DEMAND\n", desc,
		tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C));
    }
    else if (type == AR_METH_ALWAYS) {
/*----------------------------------------------------------------------------
* ALWAYS
*	'desc' at mm/dd/yy hh:mm:ss due to ALWAYS
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s at %s due to ALWAYS\n", desc,
		tsStampToText(&pArMeth->timeStamp, TS_TEXT_MMDDYY, t1C));
    }
    else
/*----------------------------------------------------------------------------
* UNDEFINED
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s undefined method type\n", desc);

/*----------------------------------------------------------------------------
*    copy the text to caller's buffer.  If text is too long to fit,
*    make sure caller's buffer is null terminated.
*---------------------------------------------------------------------------*/
    strncpy(buf, message, bufSize);
    buf[bufSize-1] = '\0';

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arMethPrintInfo - print information about an archive method
*
* DESCRIPTION
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	sprintf is used, and could possibly overrun string buffer
*
* NOTES
* 1.	This routine assumes that channel types of DBF_ENUM have been
*	specified with DBR_STRING in ca_add_event calls.
*
* EXAMPLE
*	stat = arMethPrintInfo(&pArReq->act, pReq->name);
*
*-*/
int
arMethPrintInfo(pArMeth, desc)
AR_METH	*pArMeth;	/* I pointer to method structure in request */
char	*desc;		/* I descriptor for method, used in printing info */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    char	message[80];	/* temp for log messages */
    int		type;		/* method type */
    char	t1C[28];	/* time stamp, as text */
    char	t2C[28];	/* time stamp, as text */

    assert(pArMeth != NULL);

    type = pArMeth->type;
/*----------------------------------------------------------------------------
*	`desc' `methType' queued to `action'; state is `state'
*---------------------------------------------------------------------------*/
    if (pArMeth->actionType != AR_METH_AC_UNDEF) {
	(void)sprintf(message, "%s %s queued to %s; state is %s", desc,
	    ArMethTypeToText(type), ArMethActionToText(pArMeth->actionType),
	    ArMethStateToText(pArMeth->state));
    }
    else {
	(void)sprintf(message, "%s %s; not queued", desc,
	    						ArMethTypeToText(type));
    }
    (void)arLogMessage(" ", message, NULL);

    if (type == AR_METH_EVENT) {
/*----------------------------------------------------------------------------
* EVENT
*	`desc'		`chanName' `condition'
*	`desc' 		`compVal'
*	`desc' 		every `n' times; count=`m'
*	`desc' 		channel not connected		(optional line)
*---------------------------------------------------------------------------*/
        (void)sprintf(message, "%s %s %s", desc,
				pArMeth->ctrl.event.name,
				ArEvTypeToText(pArMeth->ctrl.event.type));
	(void)arLogMessage(" ", message, NULL);
	if (  pArMeth->ctrl.event.type != AR_EV_UNDEF   &&
	      pArMeth->ctrl.event.type != AR_EV_CHANGE   ) {
	    switch (ca_field_type(pArMeth->ctrl.event.pCh)) {
		case DBF_STRING:
		case DBF_ENUM:
		    (void)sprintf(message, "%s	%s", desc,
					pArMeth->ctrl.event.compVal.strval);
		    (void)arLogMessage(" ", message, NULL);
		    break;
		case DBF_SHORT:
		    (void)sprintf(message, "%s	%d", desc,
				    pArMeth->ctrl.event.compVal.shrtval);
		    (void)arLogMessage(" ", message, NULL);
		    break;
		case DBF_FLOAT:
		    (void)sprintf(message, "%s	%f", desc,
					pArMeth->ctrl.event.compVal.fltval);
		    (void)arLogMessage(" ", message, NULL);
		    break;
		default:
		    (void)sprintf(message, "%s	illegal field type", desc);
		    (void)arLogMessage(" ", message, NULL);
		    break;
	    }
	}
	(void)sprintf(message, "%s	every %d times; count=%d; ", desc,
			pArMeth->ctrl.event.count, pArMeth->ctrl.event.counter);
	(void)arLogMessage(" ", message, NULL);
	if (pArMeth->ctrl.event.pCh == NULL ||
	    ! dbf_type_is_valid(ca_field_type(pArMeth->ctrl.event.pCh)) ) {
	    (void)sprintf(message, "%s	channel not connected", desc);
	    (void)arLogMessage(" ", message, NULL);
	}
    }
    else if (type == AR_METH_PERIODIC || type == AR_METH_TIMED) {
/*----------------------------------------------------------------------------
* PERIODIC
* TIMED
*---------------------------------------------------------------------------*/
	(void)sprintf(message, "%s	interval is %s", desc,
		tsStampToText(&pArMeth->ctrl.per.period, TS_TEXT_MMDDYY, t1C));
	(void)arLogMessage(" ", message, NULL);
	(void)sprintf(message, "%s	next action at %s", desc,
		tsStampToText(&pArMeth->ctrl.per.nextTime, TS_TEXT_MMDDYY,
									t2C));
	(void)arLogMessage(" ", message, NULL);
    }
    else if (type == AR_METH_AFTER) {
/*----------------------------------------------------------------------------
* AFTER
*----------------------------------------------------------------------------*/
	(void)sprintf(message, "%s	after %d samples; count=%d", desc,
			pArMeth->ctrl.afterCount, pArMeth->pReq->sampleCount);
	(void)arLogMessage(" ", message, NULL);
    }
    else if (type == AR_METH_MONITOR) {
/*----------------------------------------------------------------------------
* MONITOR
*----------------------------------------------------------------------------*/
	(void)sprintf(message, "%s monitoring; state is %s", desc,
					ArMethStateToText(pArMeth->state));
	(void)arLogMessage(" ", message, NULL);
    }
    else if (type == AR_METH_DEMAND ||
	     type == AR_METH_ALWAYS ||
	     type == AR_METH_NEVER) {
/*----------------------------------------------------------------------------
* DEMAND
* ALWAYS
* NEVER
*---------------------------------------------------------------------------*/
	;    /* no action */
    }
    else {
/*----------------------------------------------------------------------------
* UNDEFINED
*---------------------------------------------------------------------------*/
	;    /* no action */
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arMethSetCheck - check method lists and call action routines
*
* DESCRIPTION
*	Checks the various method lists, taking appropriate action as
*	necessary for individual methods.  When necessary, method action
*	routines are called.
*
*	Much of the handling for methods of EVENT type is done by
*	arMethEventHandlerCa.  When that routine detects a need to call
*	the method action routine, it changes the method state to SUCCESS;
*	this routine then calls the action routine.
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	doesn't allow demand for any except DEMAND method types.
*
*-*/
int
arMethSetCheck()
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    AR_METH	*pMeth;		/* pointer to a method */
    TS_STAMP	timeStamp;	/* time stamp for present time */

/*----------------------------------------------------------------------------
* PERIODIC
* TIMED
*	When the present time is greater than the 'nextTime' specified
*	in the method, call the method's action routine.  Before calling the
*	action routine, 'nextTime' is updated to schedule the next action.
*
*	Before calling the action routine, the present time stamp is
*	stored into the method structure.
*
*	In this routine, no distinction is made between PERIODIC and TIMED
*	methods.
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethClkHead;
    while (pMeth != NULL) {
	(void)tsLocalTime(&timeStamp);
	if (pMeth->state == AR_METH_ST_ARMED &&
	    TsCmpStampsGE(&timeStamp, &pMeth->ctrl.per.nextTime)) {
/*----------------------------------------------------------------------------
*		method has succeeded.  Perform several chores:
*		o  set up for next scheduled time
*		o  save time/date for when action initiated
*		o  initiate the action
*---------------------------------------------------------------------------*/
	    /* set next scheduled time */
	    TsAddStamps(&pMeth->ctrl.per.nextTime,
			&pMeth->ctrl.per.nextTime, &pMeth->ctrl.per.period);

	    (void)tsLocalTime(&pMeth->timeStamp);

	    if (pMeth->actionType == AR_METH_AC_ACT)
		    arMethActionAct(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_INACT)
		    arMethActionInact(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_SAMPLE)
		    arMethActionSample(pMeth);
	}
	pMeth = pMeth->pNext;
    }

/*----------------------------------------------------------------------------
* EVENT
*	handle the methods tagged for action by arMethEventHandlerCa
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethEvHead;
    while (pMeth != NULL) {
	if (pMeth->state == AR_METH_ST_SUCCESS) {
/*----------------------------------------------------------------------------
*               method has succeeded.  Perform several chores:
*               o  initiate the action
*               o  reset the counter.  (This must be done prior to changing
*			the state back to UNARMED.)
*               o  move method state to UNARMED (or to ARMED, for CHANGE event)
*---------------------------------------------------------------------------*/
	    if (pMeth->actionType == AR_METH_AC_ACT)
		    arMethActionAct(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_INACT)
		    arMethActionInact(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_SAMPLE)
		    arMethActionSample(pMeth);
	    pMeth->ctrl.event.counter = pMeth->ctrl.event.count;
	    if (pMeth->ctrl.event.type != AR_EV_CHANGE)
		pMeth->state = AR_METH_ST_UNARMED;
	    else
		pMeth->state = AR_METH_ST_ARMED;
	}
	pMeth = pMeth->pNext;
    }

/*----------------------------------------------------------------------------
* MONITOR
* AFTER
* ALWAYS
* DEMAND
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethMiscHead;
    while (pMeth != NULL) {
	if (pMeth->type == AR_METH_AFTER) {
	    if (pMeth->pReq->sampleCount >= pMeth->ctrl.afterCount)
		pMeth->state = AR_METH_ST_ARMED;
	}
	else if (pMeth->type == AR_METH_DEMAND) {
	    if (pMeth->demand) {
		pMeth->state = AR_METH_ST_ARMED;
		pMeth->demand = 0;
	    }
	}
	if (pMeth->state == AR_METH_ST_ARMED) {
/*----------------------------------------------------------------------------
*               method has succeeded.  Perform several chores:
*               o  initiate the action
*               o  for DEMAND methods, move to UNARMED
*---------------------------------------------------------------------------*/
	    (void)tsLocalTime(&pMeth->timeStamp);

	    if (pMeth->actionType == AR_METH_AC_ACT)
		    arMethActionAct(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_INACT)
		    arMethActionInact(pMeth);
	    else if (pMeth->actionType == AR_METH_AC_SAMPLE)
		    arMethActionSample(pMeth);
	    if (pMeth->type == AR_METH_DEMAND)
		pMeth->state = AR_METH_ST_UNARMED;
	}
	pMeth = pMeth->pNext;
    }

    return AR_S_OK;
}

/*+/subr***********************************************************************
* NAME	arMethSetPrintInfo - print info for all methods on all lists
*
* DESCRIPTION
*	Checks the various method lists, printing information for each
*	individual method.
*
* RETURNS
*	AR_S_OK
*
* BUGS
*
*-*/
int
arMethSetPrintInfo()
{
    AR_METH	*pMeth;		/* pointer to a method */

/*----------------------------------------------------------------------------
* PERIODIC
* TIMED
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethClkHead;
    (void)arLogMessage(" ", "**** CLOCK-BASED method queue ****", NULL);
    while (pMeth != NULL) {
	(void)arMethPrintInfo(pMeth, pMeth->pReq->name);
	pMeth = pMeth->pNext;
    }
/*----------------------------------------------------------------------------
* EVENT
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethEvHead;
    (void)arLogMessage(" ", "**** EVENT method queue ****", NULL);
    while (pMeth != NULL) {
	(void)arMethPrintInfo(pMeth, pMeth->pReq->name);
	pMeth = pMeth->pNext;
    }
/*----------------------------------------------------------------------------
* AFTER
* ALWAYS
* DEMAND
* MONITOR
*---------------------------------------------------------------------------*/
    pMeth = pglArCtx->pMethMiscHead;
    (void)arLogMessage(" ", "**** MISC method queue ****", NULL);
    while (pMeth != NULL) {
	(void)arMethPrintInfo(pMeth, pMeth->pReq->name);
	pMeth = pMeth->pNext;
    }

    return AR_S_OK;
}
