/*	@(#)arLibREQ.c	1.2 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03	10-21-91	rac	split into smaller files; handle REQ_STATUS:
 *				and SAMP_STATUS: commands; treat lines
 *				starting with # as comments; use nextXxx
 *				routines for scanning
 * .04	12-05-91	rac	don't use LWP
 *
 * make options
 *	-DNDEBUG	don't compile assert checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arLib - routines for handling archive `requests'
*
* DESCRIPTION
*
*-***************************************************************************/

#include <stdio.h>
#include <sys/file.h>	/* O_RDWR definition */

#include <tsDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>

/*+/subr***********************************************************************
* NAME	arReadEnabled - read the enable file, put requests on read list
*
* DESCRIPTION
*	The request names in ./arReq/enabled.dat are read.  For each request
*	named in the file, an archive request is created and the request
*	is placed on the `read list'.  Blank lines and lines starting with
*	"#" are ignored.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_FOPENERR	if the file containing the list of enabled requests
*			couldn't be opened (an error message has been
*			printed)
*	status from various other arXxx routines if an error is encountered	
*			(an error message is printed mentioning the file,
*			request, or channel name, as appropriate)
*
* SEE ALSO
*	arReadQCheck - supervises and checks the reading of archive request
*			files
*	arReadReq - reads an archive request file into a request skeleton
*	arReadReqWrapup - checks an archive request for validity after
*			its channels have been connected
*
*-*/
int
arReadEnabled()
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    char	message[80];	/* temp for error messages */
    FILE*	pStream;	/* stream pointer for enabled.dat */
    char	record[81];	/* record from file */
    char	req[AR_NAME_DIM];	/* request name from file */
    AR_REQ	*pNewReq;	/* request pointer */
    AR_REQ	*arReqFind();	/* search a list for a named request */
    int		i;

/*----------------------------------------------------------------------------
*    open and read the file, ignoring blank lines.  Duplicate requests
*    are rejected.  For each request, create a request item at the end of
*    the read list, which can be thought of as a queue of requests to read
*    archive requests files.  Lines beginning with '#' are ignored.
*---------------------------------------------------------------------------*/
    if ((pStream = fopen(AR_FILE_ENB_PATH, "r")) == NULL) {
	(void)sprintf(message, "couldn't open for enable: %s",
							AR_FILE_ENB_PATH);
	(void)arLogMessage(" ", message, NULL);
	retStat = AR_S_FOPENERR;
    }
    else {
	(void)sprintf(message, "**** reading: %s ****",
							AR_FILE_ENB_PATH);
	(void)arLogMessage(" ", message, NULL);
	while (fgets(record, 80, pStream) != NULL) {
	    if ((i = strlen(record)) > 0) {
		if (record[i-1] == '\n')
		    record[i-1] = '\0';	/* remove \n before logging record */
	    }
	    (void)arLogMessage(" ", req, NULL);
	    if (strlen(record) > 0 && record[0] != '#') {
		assert(AR_NAME_DIM == 36);
		if (sscanf(record, "%35s", req) == 1) {
		    stat = arReadQAddReq(req);
		}
	    }
	}
	(void)fclose(pStream);
	(void)sprintf(message, "**** EOF for %s ****", AR_FILE_ENB_PATH);
	(void)arLogMessage(" ", message, NULL);
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReadQAddReq - queue a request name on the list for reading
*
* DESCRIPTION
*	Create an archive request for the supplied request name, and place
*	the new request on the `read list'.  If a request already exists
*	with the specified name, nothing is done.
*
*	The new request is given status of AR_REQ_CONN_NONE to trigger reading
*	the request file.
*
* RETURNS
*	AR_S_OK, or
*	status from various other arXxx routines if an error is encountered	
*			(an error message is printed mentioning the file,
*			request, or channel name, as appropriate)
*
* BUGS
* o	only the main request list is checked for duplicates.  The read
*	list and the connect list should also be checked.
*
* SEE ALSO
*	arReadQCheck - supervises and checks the reading of archive request
*			files
*	arReadReq - reads an archive request file into a request skeleton
*	arReadReqWrapup - checks an archive request for validity after
*			its channels have been connected
*
* SPECIAL NOTE
*	arCmdProcess needs some of the same processing done by the
*	following pseudo-code, to be used when a channel is added to an
*	active request.  Since no actual read is done of a request file, and,
*	thus, arReadQAddReq can't be used, arCmdProcess "fakes" some
*	of the following processing.  If the pseudo-code is changed, then
*	the processing for the "open" command should be re-examined.
*
* PSEUDO-CODE
*	The transition to `connects all complete' in the first part of this
*	pseudo-code depends on the connection handler being called, as in
*	the second part of this pseudo-code.
*
*	arReadQAddReq(reqName);	queue request for reading
*	. arReqCreate();		create a request structure
*	..append to read list;		`queue' for reading
*	arReadQCheck();			check on requests being read
*	. if (not queued for connect)
*	. . arReadReq(pReq);		read the text request file
*	. . . arMethBuild(pMeth);	preliminary set up for each method
*	. . . arElCreateAndAppend(name); preliminary set up for elements
*	. . ..do some validation;
*	. ..arConnMake(pReq);	queue channels for connect
*	. else if (connects all complete)
*	. . arReadReqWrapup(pReq);
*	. . . arElCreateWrapup(&pEl);	wrapup for each element
*	. . ..arMethBuildCheck(pMeth);  wrapup for each method
*	....move request to main list;	reading and connecting are complete	
*
*	ca_pend_event();		wait for connect/disconnect
*	arConnHandler();
*	. if (connect)
*	. . if (method)
*	. . ..make sure this is EV method;
*	. . else if (element)
*	......set status to AR_EL_CONNOK;
*
*-*/
int
arReadQAddReq(reqName)
char	*reqName;	/* I request name */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    AR_REQ	*pNewReq;	/* request pointer */
    AR_REQ	*arReqFind();	/* search a list for a named request */

    assert(strlen(reqName) > 0);
    assert(strlen(reqName) < AR_NAME_DIM);

    if ((pNewReq = arReqFind(reqName)) != NULL)
	(void)printf("%s ignored--already exists\n", reqName);
    else {
	stat = arReqCreate(&pNewReq, reqName);

	if (stat == AR_S_OK) {
	    pNewReq->status = AR_REQ_CONN_NONE;
	    pNewReq->readQ = 1;
	    DoubleListAppend(pNewReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail); 
	}
	if (stat != AR_S_OK)
	    retStat = stat;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arReadQCheck - check the read list for needed attention
*
* DESCRIPTION
*	Check the requests on the pReqReadHead list.  The action taken
*	for a request depends on its status:
*
*	AR_REQ_CONN_NONE results in reading the request file and queueing
*	the ca_search's for the channels in the request.  If the reading
*	and processing of the request file encounters errors, then the
*	request is deleted (with appropriate clean up).
*
*	AR_REQ_CONN_CA_BAD results in removing the request from the read
*	list and placing it on the connect list.
*
*	AR_REQ_CONN_OK or AR_REQ_CONN_CF_OK results in doing channel-related
*	validation for the request.  If validation fails, then the request
*	is deleted (with appropriate clean up).  If the validation succeeds,
*	then the request is removed from the read list and appended to the
*	main list.
*
*	If the .disable flag is set in the request, the request is removed
*	from the list and deleted.
*
* RETURNS
*	AR_S_OK
*
* BUGS
* o	no notification is printed if connect never succeeds
*
* SEE ALSO
*	arReadReq - reads an archive request file into a request skeleton
*	arReadReqWrapup - checks an archive request for validity after
*			its channels have been connected
*
*-*/
int
arReadQCheck()
{
    char	message[80];	/* buffer for error messages */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    AR_REQ	*pArReq;	/* request pointer */
    AR_REQ	*pArReqNext;	/* request pointer to next request */

    pArReq = pglArCtx->pReqReadHead;
    while (pArReq != NULL) {
	pArReqNext = pArReq->pNext;	/* in case need to delete req */
	if (pArReq->disable) {
	    pArReq->status = AR_REQ_DISABLE;
	    DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
	}
	else if (pArReq->status == AR_REQ_CONN_NONE) {
	    stat = arReadReq(pArReq);
	    if (stat == AR_S_OK) {
		DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
		DoubleListAppend(pArReq, pglArCtx->pReqConnHead,
						    pglArCtx->pReqConnTail);
	    }
	    else {
		(void)sprintf(message,
			"errors while reading request: %s", pArReq->name);
		arLogMessage(" ", message, stat);
		arLogMessage(" ", "request being deleted", NULL);
		DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
	    }
	}
	else if (pArReq->status == AR_REQ_CONN_CA_BAD) {
	    DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
	    DoubleListAppend(pArReq, pglArCtx->pReqConnHead,
						    pglArCtx->pReqConnTail);
	}
	else if (pArReq->status == AR_REQ_CONN_OK ||
				pArReq->status == AR_REQ_CONN_CF_OK) {
	    stat = arReadReqWrapup(pArReq);
	    if (stat == AR_S_OK) {
		DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
		pArReq->readQ = 0;
		pArReq->mainQ = 1;
		DoubleListAppend(pArReq, pglArCtx->pReqMainHead,
						    pglArCtx->pReqMainTail);
	    }
	    else {
		(void)sprintf(message,
			"errors while enabling request: %s:", pArReq->name);
		arLogMessage(" ", message, NULL);
		arLogMessage(" ", "request being deleted", NULL);
		DoubleListRemove(pArReq, pglArCtx->pReqReadHead,
						    pglArCtx->pReqReadTail);
	    }
	}
	if (pArReq->status == AR_REQ_DISABLE) {
	    if ((stat = arReqDelete(pArReq)) != AR_S_OK) {
		(void)sprintf(message,
			"errors while disabling request: %s", pArReq->name);
		arLogMessage(" ", message, stat);
	    }
	}
	pArReq = pArReqNext;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arReadQPrintInfo - print information about read queue
*
* DESCRIPTION
*
* RETURNS
*
* BUGS
* o	text
*
* SEE ALSO
*
* EXAMPLE
*
*-*/
int
arReadQPrintInfo(abbrev)
int	abbrev;		/* I 0,1 for full,short print */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    AR_REQ	*pArReq;	/* pointer to next request */

    if (!abbrev)
	(void)arLogMessage(" ", "**** READ LIST ****", NULL);

    pArReq = pglArCtx->pReqReadHead;
    while (pArReq != NULL) {
	stat = arReqPrintInfo(pArReq, abbrev);
	if (stat != AR_S_OK)
	    retStat = stat;
	pArReq = pArReq->pNext;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReadReq - read an archive request file
*
* DESCRIPTION
*	An archive request file is read into an empty archive request
*	supplied by the caller.  The lines in the request file are parsed
*	and validated as they are read, and some cross-checks are made once
*	reading is complete.  Blank lines and those beginning with "#"
*	are ignored.
*
*	None of the channels in the request are opened by this routine, and,
*	consequently, no channel related validity checks are made by this
*	routine; such checks are done by arReadReqWrapup and
*	arMethBuildCheck.
*
*	None of the methods are queued on method lists.
*
*	If any errors are encountered, the request status will be set to
*	AR_REQ_UNDEF; otherwise, status will be AR_REQ_CONN_NONE.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_FOPENERR	if file couldn't be opened (an error message has
*			been printed)
*	AR_S_SYNERR	if a syntax error is encountered in the file (i.e.,
*			its contents and/or format violates the rules)  (an
*			error message is printed describing the specific
*			violation)
*	status from various other arXxx routines if an error is encountered
*			(an error message is printed mentioning channel name
*			for any channels which encounter errors)
*
* BUGS
* o	the file stream should be used in non-blocking mode for SunOS (as
*	long as light-weight processes are used in the implementation).
*	(Using the non-blocking I/O library doesn't work because of a conflict
*	with Channel Access.)
* o	in the header, event methods need to have the type of the comparison
*	value match the native type of the record--this isn't checked
*
* SEE ALSO
*	arReadReqWrapup, arMethBuild, arMethBuildCheck
*-*/
int
arReadReq(pArReq)
AR_REQ	*pArReq;	/* IO pointer to empty archive request */
{
    char	filePath[GEN_FNAME_DIM];/* file name, including path */
    FILE	*pStream;	/* stream pointer */

    short	act=0;		/* number of ACT: encountered */
    char	arg1[40];	/* first arg from line */
    char	channel[AR_NAME_DIM];/* channel name */
    int		i;		/* temporary index */
    short	inact=0;	/* number of INACT: encountered */
    short	inHead=1;	/* flag for 'in header section' */
    char	message[80];	/* temp for error messages */
    short	out=0;		/* number of OUTPUT: encountered */
    int		outType;	/* type of output */
    char	outTypeC[20];	/* text for type of output */
    chid	pCh;		/* channel table pointer */
    AR_EL	*pTempEl;	/* element pointer */
    AR_EL	*arElFind();	/* search request for named element */
    char	record[81];	/* record from file */
    char	recordCopy[81];	/* copy of record from file */
    int		retStat=AR_S_OK;/* return status to caller */
    short	samp=0;		/* number of SAMPLE: encountered */
    short	reqStat=0, sampStat=0; /* number of XXX_STATUS: encountered */
    int		stat;		/* status return from calls */
    char	*pTxt, *pFld, dlm;/* used by nextXxx routines */

    assert(pArReq != NULL);
    assert(strlen(pArReq->name) > 0);
    assert(strlen(pArReq->name) < AR_NAME_DIM);
    assert(strlen(AR_FILE_REQ_DIR) + AR_NAME_DIM + strlen(AR_FILE_REQ_EXT) <
							GEN_FNAME_DIM);

/*----------------------------------------------------------------------------
*    set up the path name to the file and open the file
*---------------------------------------------------------------------------*/
    (void)sprintf(filePath, "%s%s%s", AR_FILE_REQ_DIR, pArReq->name,
							AR_FILE_REQ_EXT);

    if ((pStream = fopen(filePath, "r")) == NULL) {
	(void)sprintf(message, "couldn't open: %s", filePath);
	(void)arLogMessage(" ", message, NULL);
	return AR_S_FOPENERR;
    }

/*----------------------------------------------------------------------------
*    read the file, ignoring blank lines.  A line containing
*    %endHeader% separates the header section of the file from the data
*    section.
*---------------------------------------------------------------------------*/
    (void)sprintf(message, "**** reading: %s ****", filePath);
    (void)arLogMessage(" ", message, NULL);
    while (inHead && fgets(record, sizeof(record), pStream) != NULL) {
	if ((i = strlen(record)) > 0) {
	    if (record[i-1] == '\n')
		record[i-1] = '\0';	/* remove \n before logging record */
	}
	(void)arLogMessage(" ", record, NULL);
	if (i <= 0 || record[0] == '#')
	    ;
	else if (strncmp(record,glArKwEndHeader,strlen(glArKwEndHeader)) == 0)
	    inHead = 0;
	else {
/*----------------------------------------------------------------------------
* HEADER SECTION
*
*    In the header section, attempt to match keywords from the file lines with
*    known keywords.  For valid keywords, parse the rest of the line and put
*    the resulting information into the request structure.
*---------------------------------------------------------------------------*/
	    strcpy(recordCopy, record);
	    pTxt = recordCopy;
	    if (nextNonSpaceField(&pTxt, &pFld, &dlm) <= 1)
		;	/* ignore blank lines */
	    else {
		if (strcmp(pFld, glArKwName) ==0)
/*	NAME:	*/
		    ;		/* no action */
		else if (strcmp(pFld, glArKwDesc) == 0)
/*	DESC:	*/
		    ;		/* no action */
		else if (strcmp(pFld, glArKwAct) == 0) {
/*	ACT:	*/
		    if (++act == 1) {
			stat = arMethBuild(&pArReq->act, record);
			if (stat != AR_S_OK)
			    retStat = stat;
		    }
		    else {
			(void)sprintf(message, "only one %s can appear",
								glArKwAct);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		}
		else if (strcmp(pFld, glArKwInact) == 0) {
/*	INACT:	*/
		    if (++inact == 1) {
			stat = arMethBuild(&pArReq->inact, record);
			if (stat != AR_S_OK)
			    retStat = stat;
		    }
		    else {
			(void)sprintf(message, "only one %s can appear",
								glArKwInact);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		}
		else if (strcmp(pFld, glArKwSamp) == 0) {
/*	SAMPLE:	*/
		    if (++samp == 1) {
			stat = arMethBuild(&pArReq->samp, record);
			if (stat != AR_S_OK)
			    retStat = stat;
		    }
		    else {
			(void)sprintf(message, "only one %s can appear",
								glArKwSamp);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		}
		else if (strcmp(pFld, glArKwReqStatus) == 0) {
/*	REQ_STATUS:	*/
		    if (++reqStat == 1) {
			assert(AR_NAME_DIM == 36);
			if (nextChanNameField(&pTxt, &pFld, &dlm) > 1) {
			    strcpy(pArReq->reqStChan.name, pFld);
			    pArReq->reqStChan.itemType = AR_ITEM_STAT_REQ;
			}
			else {
			    (void)sprintf(message, "%s must specify channel",
							glArKwReqStatus);
			    (void)arLogMessage(" ", message, NULL);
			    retStat = AR_S_SYNERR;
			}
		    }
		    else {
			(void)sprintf(message, "only one %s can appear",
							    glArKwReqStatus);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		}
		else if (strcmp(pFld, glArKwSampStatus) == 0) {
/*	SAMP_STATUS:	*/
		    if (++sampStat == 1) {
			assert(AR_NAME_DIM == 36);
			if (nextChanNameField(&pTxt, &pFld, &dlm) > 1) {
			    strcpy(pArReq->sampStChan.name, pFld);
			    pArReq->sampStChan.itemType = AR_ITEM_STAT_SAMP;
			}
			else {
			    (void)sprintf(message, "%s must specify channel",
							glArKwSampStatus);
			    (void)arLogMessage(" ", message, NULL);
			    retStat = AR_S_SYNERR;
			}
		    }
		    else {
			(void)sprintf(message, "only one %s can appear",
							    glArKwSampStatus);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		}
		else if (strcmp(pFld, glArKwOut) == 0) {
/*	OUTPUT:	*/
		    pArReq->output.ARType = AR_OUT_UNDEF;
		    if (++out > 1) {
			(void)sprintf(message, "only one %s can appear",
								glArKwOut);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		    assert(sizeof(outTypeC) == 20);
		    if (sscanf(record, "%*s %19s ", outTypeC) != 1) {
			(void)sprintf(message, "type is required for %s",
								glArKwOut);
			(void)arLogMessage(" ", message, NULL);
			retStat = AR_S_SYNERR;
		    }
		    else {
			for (outType=AR_OUT_PAST-1; outType>0; outType--) {
			    if (strcmp(glArOutTypeText[outType], outTypeC) == 0)
				break;
			}
			if (outType == 0) {
			    (void)sprintf(message, "illegal type for %s",
								glArKwOut);
			    (void)arLogMessage(" ", message, NULL);
			    retStat = AR_S_SYNERR;
			}
			else {
			    pArReq->output.ARType = outType;
			}
		    }
		}
		else {
/*	unrecognized	*/
		    (void)sprintf(message,
				"unrecognized keyword in header: %s", pFld);
		    (void)arLogMessage(" ", message, NULL);
		}
	    }
	}
    }
    assert(sizeof record == 81);
    while (fgets(record, 80, pStream) != NULL) {
	if ((i = strlen(record)) > 0) {
	    if (record[i-1] == '\n')
		record[i-1] = '\0';	/* remove \n before logging record */
	}
	if (i <= 0 || record[0] == '#')
	    (void)arLogMessage(" ", record, NULL);
	else {
	    assert(AR_NAME_DIM == 36);
	    while (sscanf(record, "%35s", channel) == 1) {
		strncpy(index(record, channel[0]),
		"                                        ", strlen(channel));
/*----------------------------------------------------------------------------
* CHANNEL SECTION
*
*    In the data section, create an element in the request for each channel.
*    Duplicate channels are rejected.
*---------------------------------------------------------------------------*/
		if (strlen(channel) > 0) {
		    if ((pTempEl = arElFind(pArReq, channel)) != NULL) {
			(void)sprintf(message,
				"%s ignored--already open", channel);
			(void)arLogMessage(" ", message, NULL);
		    }
		    else {
			stat = arElCreateAndAppend(pArReq, channel);
			if (stat != AR_S_OK)
			    retStat = stat;
		    }
		}
		else
		    break;
	    }
	}
    }
    (void)fclose(pStream);
    (void)sprintf(message, "**** EOF for %s ****", filePath);
    (void)arLogMessage(" ", message, NULL);

    if (inHead != 0) {		/* make sure end header was there */
	(void)sprintf(message, "%s not found", glArKwEndHeader);
	(void)arLogMessage(" ", message, NULL);
	retStat = AR_S_SYNERR;
    }

    if (pArReq->pElListHead == NULL) {
	(void)sprintf(message, "no data channels specified in request");
	(void)arLogMessage(" ", message, NULL);
	retStat = AR_S_SYNERR;
    }

/*----------------------------------------------------------------------------
*    do some initialization for the methods, over and above what was
*    done by arMethBuild.  (Detailed initialization is done by other
*    routines when a method is queued.)
*
*    also check for violations of restrictions on methods.  (Checks on
*    multiple specifications for a method were made above.)  Some checks are
*    deferred to arReadReqWrapup, following the completion of ca_search
*    for EVENT methods.
*---------------------------------------------------------------------------*/
    if (retStat == AR_S_OK) {
/*----------------------------------------------------------------------------
*	activation method
*---------------------------------------------------------------------------*/
	if (act == 0) {
	    (void)sprintf(message,
				"%s missing from request header", glArKwAct);
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	else {
	    if (  pArReq->act.type == AR_METH_AFTER    ||
		  pArReq->act.type == AR_METH_TIMED    ||
		  pArReq->act.type == AR_METH_NEVER    ||
		  pArReq->act.type == AR_METH_MONITOR) {
		(void)sprintf(message,
				"illegal method type for %s  %s", glArKwAct,
				ArMethTypeToText(pArReq->act.type));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	    if (  pArReq->act.type == AR_METH_EVENT &&
		       pArReq->act.ctrl.event.type == AR_EV_CHANGE) {
		(void)sprintf(message,
				"illegal event type for %s  %s", glArKwAct,
				ArEvTypeToText(pArReq->act.ctrl.event.type));
	        (void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	}

/*----------------------------------------------------------------------------
*	inactivation method
*---------------------------------------------------------------------------*/
	if (inact == 0) {
	    (void)sprintf(message, "%s missing from request header",
								glArKwInact);
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	else {
	    if (  pArReq->inact.type == AR_METH_MONITOR   ||
		  pArReq->inact.type == AR_METH_ALWAYS    ||
	          pArReq->inact.type == AR_METH_PERIODIC) {
		(void)sprintf(message, "illegal method type for %s  %s",
			    glArKwInact, ArMethTypeToText(pArReq->inact.type));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	    if (  pArReq->inact.type == AR_METH_EVENT &&
		       pArReq->inact.ctrl.event.type == AR_EV_CHANGE) {
		(void)sprintf(message, "illegal event type for %s  %s",
		    glArKwInact, ArEvTypeToText(pArReq->inact.ctrl.event.type));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	    if (  pArReq->inact.type == AR_METH_AFTER   &&
		  pArReq->samp.type == AR_METH_MONITOR) {
		(void)sprintf(message,
				"'%s  after' illegal with '%s  monitor'",
				glArKwInact, glArKwSamp);
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	    if (  pArReq->inact.type == AR_METH_TIMED   &&
	          pArReq->act.type == AR_METH_PERIODIC  ) {
		if (pArReq->act.ctrl.per.period.secPastEpoch <
		      pArReq->inact.ctrl.per.period.secPastEpoch+10) {
		    (void)sprintf(message, 
			"%s period must be at least 10 sec less than for %s ",
				glArKwInact, glArKwAct);
		    (void)arLogMessage(" ", message, NULL);
		    retStat = AR_S_SYNERR;
		}
	    }
	}

/*----------------------------------------------------------------------------
*	sample method and output specification
*---------------------------------------------------------------------------*/
	if (samp == 0) {
	    (void)sprintf(message, "%s missing from request header",
								glArKwSamp);
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	else {
	    if (  pArReq->samp.type == AR_METH_AFTER    ||
		  pArReq->samp.type == AR_METH_ALWAYS   ||
		  pArReq->samp.type == AR_METH_NEVER    ||
		  pArReq->samp.type == AR_METH_TIMED) {
		(void)sprintf(message, "illegal method type for %s  %s",
			    glArKwSamp, ArMethTypeToText(pArReq->samp.type));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	}

	if (out == 0) {
	    (void)sprintf(message, "%s missing from request header", glArKwOut);
	    (void)arLogMessage(" ", message, NULL);
	    retStat = AR_S_SYNERR;
	}
	else if (pArReq->output.ARType == AR_OUT_SET_TXT) {
	    if (pArReq->samp.type == AR_METH_MONITOR) {
		(void)sprintf(message, "illegal sample type for %s",
			    ArOutTypeToText(pArReq->output.ARType));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	}
	else if (pArReq->output.ARType == AR_OUT_CHAN_BIN) {
	    if (pArReq->samp.type != AR_METH_MONITOR) {
		(void)sprintf(message, "illegal sample type for %s",
			    ArOutTypeToText(pArReq->output.ARType));
		(void)arLogMessage(" ", message, NULL);
		retStat = AR_S_SYNERR;
	    }
	}
    }
    if (retStat == AR_S_OK)
	pArReq->status = AR_REQ_CONN_NONE;
    else
	pArReq->status = AR_REQ_UNDEF;

    if (retStat != AR_S_OK) {
	(void)sprintf(message, "deleting request: %s due to errors",
								pArReq->name);
	(void)arLogMessage(" ", message, NULL);
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arReadReqWrapup - validate archive request file following ca_search
*
* DESCRIPTION
*	This routine validates channel-related information in an archive
*	request and, if successful, moves the request onto the main archive
*	request list.  This routine is to be called after all channels in
*	the request are connected, as part of the `read a request file'
*	processing.
*
* RETURNS
*
* BUGS
* o	text
*
* SEE ALSO
*
* EXAMPLE
*
*-*/
int
arReadReqWrapup(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    char	message[80];	/* temp for error messages */
    AR_EL	*pArEl;		/* pointer to archive element */

    assert(pArReq != NULL);

/*----------------------------------------------------------------------------
*    check the various elements and perform wrapup
*---------------------------------------------------------------------------*/
    pArEl = pArReq->pElListHead;
    while (pArEl != NULL) {
	stat = arElCreateWrapup(&pArEl);
	if (stat != AR_S_OK)
	    retStat = stat;
	pArEl = pArEl->pNext;
    }

/*----------------------------------------------------------------------------
*    check event methods to be sure that none of the rules are violated by
*    the channels' characteristics
*---------------------------------------------------------------------------*/
    stat = arMethBuildCheck(&pArReq->act);
    if (stat != AR_S_OK) {
	(void)sprintf(message, "error on %s method", glArKwAct);
	(void)arLogMessage(" ", message, NULL);
	retStat = stat;
    }
    stat = arMethBuildCheck(&pArReq->inact);
    if (stat != AR_S_OK) {
	(void)sprintf(message, "error on %s method", glArKwInact);
	(void)arLogMessage(" ", message, NULL);
	retStat = stat;
    }
    stat = arMethBuildCheck(&pArReq->samp);
    if (stat != AR_S_OK) {
	(void)sprintf(message, "error on %s method", glArKwSamp);
	(void)arLogMessage(" ", message, NULL);
	retStat = stat;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqCreate - create an AR_REQ
*
* DESCRIPTION
*	Creates an archive request and initializes several fields:
*	o   status to AR_REQ_UNDEF
*	o   empty element list
*	o   activation, inactivation, and sample methods to AR_METH_UNDEF,
*	    with the appropriate item types for each and with a link in each
*	    to the archive request
*	o   output types (for AR and user) of AR_OUT_UNDEF
*
* RETURNS
*	AR_S_OK, or
*	AR_S_NOMEM	if no memory is available to hold new request
*
*-*/
int
arReqCreate(ppArReq, name)
AR_REQ	**ppArReq;	/* IO pointer to request pointer; if NULL, then
			no request pointer is returned to caller */
char	*name;		/* I pointer to request name */
{
    AR_REQ	*pNewReq=NULL;	/* pointer to new request created for pCh */
    int		retStat=AR_S_OK;/* return status to caller */

    assert(ppArReq != NULL);
    assert(name != NULL);

    if (strlen(name) > AR_NAME_DIM-1) {
	(void)printf("request name too long\n");
	retStat = AR_S_MAXNAMELEN;
    }
    else {
	pNewReq = (AR_REQ *)malloc(sizeof (AR_REQ));
	if (pNewReq == NULL) {
	    retStat = AR_S_NOMEM;
	}
	else {
	    strcpy(pNewReq->name, name);	/* initialize the request */
	    pNewReq->pNext            = NULL;
	    pNewReq->pPrev            = NULL;
	    pNewReq->status           = AR_REQ_UNDEF;
	    pNewReq->reqStChan.pReq   = pNewReq;
	    pNewReq->reqStChan.pCh    = NULL;
	    pNewReq->reqStChan.name[0]= '\0';
	    pNewReq->sampStChan.pReq  = pNewReq;
	    pNewReq->sampStChan.pCh   = NULL;
	    pNewReq->sampStChan.name[0]= '\0';
	    pNewReq->mainQ            = 0;
	    pNewReq->readQ            = 0;
	    pNewReq->disable          = 0;
	    pNewReq->busy             = 0;
	    pNewReq->sampleCount      = 0;
	    pNewReq->pElListHead      = NULL;
	    pNewReq->pElListTail      = NULL;
	    pNewReq->act.type         = AR_METH_UNDEF;
	    pNewReq->act.itemType     = AR_ITEM_METH_ACT;
	    pNewReq->act.actionType   = AR_METH_AC_UNDEF;
	    pNewReq->act.pReq         = pNewReq;
	    pNewReq->inact.type       = AR_METH_UNDEF;
	    pNewReq->inact.itemType   = AR_ITEM_METH_INACT;
	    pNewReq->inact.actionType = AR_METH_AC_UNDEF;
	    pNewReq->inact.pReq       = pNewReq;
	    pNewReq->samp.type        = AR_METH_UNDEF;
	    pNewReq->samp.itemType    = AR_ITEM_METH_SAMPLE;
	    pNewReq->samp.actionType  = AR_METH_AC_UNDEF;
	    pNewReq->samp.pReq        = pNewReq;
	    pNewReq->output.status    = AR_OUTS_IDLE;
	    pNewReq->output.ARType    = AR_OUT_UNDEF;
	    pNewReq->output.pARCFDesc = NULL;
	    pNewReq->output.usrType   = AR_OUT_UNDEF;
	    pNewReq->output.pUsrCFDesc= NULL;
	    pNewReq->output.usrPath[0]= '\0';
	}
    }

    if (ppArReq != NULL)
	*ppArReq = pNewReq;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqDelete - terminate request's activity and free its memory
*
* DESCRIPTION
*	Does a ca_clear_channel for each element in the request,
*	closes any archive files and/or subfiles associated with the
*	request, removes the request's methods from method queues, 
*	and free's any memory held by the request.
*
*	The request must not be on any list.
*
* RETURNS
*	AR_S_OK, or
*	status from various other routines
*
* BUGS
* o	error handling needs to be done
* o	doesn't close user archive stream
*
*-*/
int
arReqDelete(pArReq)
AR_REQ	*pArReq;	/* IO pointer to request */
{
    AR_EL	*pArEl;		/* element pointer */
    AR_EL	*pArElNext;	/* temporary element pointer */
    char	message[80];	/* temp for message */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArReq != NULL);
    assert(pArReq->pPrev == NULL);
    assert(pArReq->pNext == NULL);

    if (pArReq->reqStChan.pCh != NULL &&
		    ca_field_type(pArReq->reqStChan.pCh) != TYPENOTCONN) {
	ca_put(DBR_STRING, pArReq->reqStChan.pCh, "disabled");
	ca_flush_io();
    }

    stat = arConnBreak(pArReq);

/*----------------------------------------------------------------------------
* for each element:
*    o  free the AR_EL structure
*---------------------------------------------------------------------------*/
    pArEl = pArReq->pElListHead;
    while (pArEl != NULL) {
	pArElNext = pArEl->pNext;
	if (pArEl->pARChanDesc != NULL)
	    stat = arCFChanClose(&pArEl->pARChanDesc);
	free((char *)pArEl);
	pArEl = pArElNext;
    }
/*----------------------------------------------------------------------------
* for each method:
*    o  remove the method from its queue (if any)
*---------------------------------------------------------------------------*/
    stat = arMethDel(&pArReq->act);
    stat = arMethDel(&pArReq->inact);
    stat = arMethDel(&pArReq->samp);

    free((char *)pArReq);

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqDisable - disable a request
*
* DESCRIPTION
*	Sets the request's disable flag.  The actual disable will be done
*	by one of the arXxxxQCheck routines.
*
* RETURNS
*	AR_S_OK, or
*	status from various other routines
*
*-*/
int
arReqDisable(pArReq)
AR_REQ	*pArReq;	/* IO pointer to request */
{
    int		retStat=AR_S_OK;/* return status to caller */

    assert(pArReq != NULL);

    pArReq->disable = 1;

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqFind - search for request in archive list
*
* DESCRIPTION
*	The requests on the archive request list are searched,
*	looking for one whose name matches the specified name.
*
* RETURNS
*	NULL, or
*	a pointer (AR_REQ *) to the desired request.
*
*-*/
AR_REQ *
arReqFind(name)
char	*name;		/* I request name, null terminated string */
{
    AR_REQ	*pReq;	/* request pointer */

    assert(name != NULL);

    pReq = pglArCtx->pReqMainHead;
    while (pReq != NULL) {
	if (strcmp(name, pReq->name) == 0)
	    return pReq;
	else
	    pReq = pReq->pNext;
    }
    pReq = pglArCtx->pReqConnHead;
    while (pReq != NULL) {
	if (strcmp(name, pReq->name) == 0)
	    return pReq;
	else
	    pReq = pReq->pNext;
    }
    pReq = pglArCtx->pReqReadHead;
    while (pReq != NULL) {
	if (strcmp(name, pReq->name) == 0)
	    return pReq;
	else
	    pReq = pReq->pNext;
    }
    return NULL;
}

/*+/subr***********************************************************************
* NAME	arReqGetValues - issue a ca_get for each element in a request
*
* DESCRIPTION
*	Queues up requests to get the values for the channels associated
*	with an archive request.  The `elPendCount' for the request is
*	incremented by one for each element.  The status for each element
*	will be set to AR_EL_GETPEND.
*
*	New values won't be posted to the elements until ca_pend_event
*	has been called.  When values have been received for all elements,
*	`elPendCount' will finally be zero.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_OPSPEND if `elPendCount' isn't zero at entry, or
*	status code from arElGetValue (an error message is printed with
*			channel name for any element for which an error
*			is encountered)
*-*/
int
arReqGetValues(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    AR_EL	*pEl;		/* pointer to element */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArReq != NULL);
    if (pArReq->elPendCount != 0)
	return AR_S_OPSPEND;

    pEl = pArReq->pElListHead;
    while (pEl != NULL) {
	if (ca_element_count(pEl->pCh) > 1)
	    stat = arElGetNativeValue(pEl);
	else
	    stat = arElGetValue(pEl);
	if (stat != AR_S_OK) {
	    retStat = stat;
	    (void)arElLogStatus(pEl, stat);
	}
	pEl = pEl->pNext;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqPrintInfo - print information about a request
*
* DESCRIPTION
*	Prints information about a request, its methods, and its elements.  If
*	an abbreviated set of information is desired, a single line of
*	information is displayed.
*
* RETURNS
*	AR_S_OK, or
*	status code from other routines (an error message is printed
*			as appropriate)
*
*-*/
int
arReqPrintInfo(pArReq, abbrev)
AR_REQ   *pArReq;       /* I pointer to archive request */
int	abbrev;		/* I 0,1 for full,short print */
{
    char	message[80];	/* temp for log messages */
    AR_EL	*pEl;		/* pointer to element */
    int         retStat=AR_S_OK;/* return status to caller */
    int         stat;           /* status return from calls */
    char	flagC[10];	/* text for .flags */
    char	*conn;		/* connection state for status channel */

    assert(pArReq != NULL);

    if (abbrev) {
	(void)sprintf(message, "%s %s",
		pArReq->name,
		ArReqStatusToText(pArReq->status)
		);
	(void)arLogMessage(" ", message, NULL);
	return retStat;
    }
    /* general info about the request */
    (void)sprintf(message, "---- %s ----", pArReq->name);
    (void)arLogMessage(" ", message, NULL);
    if (pArReq->pNext == NULL)
	(void)sprintf(message, "next: NULL");
    else
	(void)sprintf(message, "next: %s", pArReq->pNext->name);
    (void)arLogMessage(" ", message, NULL);
    if (pArReq->pPrev == NULL)
	(void)sprintf(message, "prev: NULL");
    else
	(void)sprintf(message, "prev: %s", pArReq->pPrev->name);
    (void)arLogMessage(" ", message, NULL);
    if (pArReq->mainQ)
	(void)strcpy(flagC, "mainQ");
    else if (pArReq->readQ)
	(void)strcpy(flagC, "readQ");
    else if (pArReq->disable)
	(void)strcpy(flagC, "disable");
    else
	flagC[0] = '\0';
    (void)sprintf(message, "status: %s  flags: %s",
				ArReqStatusToText(pArReq->status), flagC);
    (void)arLogMessage(" ", message, NULL);
    (void)sprintf(message, "connections pending: %d", pArReq->connPendCount);
    (void)arLogMessage(" ", message, NULL);
    (void)sprintf(message, "element operations pending: %d",
							pArReq->elPendCount);
    (void)arLogMessage(" ", message, NULL);
    /* information about the request's methods */
    arMethPrintInfo(&pArReq->act, "ACT:");
    arMethPrintInfo(&pArReq->inact, "INACT:");
    arMethPrintInfo(&pArReq->samp, "SAMPLE:");
    (void)sprintf(message, "output status %s",
				ArOutStatToText(pArReq->output.status));
    (void)arLogMessage(" ", message, NULL);
    (void)sprintf(message, "AR output: %s",
				ArOutTypeToText(pArReq->output.ARType));
    (void)arLogMessage(" ", message, NULL);
    (void)sprintf(message, "user output: %s to %s",
		ArOutTypeToText(pArReq->output.usrType),
		pArReq->output.usrPath);
    (void)arLogMessage(" ", message, NULL);
    if (pArReq->reqStChan.name[0] != '\0') {
	if (pArReq->reqStChan.pCh == NULL)
	    conn = "no pCh";
	else if (ca_field_type(pArReq->reqStChan.pCh) == TYPENOTCONN)
	    conn = "not conn";
	else
	    conn = "connected";
	(void)sprintf(message, "request status channel: %s; chan is %s",
		pArReq->reqStChan.name, conn);
	(void)arLogMessage(" ", message, NULL);
    }
    /* information about the elements */
    pEl = pArReq->pElListHead;
    while (pEl != NULL) {
        (void)arElPrintInfo(pEl);
        pEl = pEl->pNext;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arReqPrintValues - print the value for each element in a request
*
* DESCRIPTION
*	For each element in an archive request, print the value.
*
* RETURNS
*	AR_S_OK or
*	status code from arElPrintValue (an error message is printed with
*			channel name for any element for which an error
*			is encountered)
*
*-*/
int
arReqPrintValues(pArReq)
AR_REQ	*pArReq;	/* I pointer to archive request */
{
    AR_EL	*pEl;		/* pointer to element */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    assert(pArReq != NULL);

    pEl = pArReq->pElListHead;
    while (pEl != NULL) {
	(void)arLogMessage(" ", pEl->name, NULL);
	stat = arElPrintValue(pEl);
	if (stat != AR_S_OK) {
	    retStat = stat;
	    (void)arElLogStatus(pEl, stat);
	}
	pEl = pEl->pNext;
    }

    return retStat;
}
