/*	@(#)arLibSAMP.c	1.3 2/6/92
 *	Author:	Roger A. Cole
 *	Date:	12-06-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991-92, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	12-06-89	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03	10-21-91	rac	split into smaller files; handle REQ_STATUS:
 *				and SAMP_STATUS: commands
 * .04	12-05-91	rac	don't use LWP; allow choosing MDEL or ADEL
 *				for monitor deadband
 * .05	02-04-92	rac	properly write SHORT waveforms in SSF
 *
 * make options
 *	-DNDEBUG	don't compile assert checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	arLibSAMP - routines for sampling and sample handling
*
* DESCRIPTION
*
*-***************************************************************************/

#define USE_MDEL 0	/* 1 to use MDEL, 0 to use ADEL for monitoring */

#include <stdio.h>
#include <sys/file.h>	/* O_RDWR definition */

#include <tsDefs.h>
#include <arCS.h>
#include <arDefs.h>
#include <arAccessLib.h>

/*+/subr***********************************************************************
* NAME	arSampleCheck - check main archive list for needed sample activities
*
* DESCRIPTION
*	This routine handles sampling activities for archive requests.  What
*	it does depends on request state and output status.  It processes
*	only requests which are AR_REQ_ACTIVE and which don't have .disable
*	set.  The processing for the various output status are:
*
*	AR_OUTS_SAMPLE
*	    o   queues gets using DBR_TIME_STRING for all the elements,
*		setting .elPendCount to the number of pending gets
*	    o   sets the .busy flag in the request.  (This flag will be
*		reset when the sample is actually written to file.)
*
*	AR_OUTS_GETPEND
*	    o   checks for .elPendCount == 0   .  If at 0, change output
*		status to AR_OUTS_GETDONE.
*
*	Note that this routine only TAKES samples, but has no involvement
*	in the writing of completed samples to files.  arSampleSetCheck
*	provides the function of writing sample set files.
*
* RETURNS
*	AR_S_OK, or
*	status from various other arXxx routines if an error is encountered
*			(an error message mentioning channel name is printed
*			for any channels for which errors are detected)
*
* BUGS
* o	detecting a request which needs sample service should probably
*	be done in arMainQCheck, with the activity serviced here for
*	individual requests as needed
* o	check for empty list
* o	check element status (??)
* o	stamp in file should be from first element's time stamp (???)
* o	for .VAL fields, the .EGU field should also be stored in the file
*
*-*/
int
arSampleCheck()
{
    char	message[80];	/* temp for error messages */
    chid	pCh;		/* channel table pointer */
    AR_REQ	*pArReq;	/* archive request pointer */
    AR_EL	*pTempEl;	/* element pointer */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */

    pArReq = pglArCtx->pReqMainHead;
    while (pArReq != NULL) {
	if (pArReq->status != AR_REQ_ACTIVE || pArReq->disable)
	    ;		/* no action */
	else if (pArReq->output.ARType == AR_OUT_SET_TXT ||
	         pArReq->output.usrType == AR_OUT_SET_TXT) {
/*----------------------------------------------------------------------------
* sample set
*---------------------------------------------------------------------------*/
	    if (pArReq->output.status == AR_OUTS_SAMPLE) {
/*----------------------------------------------------------------------------
*	SAMPLE
*	    o  queue up get requests for all elements
*	    o  mark the request as `busy'
*---------------------------------------------------------------------------*/
		pArReq->busy = 1;
		stat = arReqGetValues(pArReq);	/* queue the gets */
		if (stat == AR_S_OK)
		    pArReq->output.status = AR_OUTS_GETPEND;
		else {
		    pArReq->output.status = AR_OUTS_IDLE;
		    retStat = stat;
		    (void)arLogMessage(" ",
				"error queueing gets for sample set", NULL);
		    (void)sprintf(message, "sample skipped for: %s",
								pArReq->name);
		    (void)arLogMessage(" ", message, NULL);
		}
	    }
	    else if (pArReq->output.status == AR_OUTS_GETPEND) {
/*----------------------------------------------------------------------------
*	GETPEND
*---------------------------------------------------------------------------*/
		if (pArReq->elPendCount == 0)
		    pArReq->output.status = AR_OUTS_GETDONE;
	    }
	}
	pArReq = pArReq->pNext;
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arSampleChanAddMon - issue ca_add_event for each element in req.
*
* DESCRIPTION
*	For each element in an archive request, establishes a monitor.
*
* RETURNS
*	AR_S_OK, or
*	status code from arElMonitorValue (an error message is printed
*			with channel name for any element for which an
*			error is encountered)
*
*-*/
int
arSampleChanAddMon(pArReq)
AR_REQ   *pArReq;       /* IO pointer to archive request */
{
    AR_EL	*pEl;		/* pointer to element */
    int         retStat=AR_S_OK;/* return status to caller */
    int         stat;           /* status return from calls */
    void arSampleChanHandler();

    assert(pArReq != NULL);
    assert(pArReq->output.ARType == AR_OUT_CHAN_BIN ||
				pArReq->output.usrType == AR_OUT_CHAN_BIN);

    pEl = pArReq->pElListHead;
    while (pEl != NULL) {
	assert(pEl->pCh != NULL);
	if (pArReq->output.ARType == AR_OUT_CHAN_BIN) {
	    stat = ca_add_masked_array_event(
			dbf_type_to_DBR_TIME(ca_field_type(pEl->pCh)),
			ca_element_count(pEl->pCh), pEl->pCh,
			arSampleChanHandler, pEl, 0., 0., 0., &pEl->pEv,
#if USE_MDEL
			DBE_VALUE | DBE_ALARM);
#else
			DBE_LOG | DBE_ALARM);
#endif
	    if (stat != ECA_NORMAL) {
		(void)arCaStatusMessage(stat,
				"error on ca_add_event", pEl->pCh);
		retStat = AR_S_CAERR;
	    }
	    else
		pEl->status = AR_EL_MONITOR;
	}
        pEl = pEl->pNext;
    }

    return retStat;
}

/*+/subr**********************************************************************
* NAME	arSampleChanHandler - receive monitor values for `by channel' samples
*
* DESCRIPTION
*	If the status of the archive request isn't AR_REQ_ACTIVE, or if the
*	element status is other than indicated above, the value is discarded.
*
* RETURNS
*	void
*
* BUGS
* o	doesn't do any validity checking on element
* o	doesn't check status on write
*
*-*/
void
arSampleChanHandler(arg)
struct event_handler_args arg;	/* I handler arg */
{
    AR_EL	*pArEl;		/* I element pointer */
    void	*pBuf;		/* IO pointer to value buffer */
    char	message[80];	/* temp buffer for error messages */

    pArEl = arg.usr;
    assert(pArEl != NULL);
    assert(pArEl->itemType == AR_ITEM_EL);
    assert(dbr_type_is_TIME(arg.type));
    pBuf = arg.dbr;

    if (pArEl->status != AR_EL_MONITOR ||
    					pArEl->pReq->status != AR_REQ_ACTIVE) {
#if 0
	(void)sprintf(message, "error in arSampleChanHandler for %s",
								pArEl->name);
	(void)arLogMessage(" ", message, NULL);
	(void)sprintf(message, "in request: %s", pArEl->pReq->name);
	(void)arLogMessage(" ", message, NULL);
	if (pArEl->pReq->status != AR_REQ_ACTIVE) {
	    (void)sprintf(message, "bad request status: %s",
					ArReqStatusToText(pArEl->pReq->status));
	    (void)arLogMessage(" ", message, NULL);
	}
	else {
	    (void)sprintf(message, "bad element status: %s",
					ArElStatusToText(pArEl->status));
	    (void)arLogMessage(" ", message, NULL);
	}
	(void)arLogMessage(" ", "discarding value", NULL);
#endif
	return;
    }
    else {
	if (((struct dbr_time_string *)pBuf)->stamp.secPastEpoch == 0)
	    (void)tsLocalTime(&((struct dbr_time_string *)pBuf)->stamp);
	if (pArEl->pARChanDesc != NULL)
	    arCFChanWrite(pArEl->pARChanDesc, &arg);
	if (pArEl->pUsrChanDesc != NULL)
	    arCFChanWrite(pArEl->pUsrChanDesc, &arg);
    }

    return;
}

/*+/subr**********************************************************************
* NAME	arSampleSetCheck - check for need for a sample set write
*
* DESCRIPTION
*	For ACTIVE requests whose output status is AR_OUTS_GETDONE, write
*	a sample set to the file.  The output status is changed to
*	AR_OUTS_FILING while the file is being written, and then changed
*	to AR_OUTS_IDLE when the operation is complete.  At the time of
*	completion, the .sampleCount in the request is incremented.
*	Once the operation is complete, the .busy flag in the request is
*	reset by this routine.
*
*	This routine checks the entire main request list.
*
* RETURNS
*	AR_S_OK, or
*	other status codes if an error is encountered
*
* BUGS
* o	maybe an error message should be printed if an error is detected
*
*-*/
int
arSampleSetCheck(pQuitFlag)
int	*pQuitFlag;	/* non-zero says to quit after current sample set */
{
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;           /* status return from calls */
    AR_REQ	*pArReq;	/* request pointer */

    assert(pQuitFlag != NULL);

    pArReq = pglArCtx->pReqMainHead;
    while (pArReq != NULL && *pQuitFlag == 0) {
	if (pArReq->status == AR_REQ_ACTIVE &&
	    pArReq->output.status == AR_OUTS_GETDONE) {
	    pArReq->output.status = AR_OUTS_FILING;
	    stat = arSampleSetWrite(pArReq); 
	    if (stat != AR_S_OK) {
		retStat = stat;
	    }
	    pArReq->output.status = AR_OUTS_IDLE;
	    if (++pArReq->sampleCount >= 1000000000)
		pArReq->sampleCount = 1;
	    pArReq->busy = 0;
	}
	pArReq = pArReq->pNext; 
    }

    return retStat;
}

/*+/subr***********************************************************************
* NAME	arSampleSetWrite - write a sample set for an archive request
*
* DESCRIPTION
*	Write a sample set to a sample set file.  The path for the file
*	is determined by the request name and by the parameters
*	AR_FILE_SET_DIR (to establish the directory) and AR_FILE_SET_TXT_EXT
*	(to establish the extension).
*
*	The time stamp for the sample is taken from the request's sample
*	method, which is based on the workstation's notion of current time.
*	(This is done for various reasons.  1) a single time stamp is
*	desired for the sample; 2) if the first channel isn't changing,
*	then it's time stamp won't be changing, either; 3) it is possible
*	to get zero time stamps back with data following an IOC reboot.)
*
*	If the file already exists, the sample set is appended to the end
*	of the file.  If the file doesn't exist, it is created.
*
*	The output status of the request must be AR_OUTS_FILING at
*	entry.
*
*	If the request status isn't AR_REQ_ACTIVE, then no sample is taken
*	and AR_S_OK is returned, with no error message being printed.
*
* RETURNS
*	AR_S_OK, or
*	AR_S_FOPENERR	if file couldn't be opened (an error message has
*			been printed)
*	status from various other arXxx routines if an error is encountered
*			(an error message mentioning channel name is printed
*			for any channels for which errors are detected)
*
* BUGS
* o	only DBR_TIME_STRING and DBR_TIME_SHORT are handled
* o	check for empty list
* o	for .VAL fields, the .EGU field should also be stored in the file
* o	non-blocking I/O should be used on Sun
*
*-*/
int
arSampleSetWrite(pArReq)
AR_REQ	*pArReq;	/* IO pointer to archive request */
{
    char	filePath[80];	/* file name, including path */
    FILE	*pStream;	/* stream pointer */
    char	causeBuf[300];	/* buffer for explanation of why sample is
				being taken */

    char	channel[AR_NAME_DIM];/* channel name */
    short	inData=0;	/* flag for 'in data section' */
    char	message[80];	/* temp for error messages */
    chid	pCh;		/* channel table pointer */
    AR_EL	*pTempEl;	/* element pointer */
    char	record[81];	/* record from file */
    int		retStat=AR_S_OK;/* return status to caller */
    int		stat;		/* status return from calls */
    int		i;		/* temp for loop index */
    char	*pStrVal;	/* pointer to string value */
    short	*pShrtVal;	/* pointer to short value */
    char	timeText[28];

    assert(pArReq != NULL);
    assert(strlen(pArReq->name) > 0);
    assert(pArReq->output.status == AR_OUTS_FILING);

    (void)sprintf(filePath, "%s%s%s", AR_FILE_SET_DIR, pArReq->name,
						AR_FILE_SET_TXT_EXT);

/*----------------------------------------------------------------------------
*    open the file for append and write the snapshot
*---------------------------------------------------------------------------*/
    if (pArReq->status != AR_REQ_ACTIVE)
	;	/* no action */
    else if ((pStream = fopen(filePath, "a")) == NULL) {
	(void)printf("couldn't open: %s\n", filePath);
	retStat = AR_S_FOPENERR;
    }
    else {
	(void)fprintf(pStream, "REQUEST: %s%s\n",
			pArReq->name, AR_FILE_REQ_EXT);
	(void)arMethGetInfoAsText("SAMPLE:", &pArReq->samp,
			causeBuf, sizeof causeBuf);
	(void)fprintf(pStream, "%s", causeBuf);
	(void)fprintf(pStream, "%s\n", glArKwEndHeader);
	pTempEl = pArReq->pElListHead;
	while (pTempEl != NULL) {
	    (void)fprintf(pStream, "%s %s\t%d %d\t%d",
			ca_name(pTempEl->pCh),
			dbf_type_to_text(ca_field_type(pTempEl->pCh)),
			pTempEl->dbrVal.tstrval.severity,
			pTempEl->dbrVal.tstrval.status,
			ca_element_count(pTempEl->pCh));
	    assert(dbr_type_is_TIME(pTempEl->dbrType));
	    if (dbr_type_is_STRING(pTempEl->dbrType)) {
		pStrVal = pTempEl->dbrVal.tstrval.value;
		for (i=0; i<ca_element_count(pTempEl->pCh); i++) {
		    if (ca_field_type(pTempEl->pCh) == DBF_STRING)
			(void)fprintf(pStream, " %d", strlen(pStrVal));
		    (void)fprintf(pStream, " %s", pStrVal);
		    pStrVal += db_strval_dim;
		}
	    }
	    else if (dbr_type_is_SHORT(pTempEl->dbrType)) {
		pShrtVal = &pTempEl->dbrVal.tshrtval.value;
		for (i=0; i<ca_element_count(pTempEl->pCh); i++) {
		    (void)fprintf(pStream, " %d", *pShrtVal);
		    pShrtVal += 1;
		}
	    }
	    else
		assertAlways(0);
	    (void)fprintf(pStream,"\n");
	    pTempEl = pTempEl->pNext;
	}
	(void)fprintf(pStream, "%s\n", glArKwEndData);
	fclose(pStream);
	if (pArReq->sampStChan.pCh != NULL &&
		    ca_field_type(pArReq->sampStChan.pCh) != TYPENOTCONN) {
	    tsStampToText(&pArReq->pElListHead->dbrVal.tstrval.stamp,
			TS_TEXT_MMDDYY, timeText);
	    timeText[17] = '\0';
	    ca_put(DBR_STRING, pArReq->sampStChan.pCh, timeText);
	    ca_flush_io();
	}
    }

    if (retStat != AR_S_OK)
	(void)printf("aborting snapshot: errors encountered\n");

    return retStat;
}
