/*	@(#)reqSubr.c	1.5 11/17/92
 *	Author:	Roger A. Cole
 *	Date:	03-08-92
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1992, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 *  .01 03-08-92 rac	initial version
 *  .02 03-15-92 rac	continued development
 *  .03 08-28-92 rac	add new archive request features: `step a channel',
 *			`allow disconnects',
 *  .04 09-30-92 rac	print stamps for interval test
 *  .05 11-17-92 rac	allow printing to status channels even if they
 *			aren't connected
 * reqSubr.c,v
 * Revision 1.2  1994/05/06 19:06:15  mrk
 * conversion from sccs to cvs
 *
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	reqSubr.c - handle archive requests
*
* DESCRIPTION
*
* QUICK REFERENCE
* #include <reqSubr.h>
*
*   RAW_REQ *rawReqAlloc()
*      void  rawReqCheckConnect(pRaw, messageFn, pArg, delaySec)
*      void  rawReqCheckInterval(pRaw, messageFn, pArg, delaySec,
*                                         intervalSec, jitterSec, timeSec)
*      void  rawReqCheckSemantics(pRaw, messageFn, pArg)
*      void  rawReqFree(pRaw)
*      void  rawReqParseLine(pRaw, line, messageFn, pArg)
*   RAW_REQ *rawReqReadFile(dir, file, messageFn, pArg)
*      void  reqTrigArm(pTrig, nowTs)
*       int  reqTrigTest(pTrig, nowTs)
*
* If messageFn is NULL, then messages are written to stdout.  If present:
*              void messageFn(message, pArg)
*
* FILE SYNTAX
* # comment
* NAME:		reqName
* DESC:		description, for display on user interface
* ACT:		activation trigger:
*		always
*		allow disconnects
*		demand
*		every hh:mm:ss
*		if chanName condition
*		when chanName condition
* INACT:	inactivation trigger:
*		after hh:mm:ss
*		demand
*		if chanName condition
*		never
*		sample complete
*		when chanName condition
* TRIGGER:	sample trigger:
*		after each step
*		always
*		at least every hh:mm:ss
*		demand
*		every hh:mm:ss
*		if chanName condition
*		when chanName condition [every n times]
* MONITOR:	sample by monitoring channels:
*		forever
*		for hh:mm:ss
*		round stamps n msec
*		until chanName condition
* SNAPSHOT:	sample with synchronous snapshots:
*		allow partial samples
*		every {hh:mm:ss, n samples}
*		for n samples
*		for hh:mm:ss, n samples maximum
*		round stamps n msec
*		start {hh:mm:ss, n samples} {before, after} trigger
*				("after trigger" not yet implemented)
*		until chanName condition, n samples maximum
* STEP:		step a channel through a sequence of values:
*		chanName from startVal to endVal by stepSize
* OUTPUT:	output specifications, one of:
*		chan:bin [fileName] n Mbytes
*		set:txt [fileName]
* SAMP_STATUS:	chanName
* REQ_STATUS:	chanName
* %endHeader%
* [F] chanName  [F] chanName
* [F] chanName
*
* condition can be:
*	chanName cond value	or
*	chanName change
*
*	where `cond' is one of:  ==  !=  <  <=  >  >=
*
* BUGS
* o	"after trigger" not yet implemented
* o	the `start before' time is based on data time stamps, while the
*	`for hh:mm:ss' time is based on UNIX time
*
* WISH LIST
* o	allow specifying whether activate requires all channels to be
*	connected
* o	provide a "grab" sample type, which would use ca_get
*
*-***************************************************************************/
#include <stdio.h>
#include <varargs.h>
#include <string.h>
#include <db_access.h>
#include <cadef.h>
#include <genDefs.h>
#include <tsDefs.h>

#define REQ_PRIVATE
#include <reqSubr.h>


#ifdef REQ_TEST
main()
{
    RAW_REQ	*pRaw;
    char	line[120];

    pRaw = rawReqAlloc();
    while (fgets(line, sizeof(line), stdin) != NULL) {
	rawReqParseLine(pRaw, line, NULL);
    }
    rawReqCheckConnect(pRaw, NULL);
    rawReqFree(pRaw);
}
#endif

/*-----------------------------------------------------------------------------
* global
*----------------------------------------------------------------------------*/
static int errorCount;
static char *line_u;
static void (*msgFn_u)();
static void *pArg_u;
static RAW_REQ *pRaw_u;
static char *pLine, *tempLine, *pField, dlm;
static int L;
static char message[400];


/*+/subr**********************************************************************
* NAME	rawParseXxx - individual parsing routines
*
*-*/
#define RAW_NEXT_FLD	\
	(L = nextNonSpaceField(&pLine, &pField, &dlm))
/*-----------------------------------------------------------------------------
*    RAW_ERROR("...channel name expected\n", pArg_u);
*----------------------------------------------------------------------------*/
#define RAW_ERROR \
	if (errorCount++ == 0) msgFn_u(line_u, pArg_u); msgFn_u

#define RAW_EOL_CHECK \
    if (dlm != '\n' && dlm != '\0') { \
	RAW_ERROR("...end of line expected\n", pArg_u); \
    }

/*+/internal******************************************************************
* NAME	rawParseEvent
*
* BUGS
* o	doesn't allow 'every n times' with 'change' (and doesn't enforce
*	no 'value' following 'change')
*
*-*/
void rawParseEvent_1();

static void rawParseEvent(pTrig)
REQ_TRIG *pTrig;
{
    rawParseEvent_1(pTrig, 1, 1);
}
static void rawParseIf(pTrig)
REQ_TRIG *pTrig;
{
    rawParseEvent_1(pTrig, 0, 0);
}
static void rawParseWhile(pTrig)
REQ_TRIG *pTrig;
{
    rawParseEvent_1(pTrig, 0, 1);
}
static void rawParseEvent_1(pTrig, allowEvery, allowChange)
REQ_TRIG *pTrig;
int	allowEvery, allowChange;
{
    int		i;

    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...channel name expected\n", pArg_u);
	return;
    }
    if (strlen(pField) >= db_name_dim) {
	RAW_ERROR("...channel name too long\n", pArg_u);
	return;
    }
    strcpy(pTrig->eventName, pField);
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...event condition expected\n", pArg_u);
	return;
    }
    for (i=1; i<REQ_EV_PAST; i++) {
	if (strcmp(pField, glReqEventText[i]) == 0)
	    break;
    }
    if (i >= REQ_EV_PAST) {
	RAW_ERROR("...illegal event condition\n", pArg_u);
	return;
    }
    pTrig->eventCond = i;
    if (i == REQ_EV_CHANGE) {
	if (allowChange)
	    return;
	RAW_ERROR("...`change' not allowed\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...comparison value expected\n", pArg_u);
	return;
    }
    strcpy(pTrig->eventValCmp.strVal, pField);
    pTrig->everyComp = 1;
    if (RAW_NEXT_FLD <= 1 || *pField == '#')
	return;
    if (strcmp("every", pField) != 0) {
	RAW_ERROR("...`every' keyword expected\n", pArg_u);
	return;
    }
    if (!allowEvery) {
	RAW_ERROR("...`every' not allowed\n", pArg_u);
	return;
    }
    if (rawParseGetInt(&pTrig->everyComp, 1, 1000) != 0) {
	RAW_ERROR("...illegal `every' count\n", pArg_u);
	return;
    }
    if (rawParseStrcmp("times", 1) != 0) {
	RAW_ERROR("...`times' keyword expected\n", pArg_u);
	return;
    }
}

/*+/internal******************************************************************
* NAME	rawParseGetDbl - get next field as an double
*
* RETURNS
*	0	if a valid value within the specified range is found, or
*	-1
*
*-*/
static double
rawParseGetDbl(pDbl, minVal, maxVal)
double	*pDbl, minVal, maxVal;
{
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...double value expected\n", pArg_u);
	return -1;
    }
    if (sscanf(pField, "%lf", pDbl) != 1) {
	RAW_ERROR("...double value expected\n", pArg_u);
	return -1;
    }
    if (minVal == maxVal)
	;	/* don't do a range check */
    else if (*pDbl < minVal || *pDbl > maxVal) {
	sprintf(message, "...value must be between %f and %f\n", minVal, maxVal);
	RAW_ERROR(message, pArg_u);
	return -1;
    }
    return 0;
}

/*+/internal******************************************************************
* NAME	rawParseGetInt - get next field as an integer
*
* RETURNS
*	0	if a valid value within the specified range is found, or
*	-1
*
*-*/
static int
rawParseGetInt(pInt, minVal, maxVal)
int	*pInt, minVal, maxVal;
{
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...integer value expected\n", pArg_u);
	return -1;
    }
    if (sscanf(pField, "%d", pInt) != 1) {
	RAW_ERROR("...integer value expected\n", pArg_u);
	return -1;
    }
    if (minVal == maxVal)
	;	/* don't do a range check */
    else if (*pInt < minVal || *pInt > maxVal) {
	sprintf(message, "...value must be between %d and %d\n", minVal, maxVal);
	RAW_ERROR(message, pArg_u);
	return -1;
    }
    return 0;
}

/*+/internal******************************************************************
* NAME	rawParseStampOrCount - handle xxx {hh:mm:ss, n samples}
*
* DESCRIPTION
*	Scans forward on the line looking for either a count or a time.
*
*	If a count is found, type is set to REQ_TYPE_SAMPLES.
*
*	If a time is found, type is set to REQ_TYPE_TIMED.
*
*-*/
static void rawParseStampOrCount(key, pType, pCount, pStamp)
char	*key;
REQ_TYPE *pType;
int	pCount;
TS_STAMP *pStamp;
{
    if (RAW_NEXT_FLD <= 1) {
	sprintf(message, "...%s parameters expected\n", key);
	RAW_ERROR(message, pArg_u);
	return;
    }
    if (strchr(pField, ':') != NULL) {
	if (tsTimeTextToStamp(pStamp, &pField) != S_ts_OK) {
	    RAW_ERROR("...invalid time\n", pArg_u);
	    return;
	}
	*pType = REQ_TYPE_TIMED;
    }
    else {
	if (sscanf(pField, "%d", pCount) != 1) {
	    RAW_ERROR("...invalid number of samples\n", pArg_u);
	    return;
	}
	if (rawParseStrcmp("samples", 1, 1000) != 0) 
	    return;
	*pType = REQ_TYPE_SAMPLES;
    }
}

/*+/internal******************************************************************
* NAME	rawParseStamp
*
*-*/
static void
rawParseStamp(pStamp)
TS_STAMP *pStamp;
{
    if (RAW_NEXT_FLD <= 1 || strchr(pField, ':') == NULL) {
	RAW_ERROR("...hh:mm:ss expected\n", pArg_u);
	return;
    }
    if (tsTimeTextToStamp(pStamp, &pField) != S_ts_OK) {
	RAW_ERROR("...invalid hh:mm:ss\n", pArg_u);
	return;
    }
#if 0
    if (pStamp->secPastEpoch < 10) {
	RAW_ERROR("...time can't be less than 10 seconds\n", pArg_u);
	return;
    }
#endif
    if (pStamp->secPastEpoch > 86400) {
	RAW_ERROR("...time can't be more than 24 hours\n", pArg_u);
	return;
    }
}

/*+/internal******************************************************************
* NAME	rawParseTrigStamp
*
*-*/
static void rawParseTrigStamp(pTrig)
REQ_TRIG *pTrig;
{
    rawParseStamp(&pTrig->timeStamp);
}

/*+/subr**********************************************************************
* NAME	rawParseStrcmp - get next field and check if it matches keyword
*
* RETURNS
*	0	if there is a match
*	-1	if there isn't a match
*
*-*/
static int
rawParseStrcmp(keyword, required)
char	*keyword;
int	required;
{
    if (RAW_NEXT_FLD <= 1 || strcmp(pField, keyword) != 0) {
	if (!required && (*pField == '#' || L < 1))
	    return -1;
	RAW_ERROR("...keyword expected: ", pArg_u);
	msgFn_u(keyword, pArg_u);
	msgFn_u("\n");
	return -1;
    }
    return 0;
}

/*+/internal******************************************************************
* NAME	rawReqAct - handle ACT: lines
*
*-*/
static void rawReqAct()
{
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...activation keyword expected\n", pArg_u);
	return;
    }
    if (strcmp(pField, "allow") == 0) {
	if (rawParseStrcmp("disconnects", 1) != 0) {
	    RAW_ERROR("...legal form is \"allow disconnects\"\n", pArg_u);
	    return;
	}
	else
	    pRaw_u->allowDiscon = 1;
	    return;
    }
    if (++pRaw_u->hadAct > 1) {
	RAW_ERROR("...activation type already specified\n", pArg_u);
	return;
    }
    if (strcmp(pField, glReqTypeText[REQ_TYPE_ALWAYS]) == 0)
	pRaw_u->actTrig.type = REQ_TYPE_ALWAYS;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_DEMAND]) == 0)
	pRaw_u->actTrig.type = REQ_TYPE_DEMAND;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_EVENT]) == 0) {
	pRaw_u->actTrig.type = REQ_TYPE_EVENT;
	rawParseEvent(&pRaw_u->actTrig);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_IF]) == 0) {
	pRaw_u->actTrig.type = REQ_TYPE_IF;
	rawParseIf(&pRaw_u->actTrig);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_EVERY]) == 0) {
	pRaw_u->actTrig.type = REQ_TYPE_PERIODIC;
	rawParseTrigStamp(&pRaw_u->actTrig);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_WHEN]) == 0) {
	pRaw_u->actTrig.type = REQ_TYPE_EVENT;
	rawParseEvent(&pRaw_u->actTrig);
    }
    else {
	RAW_ERROR("...illegal activation type\n", pArg_u);
	return;
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqChan - process a "channels" line (following %endHeader%)
*
*-*/
static void rawReqChan()
{
    RAW_CHAN	*pChan, *pChan1;

    while (RAW_NEXT_FLD > 1) {
	if (*pField == '#')
	    return;
	pChan = (RAW_CHAN *)genMalloc(sizeof(RAW_CHAN));
	assertAlways(pChan != NULL);
	pChan->pCh = NULL;
	if (strcmp(pField, "F") == 0) {
	    pChan->filled = 1;
	    RAW_NEXT_FLD;
	}
	else
	    pChan->filled = 0;
	strcpy(pChan->name, pField);
	for (pChan1=pRaw_u->pChanHead; pChan1!=NULL; pChan1=pChan1->pNext) {
	    if (strcmp(pChan1->name, pField) == 0) {
sprintf(message, "...ignoring channel %s--it was already used\n", pField);
		RAW_ERROR(message, pArg_u);
		genFree((char *)pChan);
		return;
	    }
	}
	DoubleListAppend(pChan, pRaw_u->pChanHead, pRaw_u->pChanTail);
	pChan->pRaw = pRaw_u;
	pChan->isAField = 0;
    }
}

/*+/internal******************************************************************
* NAME	rawReqDesc - handle DESC: item
*
*-*/
static void rawReqDesc()
{
    if (strlen(pLine) >= 120)
	*(pLine+119) = '\0';
    strcpy(pRaw_u->desc, pLine);
}

/*+/internal******************************************************************
* NAME	rawReqExport - handle EXPORT: line
*
*-*/
static void rawReqExport()
{
    if (++pRaw_u->hadExport > 1) {
	RAW_ERROR("...export file already specified\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1)
	sprintf(pRaw_u->exportFile, "./data/%s.exp", pRaw_u->name);
    else
	strcpy(pRaw_u->exportFile, pField);
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqInact - handle INACT: line
*
*-*/
static void rawReqInact()
{
    if (++pRaw_u->hadInact > 1) {
	RAW_ERROR("...inactivation type already specified\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...inactivation type expected\n", pArg_u);
	return;
    }
    if (strcmp(pField, glReqTypeText[REQ_TYPE_NEVER]) == 0)
	pRaw_u->inactTrig.type = REQ_TYPE_NEVER;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_AFTER]) == 0) {
	pRaw_u->inactTrig.type = REQ_TYPE_TIMED;
	rawParseTrigStamp(&pRaw_u->inactTrig);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_DEMAND]) == 0)
	pRaw_u->inactTrig.type = REQ_TYPE_DEMAND;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_EVENT]) == 0) {
	pRaw_u->inactTrig.type = REQ_TYPE_EVENT;
	rawParseEvent(&pRaw_u->inactTrig);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_IF]) == 0) {
	pRaw_u->inactTrig.type = REQ_TYPE_IF;
	rawParseIf(&pRaw_u->inactTrig);
    }
    else if (strcmp(pField, "sample") == 0) {
	if (rawParseStrcmp("complete", 1) != 0) return;
	pRaw_u->inactTrig.type = REQ_TYPE_DONE;
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_WHEN]) == 0) {
	pRaw_u->inactTrig.type = REQ_TYPE_EVENT;
	rawParseEvent(&pRaw_u->inactTrig);
    }
    else {
	RAW_ERROR("...illegal inactivation type\n", pArg_u);
	return;
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqMon - handle MONITOR: lines
*
* DESCRIPTION
*	Set up a sample structure for monitor sampling.
*
*-*/
static void rawReqMon()
{
    REQ_SAMPLE *pSamp=&pRaw_u->sample;

    pRaw_u->hadMonitor++;
    if (pRaw_u->hadSnapshot) {
	RAW_ERROR("...can't use MONITOR: with SNAPSHOT:\n", pArg_u);
	return;
    }
    pSamp->type = REQ_TYPE_MONITOR;
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...monitor keyword expected\n", pArg_u);
	return;
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_FOREVER]) == 0) {
	pSamp->forType = REQ_TYPE_FOREVER;
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_FOR]) == 0) {
	if (pRaw_u->hadFor++ > 0) {
	    RAW_ERROR("...\"MONITOR: for\" already used\n", pArg_u);
	    return;
	}
	rawParseStamp(&pSamp->forStamp);
	pSamp->forType = REQ_TYPE_TIMED;
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_ROUND]) == 0) {
	if (rawParseStrcmp("stamps", 1) != 0) return;
	if (rawParseGetInt(&pRaw_u->sample.roundMsec, 0, 5000) != 0) return;
	rawParseStrcmp("msec", 1);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_UNTIL]) == 0) {
	pSamp->forType = REQ_TYPE_EVENT;
	rawParseWhile(&pSamp->forEvent);
    }
    else {
	RAW_ERROR("...illegal monitor keyword\n", pArg_u);
	return;
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqName - handle NAME: item
*
*-*/
static void rawReqName()
{
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...name expected\n", pArg_u);
	return;
    }
    if (dlm != '\n') {
	RAW_ERROR("...\"name\" must be a single word\n", pArg_u);
	return;
    }
    strcpy(pRaw_u->name, pField);
}

/*+/internal******************************************************************
* NAME	rawReqOutput - handle OUTPUT: line
*
*-*/
static void rawReqOutput()
{
    int		i;
    char	*pC, path[120];

    if (++pRaw_u->hadOutput > 1) {
	RAW_ERROR("...output type already specified\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...output type expected\n", pArg_u);
	return;
    }
    for (i=1; i<REQ_OUT_PAST; i++) {
	if (strcmp(glReqOutText[i], pField) == 0)
	    break;
    }
    if (i >= REQ_OUT_PAST) {
	RAW_ERROR("...output type is illegal\n", pArg_u);
	return;
    }
    pRaw_u->sample.outputType = i;
    if (pRaw_u->sample.outputType == REQ_OUT_BCF) {
	char	*fileName=NULL, *fileSize=NULL, *keyword=NULL;
	if (RAW_NEXT_FLD > 1) fileName = pField;
	if (RAW_NEXT_FLD > 1) fileSize = pField;
	if (RAW_NEXT_FLD > 1) keyword = pField;
	if (keyword == NULL) {
	    if (fileSize == NULL || fileName == NULL) {
		RAW_ERROR("...file size in Mbytes must be specified\n");
		return;
	    }
	    keyword = fileSize;
	    fileSize = fileName;
	    fileName = NULL;
	}
	if (strcmp(keyword, "Mbytes") != 0) {
	    RAW_ERROR("...\"Mbytes\" keyword expected\n");
	    return;
	}
	if (sscanf(fileSize, "%d", &pRaw_u->sample.outSize) != 1) {
	    RAW_ERROR("...file size is invalid\n");
	    return;
	}
	if (pRaw_u->sample.outSize < 1 || pRaw_u->sample.outSize > 100) {
	    RAW_ERROR("...file size must be between 1 and 100\n");
	    return;
	}
	pRaw_u->sample.outSize *= 1000000;
	if (fileName == NULL)
	    sprintf(path, "./arChan/%s.bin", pRaw_u->name);
	else
	    strcpy(path, fileName);
    }
    else {
	if (RAW_NEXT_FLD <= 1)
	    sprintf(path, "./arSet/%s.txt", pRaw_u->name);
	else
	    strcpy(path, pField);
    }
    if ((pC = strrchr(path, '/')) != NULL) {
	*pC++ = '\0';
	strcpy(pRaw_u->sample.outDir, path);
	strcpy(pRaw_u->sample.outFile, pC);
    }
    else {
	strcpy(pRaw_u->sample.outDir, ".");
	strcpy(pRaw_u->sample.outFile, path);
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqSnap - handle SNAPSHOT: lines
*
* DESCRIPTION
*	Set up a sample structure for snapshot sampling.
*
*-*/
static void rawReqSnap()
{
    REQ_SAMPLE *pSamp=&pRaw_u->sample;
    int		n;

    pRaw_u->hadSnapshot++;
    if (pRaw_u->hadMonitor) {
	RAW_ERROR("...can't use SNAPSHOT: with MONITOR:\n", pArg_u);
	return;
    }
    pSamp->type = REQ_TYPE_SNAP;
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...snapshot keyword expected\n", pArg_u);
	return;
    }
    if (strcmp(pField, glReqTypeText[REQ_TYPE_EVERY]) == 0 ||
		strcmp(pField, glReqTypeText[REQ_TYPE_PERIODIC]) == 0) {
	rawParseStampOrCount("every",
		&pSamp->everyType, &pSamp->everyCmp, &pSamp->everyStamp);
	if (pSamp->everyType == REQ_TYPE_SAMPLES) {
	    pSamp->everyType = REQ_TYPE_NTH;
	    if (pSamp->everyCmp > pRaw_u->sample.bufNsamples)
		pRaw_u->sample.bufNsamples = pSamp->everyCmp;
	}
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_FOR]) == 0) {
	if (pRaw_u->hadFor++ > 0) {
	    RAW_ERROR("...\"SNAPSHOT: for\" already used\n", pArg_u);
	    return;
	}
	rawParseStampOrCount("for",
		&pSamp->forType, &pSamp->forCmp, &pSamp->forStamp);
	if (pSamp->forType == REQ_TYPE_TIMED)
	    goto nSamplesMax;
	else {
	    if (pSamp->forCmp > pRaw_u->sample.bufNsamples)
		pRaw_u->sample.bufNsamples = pSamp->forCmp;
	}
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_ROUND]) == 0) {
	if (rawParseStrcmp("stamps", 1) != 0) return;
	if (rawParseGetInt(&pRaw_u->sample.roundMsec, 0, 5000) != 0) return;
	rawParseStrcmp("msec", 1);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_START]) == 0) {
	rawParseStampOrCount("start",
		&pSamp->startType, &pSamp->startCount, &pSamp->startStamp);
	if (pSamp->startType == REQ_TYPE_SAMPLES) {
	    pSamp->startType = REQ_TYPE_NTH;
	    if (pSamp->startCount > pRaw_u->sample.bufNsamples)
		pRaw_u->sample.bufNsamples = pSamp->startCount;
	}
	if (RAW_NEXT_FLD <= 1) {
	    RAW_ERROR("\"before trigger\" (or \"after\") expected\n", pArg_u);
	    return;
	}
	if (strcmp(pField, glReqTypeText[REQ_TYPE_BEFORE]) == 0)
	    pSamp->startWhen = REQ_TYPE_BEFORE;
	else if (strcmp(pField, glReqTypeText[REQ_TYPE_AFTER]) == 0)
	    pSamp->startWhen = REQ_TYPE_AFTER;
	else {
	    RAW_ERROR("\"before trigger\" (or \"after\") expected\n", pArg_u);
	    return;
	}
	if (rawParseStrcmp("trigger", 1) != 0) {
	    RAW_ERROR("...\"before\" or \"after\" expected\n", pArg_u);
	    RAW_ERROR("\"before trigger\" (or \"after\") expected\n", pArg_u);
	    return;
	}
    }
    else if (strcmp(pField, "allow") == 0) {
	if (rawParseStrcmp("partial", 1) != 0 ||
			rawParseStrcmp("samples", 1) != 0) {
	    RAW_ERROR("...legal form is \"allow partial samples\"\n", pArg_u);
	    return;
	}
	else
	    pSamp->ignorePartial = 0;
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_UNTIL]) == 0) {
	pSamp->forType = REQ_TYPE_EVENT;
	rawParseWhile(&pSamp->forEvent);
	goto nSamplesMax;
    }
    else {
	RAW_ERROR("...illegal snapshot keyword\n", pArg_u);
	return;
    }
    RAW_EOL_CHECK;
    return;
nSamplesMax:
    if (rawParseGetInt(&n, 0, 1000) != 0) {
	RAW_ERROR("...\"n samples maximum\" expected\n", pArg_u);
	return;
    }
    if (rawParseStrcmp("samples", 1) != 0) return;
    if (rawParseStrcmp("maximum", 1) != 0) return;
    if (n > pRaw_u->sample.bufNsamples)
	pRaw_u->sample.bufNsamples = n;
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqSampleStat - handle SAMP_STATUS: line
*
*-*/
static void rawReqSampleStat()
{
    if (++pRaw_u->hadSampleStat > 1) {
	RAW_ERROR("...sample status channel already specified\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...sample status channel expected\n", pArg_u);
	return;
    }
    strcpy(pRaw_u->sampleStatChan.name, pField);
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqStep - handle STEP: lines
*
*-*/
static void rawReqStep()
{
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...step parameters expected\n", pArg_u);
	return;
    }
    if (strcmp(pField, glReqTypeText[REQ_TYPE_DELAY]) == 0) {
	rawParseGetDbl(&pRaw_u->step.delay, .1, 60.);
	if (rawParseStrcmp("seconds", 1) != 0) return;
    }
    else {
	pRaw_u->hadStep++;
	if (strlen(pField) >= db_name_dim) {
	    RAW_ERROR("...channel name too long\n", pArg_u);
	    return;
	}
	strcpy(pRaw_u->step.name, pField);
	if (rawParseStrcmp("from", 1) != 0) return;
	if (rawParseGetDbl(&pRaw_u->step.beg, 0., 0.) != 0) return;
	if (rawParseStrcmp("to", 1) != 0) return;
	if (rawParseGetDbl(&pRaw_u->step.end, 0., 0.) != 0) return;
	if (rawParseStrcmp("by", 1) != 0) return;
	if (rawParseGetDbl(&pRaw_u->step.step, 0., 0.) != 0) return;
	if (pRaw_u->step.beg == pRaw_u->step.end) {
	    RAW_ERROR("...begin and end are equal\n", pArg_u);
	    return;
	}
	pRaw_u->step.type = REQ_TYPE_STEP;
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqSampleTrigger - handle TRIGGER: lines
*
*-*/
static void rawReqSampleTrigger()
{
    pRaw_u->hadTrigger++;
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...trigger type expected\n", pArg_u);
	return;
    }
    if (strcmp(pField, glReqTypeText[REQ_TYPE_ALWAYS]) == 0)
	pRaw_u->sample.trigger.type = REQ_TYPE_ALWAYS;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_DEMAND]) == 0)
	pRaw_u->sample.trigger.type = REQ_TYPE_DEMAND;
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_EVERY]) == 0) {
	pRaw_u->sample.trigger.type = REQ_TYPE_PERIODIC;
	rawParseTrigStamp(&pRaw_u->sample.trigger);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_IF]) == 0) {
	pRaw_u->sample.trigger.type = REQ_TYPE_IF;
	rawParseIf(&pRaw_u->sample.trigger);
    }
    else if (strcmp(pField, glReqTypeText[REQ_TYPE_WHEN]) == 0) {
	pRaw_u->sample.trigger.type = REQ_TYPE_EVENT;
	rawParseEvent(&pRaw_u->sample.trigger);
    }
    else if (strcmp(pField, "after") == 0) {
	pRaw_u->sample.trigger.type = REQ_TYPE_AFTER_STEP;
	rawParseStrcmp("each", 1);
	rawParseStrcmp("step", 1);
    }
    else if (strcmp(pField, "at") == 0) {
	rawParseStrcmp("least", 1);
	rawParseStrcmp("every", 1);
	rawParseStamp(&pRaw_u->sample.trigger.atleastStamp);
	pRaw_u->sample.trigger.atleastType = REQ_TYPE_TIMED;
    }
    else {
	RAW_ERROR("...illegal trigger keyword\n", pArg_u);
	return;
    }
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqReqStat - handle REQ_STATUS: line
*
*-*/
static void rawReqReqStat()
{
    if (++pRaw_u->hadReqStat > 1) {
	RAW_ERROR("...request status channel already specified\n", pArg_u);
	return;
    }
    if (RAW_NEXT_FLD <= 1) {
	RAW_ERROR("...request status channel expected\n", pArg_u);
	return;
    }
    strcpy(pRaw_u->reqStatChan.name, pField);
    RAW_EOL_CHECK;
}

/*+/internal******************************************************************
* NAME	rawReqEndHdr - process the %endHeader% line
*
*-*/
static void rawReqEndHdr()
{
    if (++pRaw_u->hadEndHead > 1) {
	RAW_ERROR("...%endHeader% already encountered\n", pArg_u);
    }
    RAW_EOL_CHECK;
}


/*+/internal******************************************************************
* NAME	rawMessageFn - default error message writing function
*
*-*/
static void rawMessageFn(message, pArg) char *message; void *pArg; {
    fputs(message, stderr);
}

/*+/subr**********************************************************************
* NAME	rawReqParseLine
*
*-*/
int
rawReqParseLine(pRaw, line, messageFn, pArg)
RAW_REQ	*pRaw;		/* I pointer to raw request structure */
char	*line;		/* I pointer to input line */
void	(*messageFn)();	/* I ptr to fn to call with error messages, or NULL */
void	*pArg;		/* I pointer to arg for messageFn */
{
    int		i;
    static struct {
	char	*keyword;
	void	(*fn)();
    } rawItems[]={
	"NAME:", rawReqName, "DESC:", rawReqDesc,
	"ACT:", rawReqAct, "INACT:", rawReqInact,
	"SNAPSHOT:", rawReqSnap, "MONITOR:", rawReqMon,
	"TRIGGER:", rawReqSampleTrigger,
	"STEP:", rawReqStep,
	"OUTPUT:", rawReqOutput,
	"SAMP_STATUS:", rawReqSampleStat, "REQ_STATUS:", rawReqReqStat,
	"EXPORT:", rawReqExport,
	"%endHeader%", rawReqEndHdr,
	"", NULL
    };

    errorCount = 0;
    if (messageFn == NULL)
	messageFn = rawMessageFn;
    tempLine = (char *)malloc(strlen(line)+1);
    assertAlways(tempLine != NULL);
    strcpy(tempLine, line);
    pLine = tempLine;
    line_u = line;
    pRaw_u = pRaw;
    msgFn_u = messageFn;
    pArg_u = pArg;
    if (pRaw->hadEndHead) {
	rawReqChan(line, messageFn, pArg, pRaw, pLine);
	goto rawReqParseLineDone;
    }
    if (RAW_NEXT_FLD <= 1 || *pField == '#')
	goto rawReqParseLineDone;
    for (i=0; rawItems[i].keyword[0]!='\0'; i++) {
	if (strcmp(rawItems[i].keyword, pField) == 0)
	    break;
    }
    if (rawItems[i].keyword[0] == '\0') {
	errorCount++;
	messageFn(line, pArg);
	messageFn("...unrecognized keyword: ", pArg);
	messageFn(pField, pArg);
	messageFn("\n", pArg);
	goto rawReqParseLineDone;
    }
    rawItems[i].fn();

rawReqParseLineDone:
    free((char *)tempLine);
    pRaw->totalErrors += errorCount;
    return errorCount;
}

/*+/subr**********************************************************************
* NAME	rawReqTrigInit - initialize fields in a method structure
*
*-*/
void
rawReqTrigInit(pTrig)
REQ_TRIG *pTrig;
{
    pTrig->type = REQ_TYPE_ALWAYS;
    pTrig->atleastType = REQ_TYPE_NEVER;
    pTrig->eventCond = REQ_EV_UNDEF;
    pTrig->eventName[0] = '\0';
    pTrig->event_pCh = NULL;
}

/*+/subr**********************************************************************
* NAME	rawReqAlloc - create and initialize a raw request structure
*
*-*/
RAW_REQ *
rawReqAlloc()
{
    RAW_REQ	*pRaw;

    pRaw = (RAW_REQ *)genMalloc(sizeof(RAW_REQ));
    assertAlways(pRaw != NULL);
    bzero((char *)pRaw, sizeof(RAW_REQ));
    pRaw->pChanHead = pRaw->pChanTail = NULL;
    pRaw->reqStatChan.pCh = NULL;
    pRaw->reqStatChan.name[0] = '\0';
    pRaw->sampleStatChan.pCh = NULL;
    pRaw->sampleStatChan.name[0] = '\0';
    pRaw->actChan.pCh = NULL;
    pRaw->inactChan.pCh = NULL;
    pRaw->forChan.pCh = NULL;
    pRaw->triggerChan.pCh = NULL;

    rawReqTrigInit(&pRaw->actTrig);
    pRaw->allowDiscon = 0;
    rawReqTrigInit(&pRaw->inactTrig);

    pRaw->sample.type = REQ_TYPE_UNDEF;
    pRaw->sample.everyType = REQ_TYPE_CONTIN;
    pRaw->sample.everyCount = 0;
    pRaw->sample.forType = REQ_TYPE_FOREVER;
    pRaw->sample.forCmp = 0;
    rawReqTrigInit(&pRaw->sample.trigger);
    rawReqTrigInit(&pRaw->sample.forEvent);
    pRaw->sample.startType = REQ_TYPE_UNDEF;
    pRaw->sample.startCount = 0;
    pRaw->sample.startWhen = REQ_TYPE_AFTER;
    pRaw->sample.bufNsamples = 0;
    pRaw->sample.roundMsec = 0;
    pRaw->sample.ignorePartial = 1;

    pRaw->step.type = REQ_TYPE_UNDEF;
    pRaw->step.delay = 0.;
    pRaw->step.pCh = pRaw->stepChan.pCh = NULL;

    return pRaw;
}

void rawReqConnHand();
void rawReqConnHand_ADEL();
void rawReqMonHand(), rawReqMonHand_ADEL();

/*+/subr**********************************************************************
* NAME	rawReqCheckConnect_init - initialize check connections for channels
*
* DESCRIPTION
*	Initialize for checking the connections for channels in an archive
*	request.
*
* EXAMPLE
*	rawReqCheckConnect_init(pRaw, myPrintFn, pMyCtx);
*	while (1) {
*	    if (maximum delay time expired)
*		break;
*	}
*	rawReqCheckConnect_wrap(pRaw, myPrintFn, pMyCtx);
*
*-*/
void
rawReqCheckConnect_init(pRaw, messageFn, pArg)
RAW_REQ	*pRaw;		/* I pointer to raw request structure */
void	(*messageFn)();	/* I pointer to function to print messages */
void	*pArg;		/* I pointer to arg for messageFn */
{
    RAW_CHAN	*pChan;

    if (messageFn == NULL)
	messageFn = rawMessageFn;

    strcpy(pRaw->actChan.name, pRaw->actTrig.eventName);
    strcpy(pRaw->inactChan.name, pRaw->inactTrig.eventName);
    strcpy(pRaw->forChan.name, pRaw->sample.forEvent.eventName);
    strcpy(pRaw->triggerChan.name, pRaw->sample.trigger.eventName);
    strcpy(pRaw->stepChan.name, pRaw->step.name);

    rawReqConnect_1(&pRaw->actChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->inactChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->forChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->triggerChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->sampleStatChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->reqStatChan, messageFn, pArg);
    rawReqConnect_1(&pRaw->stepChan, messageFn, pArg);

    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqConnect_1(pChan, messageFn, pArg);
}

/*+/subr**********************************************************************
* NAME	rawReqCheckConnect_wrap - wrapup check for connections for channels
*
*-*/
void
rawReqCheckConnect_wrap(pRaw, messageFn, pArg)
RAW_REQ	*pRaw;		/* I pointer to raw request structure */
void	(*messageFn)();	/* I pointer to function to print messages */
void	*pArg;		/* I pointer to arg for messageFn */
{
    RAW_CHAN	*pChan;

    sprintf(message, "%-10.10s %-8.8s %30s %c ADEL\n",
				"host", "type", "name", 'F');
    messageFn(message, pArg);
    rawReqConnCheck_1(&pRaw->actChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->inactChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->forChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->triggerChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->sampleStatChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->reqStatChan, messageFn, pArg);
    rawReqConnCheck_1(&pRaw->stepChan, messageFn, pArg);

    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqConnCheck_1(pChan, messageFn, pArg);

    rawReqConnClear_1(&pRaw->actChan);
    rawReqConnClear_1(&pRaw->inactChan);
    rawReqConnClear_1(&pRaw->forChan);
    rawReqConnClear_1(&pRaw->triggerChan);
    rawReqConnClear_1(&pRaw->sampleStatChan);
    rawReqConnClear_1(&pRaw->reqStatChan);
    rawReqConnClear_1(&pRaw->stepChan);

    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqConnClear_1(pChan);
}

/*+/subr**********************************************************************
* NAME	rawReqCheckInterval_init - init for check of data time interval
*
*-*/
void
rawReqCheckInterval_init(pRaw, messageFn, pArg, intervalSec, jitterSec)
RAW_REQ	*pRaw;		/* I pointer to raw request structure */
void	(*messageFn)();	/* I pointer to function to print messages */
void	*pArg;		/* I pointer to arg for messageFn */
double	intervalSec;	/* I expected interval between samples */
double	jitterSec;	/* I jitter allowed in interval */
{
    RAW_CHAN	*pChan;
    
    if (messageFn == NULL)
	messageFn = rawMessageFn;
    
    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqConnect_1(pChan, messageFn, pArg);
    
    pRaw->jitterSec = jitterSec;
    pRaw->intervalSec = intervalSec;
}
    
/*+/subr**********************************************************************
* NAME	rawReqCheckInterval_wrap - wrapup for check of data time interval
*
*-*/
void
rawReqCheckInterval_wrap(pRaw, messageFn, pArg)
RAW_REQ	*pRaw;		/* I pointer to raw request structure */
void	(*messageFn)();	/* I pointer to function to print messages */
void	*pArg;		/* I pointer to arg for messageFn */
{
    RAW_CHAN	*pChan;

    sprintf(message,
"%-10.10s %-8.8s %30s %c ADEL      n  nOK    min    max %-17.17s %-17.17s\n",
			"host", "type", "name", 'F', "min stamp", "max stamp");
    messageFn(message, pArg);

    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqCheckInterval_1(pChan, messageFn, pArg);

    for (pChan=pRaw->pChanHead; pChan!=NULL; pChan=pChan->pNext)
	rawReqConnClear_1(pChan);
}

/*+/internal******************************************************************
* NAME	rawReqCheckInterval_1 - start a Channel Access monitor for 1 channel
*
*-*/
static rawReqCheckInterval_1(pChan, messageFn, pArg)
RAW_CHAN *pChan;
void	(*messageFn)();
void	*pArg;		/* I pointer to arg for messageFn */
{
    char	*typeText, *hostText, filled=' ';
    char	deadbandText[20], typeFromCA[40];
    char	*pC;
    char	minTsText[28], maxTsText[28];

    if (pChan->name[0] == NULL)
	return;

    typeText = hostText = "notConn";
    if (pChan->pCh != NULL) {
	strcpy(typeFromCA, dbf_type_to_text(ca_field_type(pChan->pCh)));
	typeText = &typeFromCA[4];
	hostText = ca_host_name(pChan->pCh);
	if ((pC = strchr(hostText, '.')) != NULL)
	    *pC = '\0';
	if (pChan->filled)	filled = 'F';
    }
    if (pChan->pCh == NULL || ca_field_type(pChan->pCh) == TYPENOTCONN)
	strcpy(deadbandText, "noConn");
    else if (pChan->pChADEL == NULL)
	strcpy(deadbandText, "err");
    else if (ca_field_type(pChan->pChADEL) == TYPENOTCONN)
	strcpy(deadbandText, "n/a");
    else if (pChan->isAField)
	strcpy(deadbandText, "anyChg");
    else if (pChan->deadband == -10.)
	strcpy(deadbandText, "noGet");
    else
	cvtDblToTxt(deadbandText, 6, (double)pChan->deadband, 3);

    tsStampToText(&pChan->minTs, TS_TEXT_MMDDYY, minTsText);
    tsStampToText(&pChan->maxTs, TS_TEXT_MMDDYY, maxTsText);
    sprintf(message,
    "%-10.10s %-8.8s %30s %c %-6s %4d %4d %6.2f %6.2f %-17.17s %-17.17s\n",
			hostText, typeText, pChan->name, filled, deadbandText,
			pChan->nSamp, pChan->nSampOK,
			pChan->minInterval, pChan->maxInterval,
			minTsText, maxTsText);
    messageFn(message, pArg);
}

/*+/subr**********************************************************************
* NAME	rawReqCheckSemantics - check semantics for a raw archive request
*
* BUGS
* o	check for using AFTER with MONITOR isn't made
*
*-*/
void
rawReqCheckSemantics(pRaw, messageFn, pArg)
RAW_REQ	*pRaw;		/* I pointer to raw archive request to check */
void	(*messageFn)();	/* I pointer to function to print messages, or NULL */
void	*pArg;		/* I pointer to arg for messageFn */
{
    REQ_TRIG	*pA=&pRaw->actTrig, *pI=&pRaw->inactTrig;
    REQ_TRIG	*pT=&pRaw->sample.trigger;
    REQ_SAMPLE	*pS=&pRaw->sample;
    double	diff;

    if (messageFn == NULL)
	messageFn = rawMessageFn;
    if (pRaw->hadAct == 0)
	pRaw->totalErrors++, messageFn("...ACT: wasn't found\n", pArg);
    if (pRaw->hadInact == 0)
	pRaw->totalErrors++, messageFn("...INACT: wasn't found\n", pArg);
    if (pRaw->hadTrigger == 0)
	pRaw->totalErrors++, messageFn("...TRIGGER: wasn't found\n", pArg);
    if (pRaw->hadSnapshot == 0 && pRaw->hadMonitor == 0)
	pRaw->totalErrors++, messageFn(
	"...neither SNAPSHOT: nor MONITOR: was found\n", pArg);
    if (pRaw->hadOutput == 0)
	pRaw->totalErrors++, messageFn("...OUTPUT: wasn't found\n", pArg);
    if (pRaw->pChanHead == NULL)
	pRaw->totalErrors++, messageFn("...no channels were named\n", pArg);

    if (pA->type == REQ_TYPE_EVENT && pA->eventCond == REQ_EV_CHANGE) {
	sprintf(message, "...%s illegal for ACT: %s\n",
		glReqEventText[REQ_EV_CHANGE], glReqTypeText[pA->type]);
	pRaw->totalErrors++, messageFn(message, pArg);
    }
    if (pI->type == REQ_TYPE_EVENT && pI->eventCond == REQ_EV_CHANGE) {
	sprintf(message, "...%s illegal for INACT: %s\n",
		glReqEventText[REQ_EV_CHANGE], glReqTypeText[pI->type]);
	pRaw->totalErrors++, messageFn(message, pArg);
    }
    if (pI->type == REQ_TYPE_TIMED && pA->type == REQ_TYPE_PERIODIC) {
	TsDiffAsDouble(&diff, &pA->timeStamp, &pI->timeStamp);
	if (diff < 10.) {
	    pRaw->totalErrors++;
messageFn("...INACT: time must be at least 10 sec less than ACT: period\n", pArg);
	}
    }
    if (pS->trigger.type == REQ_TYPE_AFTER_STEP && pRaw->hadStep == 0)
	pRaw->totalErrors++, messageFn(
	"...TRIGGER: after each step used, but no STEP: command found\n", pArg);
    if (pS->type == REQ_TYPE_MONITOR || pRaw->hadMonitor) {
	if (pRaw->sample.outputType == REQ_OUT_SSF) {
	    sprintf(message, "...MONITOR: is illegal for OUTPUT: %s\n",
		glReqOutText[REQ_OUT_SSF]);
	    pRaw->totalErrors++, messageFn(message, pArg);
	}
    }
    else if (pS->type == REQ_TYPE_SNAP || pRaw->hadSnapshot) {
	if (pS->bufNsamples == 0) {
	    pRaw->totalErrors++;
	    messageFn("...max number of samples wasn't set\n", pArg);
	}
	if (pS->forType == REQ_TYPE_FOREVER) {
	    pRaw->totalErrors++;
messageFn("...for SNAPSHOT:, either for or until must be used\n", pArg);
	}
    }

#if 0
    jjj

don't allow both `for' and `until'
dont allow REQ_TYPE_IF with `change'

    more checks are needed

#endif
}

/*+/internal******************************************************************
* NAME	rawReqConnCheck_1 - do connect check for 1 channel
*
*-*/
static rawReqConnCheck_1(pChan, messageFn, pArg)
RAW_CHAN *pChan;
void	(*messageFn)();
void	*pArg;		/* I pointer to arg for messageFn */
{
    char	*typeText, *hostText, filled=' ';
    char	deadbandText[20], typeFromCA[40];
    char	*pC;

    typeText = hostText = "notConn";
    if (pChan->name[0] != NULL) {
	if (pChan->pCh != NULL) {
	    strcpy(typeFromCA, dbf_type_to_text(ca_field_type(pChan->pCh)));
	    typeText = &typeFromCA[4];
	    hostText = ca_host_name(pChan->pCh);
	    if ((pC = strchr(hostText, '.')) != NULL)
		*pC = '\0';
	    if (pChan->filled)	filled = 'F';
	}
	if (pChan->pCh == NULL || ca_field_type(pChan->pCh) == TYPENOTCONN)
	    strcpy(deadbandText, "noConn");
	else if (pChan->pChADEL == NULL)
	    strcpy(deadbandText, "err");
	else if (ca_field_type(pChan->pChADEL) == TYPENOTCONN)
	    strcpy(deadbandText, "n/a");
	else if (pChan->isAField)
	    strcpy(deadbandText, "anyChg");
	else if (pChan->deadband == -10.)
	    strcpy(deadbandText, "noGet");
	else
	    cvtDblToTxt(deadbandText, 6, (double)pChan->deadband, 3);
	sprintf(message, "%-10.10s %-8.8s %30s %c %-s\n",
			hostText, typeText, pChan->name, filled, deadbandText);
	messageFn(message, pArg);
    }
}

/*+/internal******************************************************************
* NAME	rawReqConnClear_1 - do ca_clear_channel for 1 channel
*
*-*/
static rawReqConnClear_1(pChan)
RAW_CHAN *pChan;
{
    int		stat;

    if (pChan->pCh == NULL)
	return;
    stat = ca_clear_channel(pChan->pCh);
    if (pChan->pChADEL != NULL)
	stat = ca_clear_channel(pChan->pChADEL);
    pChan->pCh = NULL;
}

/*+/internal******************************************************************
* NAME	rawReqConnect_1 - do ca_search for 1 channel
*
* DESCRIPTION
*	Does a search for the channel, with rawReqConnHand as the connect
*	handler.  For all channels, a search is also done for the .ADEL
*	field, with rawReqConnHand_ADEL as the connect handler.  This
*	strategy allows distinguishing records which have an ADEL field
*	and those which don't.  For those that don't, "n/a" will appear
*	on the listing.  For fields other than .VAL for records that have
*	an ADEL field, the listing will show "anyChg".
*
*-*/
static rawReqConnect_1(pChan, messageFn, pArg)
RAW_CHAN *pChan;
void	(*messageFn)();
void	*pArg;		/* I pointer to arg for messageFn */
{
    int		stat;
    char	deadName[db_name_dim];
    char	*dot;

    if (pChan->name[0] == NULL)
	return;
    assert(pChan->pCh == NULL);
    stat = ca_build_and_connect(pChan->name, TYPENOTCONN, 0, &pChan->pCh,
			NULL, rawReqConnHand, pChan);
    assertAlways(stat == ECA_NORMAL);
    stat = ca_add_masked_array_event(DBR_TIME_STRING, 1, pChan->pCh,
	rawReqMonHand, pChan, 0., 0., 0., &pChan->pEv, DBE_LOG);
    assertAlways(stat == ECA_NORMAL);
#if 0
    ca_flush_io();
#endif
    dot=strchr(pChan->name, '.');
    if (dot == NULL || strcmp(dot, ".VAL") == 0)
	pChan->isAField = 0;
    else
	pChan->isAField = 1;
    if (dot != NULL) *dot = '\0';
    sprintf(deadName, "%s.ADEL", pChan->name);
    if (dot != NULL) *dot = '.';
    pChan->pEv = NULL;
    stat = ca_build_and_connect(deadName, TYPENOTCONN, 0, &pChan->pChADEL,
			NULL, rawReqConnHand_ADEL, pChan);
    assertAlways(stat == ECA_NORMAL);

    pChan->nSamp = pChan->nSampOK = 0;
    pChan->maxInterval = pChan->minInterval = 0.;
    pChan->deadband = -10.;
#if 0
    ca_flush_io();
#endif
}

/*+/subr**********************************************************************
* NAME	rawReqConnHand - connect handler for channel
*
*-*/
void rawReqConnHand(arg)
struct connection_handler_args arg;
{
    chid	pCh=arg.chid;
    RAW_CHAN	*pChan=(RAW_CHAN *)ca_puser(pCh);
    int		stat;

#if 0
    if (arg.op == CA_OP_CONN_UP) {
	stat = ca_add_masked_array_event(DBR_TIME_STRING, 1, pChan->pCh,
	    rawReqMonHand, pChan, 0., 0., 0., &pChan->pEv, DBE_LOG);
	assertAlways(stat == ECA_NORMAL);
    }
    else {
	if (pChan->pEv != NULL) {
	    stat = ca_clear_event(pChan->pEv);
	    assertAlways(stat == ECA_NORMAL);
	    pChan->pEv = NULL;
	}
    }
#endif
}

/*+/internal******************************************************************
* NAME	rawReqConnHand_ADEL - connect handler for channel.ADEL
*
*-*/
void rawReqConnHand_ADEL(arg)
struct connection_handler_args arg;
{
    chid	pCh=arg.chid;
    RAW_CHAN	*pChan=(RAW_CHAN *)ca_puser(pCh);
    if (arg.op != CA_OP_CONN_UP)
	return;
    ca_array_get_callback(DBR_FLOAT, 1, pCh, rawReqMonHand_ADEL, pChan);
#if 0
    ca_flush_io();
#endif
}

/*+/internal******************************************************************
* NAME	rawReqMonHand - handle Channel Access monitor callbacks
*
*-*/
void rawReqMonHand(arg)
struct event_handler_args arg;
{   chid	pCh=arg.chid;
    RAW_CHAN	*pChan=(RAW_CHAN *)arg.usr;
    TS_STAMP	*pTs=&((struct dbr_time_string *)arg.dbr)->stamp;
    double	diff, interval;

    if (++pChan->nSamp == 1)
	pChan->minTs = pChan->maxTs = *pTs;
    else {
	TsDiffAsDouble(&interval, pTs, &pChan->prevTs);
	diff = pChan->pRaw->intervalSec - interval;
	if (diff < 0.) diff *= -1.;
	if (diff <= pChan->pRaw->jitterSec)
	    ++pChan->nSampOK;
	if (pChan->nSamp == 2)
	    pChan->maxInterval = pChan->minInterval = interval;
	else {
	    if (pChan->maxInterval < interval)
		pChan->maxInterval = interval;
	    if (pChan->minInterval > interval)
		pChan->minInterval = interval;
	}
	if (TsCmpStampsLT(pTs, &pChan->minTs))
	    pChan->minTs = *pTs;
	if (TsCmpStampsGT(pTs, &pChan->maxTs))
	    pChan->maxTs = *pTs;
    }
    pChan->prevTs = *pTs;
}

/*+/subr**********************************************************************
* NAME	rawReqMonHand_ADEL - handle callback for ADEL field
*
*-*/
static void rawReqMonHand_ADEL(arg)
struct event_handler_args arg;
{
    RAW_CHAN	*pChan=(RAW_CHAN *)arg.usr;
    chid	pCh=arg.chid;

    pChan->deadband = *(float *)arg.dbr;
}


/*+/subr**********************************************************************
* NAME	rawReqFree - wrapup and deallocate a raw archive request
*
*-*/
void
rawReqFree(pRaw)
RAW_REQ	*pRaw;		/* I pointer to raw archive request to check */
{
    RAW_CHAN	*pChan, *pChan1;

    rawReqConnClear_1(&pRaw->actChan);
    rawReqConnClear_1(&pRaw->inactChan);
    rawReqConnClear_1(&pRaw->forChan);
    rawReqConnClear_1(&pRaw->triggerChan);
    rawReqConnClear_1(&pRaw->sampleStatChan);
    rawReqConnClear_1(&pRaw->reqStatChan);
    rawReqConnClear_1(&pRaw->stepChan);
    for (pChan=pRaw->pChanHead; pChan!=NULL; ) {
	pChan1 = pChan;
	pChan = pChan->pNext;
	DoubleListRemove(pChan1, pRaw->pChanHead, pRaw->pChanTail);
	rawReqConnClear_1(pChan1);
	genFree((char *)pChan1);
    }
    genFree((char *)pRaw);
}
/*+/subr**********************************************************************
* NAME	rawReqReadFile - build a raw request by reading a file
*
* DESCRIPTION
*       Reads the lines in a file, parsing them as an archive request.
*	Error messages are sent to the message writing function.
*
*       When complete, the archive request is in the "raw request" structure.
*       The ->totalErrors item in the structure indicates whether any errors
*       were encountered during the parse.
*
* RETURNS
*       RAW_REQ *, or
*	NULL if malloc failed
*
*-*/
RAW_REQ *
rawReqReadFile(dir, file, messageFn, pArg)
char	*dir;		/* I dir of file containing archive request */
char	*file;		/* I name of file containing archive request */
void	(*messageFn)();	/* I pointer to function to print messages, or NULL */
void	*pArg;		/* I pointer to arg for messageFn */
{
    RAW_REQ     *pRaw;
    char        line[120];
    FILE	*pFile;
    char	path[400];
    int		i;

    if (messageFn == NULL)
	messageFn = rawMessageFn;

    if ((pRaw = rawReqAlloc()) == NULL)
        return NULL;
    strcpy(pRaw->name, file);
    strcpy(pRaw->file, file);
    strcpy(pRaw->dir, dir);
    sprintf(path, "%s/%s", dir, file);
    if ((pFile = fopen(path, "r")) == NULL) {
	sprintf(message, "couldn't open: %s\n", path);
	pRaw->totalErrors++;
	messageFn(message, pArg);
	return pRaw;
    }
    while (fgets(line, sizeof(line), pFile) != NULL)
        rawReqParseLine(pRaw, line, messageFn, pArg);
    fclose(pFile);
    rawReqCheckSemantics(pRaw, messageFn, pArg);
    return pRaw;
}

void reqEventConn();
void reqEventMonitor();

/*+/internal******************************************************************
* NAME	reqEventConn - connection handler for trigger channels
*
* NOTES
* 1.	The .ADEL deadband is used in monitoring the channels
*
*-*/
static void
reqEventConn(arg)
struct connection_handler_args arg;
{
    chid	pCh=arg.chid;
    REQ_TRIG	*pTrig=(REQ_TRIG *)ca_puser(pCh);
    chtype	dbfType=ca_field_type(pCh);
    chtype	dbrType=dbf_type_to_DBR_TIME(dbfType);
    int		stat;
    char	strVal[db_strval_dim];

    if (arg.op == CA_OP_CONN_UP) {
	if (pTrig->notYetInit) {
	    if (dbrType == DBR_TIME_ENUM)
		dbrType = DBR_TIME_STRING;
	    stat = ca_add_masked_array_event(dbrType, 1, pCh,
			reqEventMonitor, pTrig, 0., 0., 0., NULL, DBE_LOG);
	    assertAlways(stat == ECA_NORMAL);
	    if (dbfType == DBF_ENUM && pTrig->eventCond != REQ_EV_CHANGE) {
		stat = ca_array_get_callback(DBR_GR_ENUM, 1, pCh,
			reqEventMonitor, pTrig);
		assertAlways(stat == ECA_NORMAL);
		pTrig->enumGetGR = 1;
	    }
#if 0
	    ca_flush_io();
#endif
	    strcpy(strVal, pTrig->eventValCmp.strVal);
	    if (dbfType == DBF_FLOAT)
		sscanf(strVal, "%f", &pTrig->eventValCmp.fltVal);
	    else if (dbfType == DBF_SHORT)
		sscanf(strVal, "%d", &pTrig->eventValCmp.shtVal);
	    else if (dbfType == DBF_DOUBLE)
		sscanf(strVal, "%lf", &pTrig->eventValCmp.dblVal);
	    else if (dbfType == DBF_LONG)
		sscanf(strVal, "%ld", &pTrig->eventValCmp.lngVal);
	    else if (dbfType == DBF_CHAR)
		sscanf(strVal, "%c", &pTrig->eventValCmp.chrVal);
	    pTrig->notYetInit = 0;
	}
    }
    pTrig->eventNoValYet = 0;
}

/*+/internal******************************************************************
* NAME	reqEventMonitor - monitor handler for trigger channels
*
*-*/
static void
reqEventMonitor(arg)
struct event_handler_args arg;
{
    chid	pCh=arg.chid;
    REQ_TRIG	*pTrig=(REQ_TRIG *)arg.usr;
    
    if (arg.type != DBR_GR_ENUM)
	reqEventTest(pTrig, arg.dbr);
    else {
	pTrig->enumGr = *(struct dbr_gr_enum *)arg.dbr;
	pTrig->enumGetGR = 2;
    }
}

/*+/subr**********************************************************************
* NAME	reqEventTest - test a trigger event condition
*
* DESCRIPTION
*	Tests the trigger event condition.
*
*	This routine is intended to be called for either
*	1)  storing a new value from Channel Access, and testing for
*	    satisfying the condition; or
*	2)  testing to see is the condition has been satisfied.
*
*	For the first case, the second argument is non-NULL, and the new
*	value is stored in the trigger.  If the trigger isn't armed, then
*	the `store' is all that's done, and this routine returns 0.
*
*	For the second case, if the trigger isn't armed, or if it hasn't
*	received a value yet, this routine returns 0.
*
* RETURNS
*	1 if condition is satisfied, or
*	0
*
*-*/
int
reqEventTest(pTrig, pBuf)
REQ_TRIG *pTrig;
void	*pBuf;		/* pointer to new value, or NULL */
{
    int		condMet;

    if (pTrig->notYetInit)
	return 0;
    if (pBuf == NULL && pTrig->eventNoValYet)
	return 0;
    condMet = reqEventTest_sm(pTrig, pBuf);

    if (pTrig->armed == 0)
	return 0;
    if (pTrig->eventHappened)
	return 1;

    if (pTrig->type == REQ_TYPE_IF) {
	if (condMet)
	    pTrig->eventHappened = 1;
    }
    else {
/*-----------------------------------------------------------------------------
*    Handle the "every n times" logic.  Only on a transition from "condition
*    not met" to "condition met" will everyCount be incremented.  If the
*    initial value for the event channel causes the condition to be met,
*    then that is considered to be one transition from "not met" to "met".
*----------------------------------------------------------------------------*/
	if (pTrig->notYetPrimed && condMet == 0) {
	    pTrig->everyPrimed = 1;
	    pTrig->notYetPrimed = 0;
	}
	if (condMet) {
	    if (pTrig->everyPrimed == 1) {
		if (++pTrig->everyCount >= pTrig->everyComp) {
		    pTrig->everyCount = 0;
		    pTrig->eventHappened = 1;
		}
		pTrig->everyPrimed = 0;
	    }
	}
	else if (pTrig->notYetPrimed == 0)
	    pTrig->everyPrimed = 1;
    }

    return pTrig->eventHappened;
}

/*+/subr**********************************************************************
* NAME	reqEventTest_sm - simple minded test, ignoring nuances
*
* RETURNS
*	1 if test passes, or
*	0
*
* BUGS
* o	maybe this should handle DBR_TIME_xxx buffers
*
*-*/
int
reqEventTest_sm(pTrig, pBuf)
REQ_TRIG *pTrig;
void	*pBuf;		/* pointer to new value, or null */
{
    int		condMet=0;
    chtype	type;
    int		change;		/* -1,0,1 for new <,==,> {old, cmpVal} */

#define CURR pTrig->eventValCurr
#define PREV pTrig->eventValPrev
#define CMP pTrig->eventValCmp
#define COND pTrig->eventCond

    type = ca_field_type(pTrig->event_pCh);
    if (type ==DBF_SHORT)	change = reqEventTest_sht(pTrig, pBuf);
    else if (type ==DBF_LONG)	change = reqEventTest_lng(pTrig, pBuf);
    else if (type ==DBF_FLOAT)	change = reqEventTest_flt(pTrig, pBuf);
    else if (type ==DBF_DOUBLE)	change = reqEventTest_dbl(pTrig, pBuf);
    else if (type ==DBF_STRING)	change = reqEventTest_str(pTrig, pBuf);
    else if (type ==DBF_ENUM)	change = reqEventTest_str(pTrig, pBuf);
    else if (type ==DBF_CHAR)	change = reqEventTest_chr(pTrig, pBuf);
    if (change == -2)
	;	/* no value yet for event; condition not met */
    else if (change == 0) {
	if (COND == REQ_EV_EQ || COND == REQ_EV_LE || COND == REQ_EV_GE)
	    condMet = 1;
    }
    else if (change < 0) {
	if (COND == REQ_EV_NEQ || COND == REQ_EV_LE || COND == REQ_EV_LT)
	    condMet = 1;
    }
    else {
	if (COND == REQ_EV_CHANGE ||
		COND == REQ_EV_NEQ || COND == REQ_EV_GE || COND == REQ_EV_GT)
	    condMet = 1;
    }
    return condMet;
}

/*+/internal******************************************************************
* NAME	REQ_EVENT_TEST_type - test events for various types of data
*
*-*/
#define REQ_EVENT_TEST_type(typ, type) \
static int \
reqEventTest_/**/typ(pTrig, pBuf) \
REQ_TRIG *pTrig; \
void	*pBuf; \
{ \
    int		change; \
    if (pBuf != NULL) { \
	if (pTrig->eventNoValYet) { \
	    pTrig->eventNoValYet = 0; \
	    PREV.typ/**/Val = ((struct type *)pBuf)->value; \
	} \
	else \
	    PREV.typ/**/Val = CURR.typ/**/Val; \
	CURR.typ/**/Val = ((struct type *)pBuf)->value; \
	pTrig->dataTs = ((struct type *)pBuf)->stamp; \
    } \
    if (pTrig->eventNoValYet) \
	return -2; \
    if (COND == REQ_EV_CHANGE) { \
	if (CURR.typ/**/Val == PREV.typ/**/Val)		change = 0; \
	else						change = 1; \
    } \
    else { \
	if (CURR.typ/**/Val == CMP.typ/**/Val)		change = 0; \
	else if (CURR.typ/**/Val < CMP.typ/**/Val)	change = -1; \
	else						change = 1; \
    } \
    return change; \
}

REQ_EVENT_TEST_type(sht, dbr_time_short)
REQ_EVENT_TEST_type(lng, dbr_time_long)
REQ_EVENT_TEST_type(flt, dbr_time_float)
REQ_EVENT_TEST_type(dbl, dbr_time_double)
REQ_EVENT_TEST_type(chr, dbr_time_char)

static int
reqEventTest_str(pTrig, pBuf)
REQ_TRIG *pTrig;
void	*pBuf;
{
    int		change;
    if (pBuf != NULL) {
	if (pTrig->eventNoValYet) {
	    pTrig->eventNoValYet = 0;
		strcpy(PREV.strVal, ((struct dbr_time_string *)pBuf)->value);
	}
	else
	    strcpy(PREV.strVal, CURR.strVal);
	strcpy(CURR.strVal, ((struct dbr_time_string *)pBuf)->value);
	pTrig->dataTs = ((struct dbr_time_string *)pBuf)->stamp;
    }
    if (pTrig->eventNoValYet)
	return -2;
    if (COND == REQ_EV_CHANGE) {
	if (strcmp(CURR.strVal, PREV.strVal) == 0)	change = 0;
	else						change = 1;
    }
    else {
	if ((change = strcmp(CURR.strVal, CMP.strVal)) == 0)
	    ;	/* "change" already set */
	else if (change < 0)				change = -1;
	else						change = 1;
    }
    return change;
}

/*+/internal******************************************************************
* NAME	reqStatConn - connection handler for status channels
*
* DESCRIPTION
*	This routine handles connect and disconnect for status channels.
*	When a connection is made, and if there is a message stored in the
*	structure, this routine sends it to the channel.
*
*	(No action is taken for disconnects.)
*
*-*/
static void
reqStatConn(arg)
struct connection_handler_args arg;
{
    chid	pCh=arg.chid;
    REQ_STAT	*pStat=(REQ_STAT *)ca_puser(pCh);
    int		stat;

    if (arg.op == CA_OP_CONN_UP) {
	if (pStat->msg[0] != '\0') {
	    ca_put(DBR_STRING, pCh, pStat->msg);
#if 0
	    ca_flush_io();
#endif
	    
	}
    }
}
/*+/subr**********************************************************************
* NAME	reqStatInit - initialize a `status' structure for operations
*
* DESCRIPTION
*	Initialize a status channel.  The intent is that the status
*	channel will be an EPICS state record.  If the name in the
*	status structure is empty, no initialization is done.
*
*-*/
void
reqStatInit(pStat)
REQ_STAT *pStat;	/* I pointer to status structure */
{
    int	stat;

    if (pStat->name[0] != '\0') {
	stat = ca_build_and_connect(pStat->name, TYPENOTCONN, 0,
			&pStat->pCh, NULL, reqStatConn, pStat);
	assertAlways(stat == ECA_NORMAL);
#if 0
	ca_flush_io();
#endif
    }
    else
	pStat->pCh = NULL;
    pStat->msg[0] = '\0';
}

#if 0
/*+/macros********************************************************************
* NAME  reqStatPrintf - do a printf to a status channel
*
* DESCRIPTION
*	Print a message to a status channel.  If there isn't a channel
*	for the structure no print is done.  If there is a channel, the
*	message is sent immediately if the channel is connected, or else
*	sending the message is deferred until the channel connects.
*
*	Since state records have a relatively small size, care must
*	be taken that the format produces a relatively small text string.
*
* SYNOPSIS
*       void
*       reqStatPrintf(pStat, fmt, ...)
*       REQ_STAT *pStat;/* I pointer to status structure */
*       char    *fmt;   /* I printf format string */
*       ...             /* I other arguments, for printing */
*
*-*/
#endif
void reqStatPrintf(va_alist)
va_dcl
{
    va_list     pArgs;
    REQ_STAT	*pStat;
    char        *fmt;
    char        message[500];

    va_start(pArgs);
    pStat = va_arg(pArgs, REQ_STAT *);
    fmt = va_arg(pArgs, char *);
    vsprintf(message, fmt, pArgs);
    va_end(pArgs);
    if (pStat->name[0] == '\0')
	return;
    if (strlen(message) > 20)
	message[20] = '\0';
    strcpy(pStat->msg, message);
    if (pStat->pCh != NULL && ca_field_type(pStat->pCh) >= 0) {
	ca_put(DBR_STRING, pStat->pCh, message);
#if 0
	ca_flush_io();
#endif
    }
}

/*+/subr**********************************************************************
* NAME	reqStatWrap - wrap up a status structure when operations are complete
*
*-*/
void
reqStatWrap(pStat)
REQ_STAT *pStat;	/* I pointer to status structure */
{
    if (pStat->pCh != NULL)
	ca_clear_channel(pStat->pCh);
}


/*+/internal******************************************************************
* NAME	reqStepConn - connection handler for step channels
*
*
*-*/
static void
reqStepConn(arg)
struct connection_handler_args arg;
{
    chid	pCh=arg.chid;
    REQ_STEP	*pStep=(REQ_STEP *)ca_puser(pCh);
    chtype	type=dbf_type_to_DBR_TIME(ca_field_type(pCh));
    int		stat;

    if (arg.op == CA_OP_CONN_UP) {
	;	/* no action */
    }
}

/*+/internal******************************************************************
* NAME	reqStepGet - handler for ca_get_callback for a step
*
* DESCRIPTION
*	Mark `getDone'.  If a delay is specified, start it, and mark
*	`delayPending'.
*
*-*/
static void
reqStepGet(arg)
struct event_handler_args arg;
{
    REQ_STEP	*pStep=(REQ_STEP *)arg.usr;

    if (pStep->getPending == 1) {
	pStep->getDone = 1;
	pStep->getPending = 0;
	if (pStep->delay > 0.) {
	    tsLocalTime(&pStep->delayTs);
	    tsAddDouble(&pStep->delayTs, &pStep->delayTs, pStep->delay);
	    pStep->delayPending = 1;
	}
    }
}

/*+/subr**********************************************************************
* NAME	reqStepInit - initialize a `step' structure for operations
*
*-*/
void
reqStepInit(pStep)
REQ_STEP *pStep;	/* I pointer to step structure */
{
    int	stat;

    pStep->delayPending = pStep->delayDone = 0;
    pStep->getPending = pStep->getDone = 0;
    pStep->samplePending = pStep->sampleActive = pStep->sampleDone = 0;
    pStep->active = pStep->complete = 0;
    if (pStep->type == REQ_TYPE_STEP) {
	stat = ca_build_and_connect(pStep->name, TYPENOTCONN, 0,
			&pStep->pCh, NULL, reqStepConn, pStep);
	assertAlways(stat == ECA_NORMAL);
#if 0
	ca_flush_io();
#endif
    }
}

/*+/subr**********************************************************************
* NAME	reqStepWrap - wrap up a step structure when operations are complete
*
*-*/
void
reqStepWrap(pStep)
REQ_STEP *pStep;	/* I pointer to step structure */
{
    if (pStep->pCh != NULL)
	ca_clear_channel(pStep->pCh);
}

/*+/subr**********************************************************************
* NAME	reqStep_seqStart - start a step sequence
*
*-*/
void
reqStep_seqStart(pStep)
REQ_STEP *pStep;	/* I pointer to step structure */
{
    pStep->present = pStep->beg - pStep->step;
    pStep->active = 1;
    pStep->complete = 0;
    pStep->delayPending = pStep->delayDone = 0;
    pStep->getPending = pStep->getDone = 0;
    pStep->samplePending = pStep->sampleActive = pStep->sampleDone = 0;
    reqStep_stepNext(pStep);
}

/*+/subr**********************************************************************
* NAME	reqStep_stepDoneCheck - check to see if a step is done
*
*-*/
int
reqStep_stepDoneCheck(pStep, pNowTs)
REQ_STEP *pStep;	/* I pointer to step structure */
TS_STAMP *pNowTs;	/* I pointer to present time, or NULL */
{
    TS_STAMP	nowTs;

    if (pStep->getPending)
	return 0;
    if (pStep->delayPending) {
	if (pNowTs == NULL)
	    tsLocalTime(&nowTs);
	else
	    nowTs = *pNowTs;
	if (TsCmpStampsLT(&nowTs, &pStep->delayTs))
	    return 0;
	pStep->delayPending = 0;
	pStep->delayDone = 1;
    }
    if (pStep->samplePending || pStep->sampleActive)
	return 0;
    return 1;
}

/*+/subr**********************************************************************
* NAME	reqStep_seqDoneCheck - check to see if the sequence of steps is done
*
* DESCRIPTION
*	Checks to see if a sequence of steps is complete.  The `complete'
*	flag must be set, and no `pending' flags can be set.
*
*-*/
int
reqStep_seqDoneCheck(pStep)
REQ_STEP *pStep;	/* I pointer to step structure */
{
    if (pStep->complete == 0)
	return 0;
    if (pStep->getPending || pStep->delayPending)
	return 0;
    if (pStep->samplePending || pStep->sampleActive)
	return 0;
    pStep->active = 0;
    return 1;
}

/*+/subr**********************************************************************
* NAME	reqStep_stepNext - do the next step in a sequence
*
* DESCRIPTION
*	Sends out the next value in the sequence.  If this is the last one,
*	`complete' is set.
*
*	After the value is sent, `getPending' is set.
*
*-*/
void
reqStep_stepNext(pStep)
REQ_STEP *pStep;	/* I pointer to step structure */
{
    double	val;
    long	stat;

    val = pStep->present + pStep->step;
    if (pStep->step > 0.) {
	if (val > pStep->end) 
	    val = pStep->end;
    }
    else {
	if (val < pStep->end)
	    val = pStep->end;
    }
    if (val == pStep->end)
	pStep->complete = 1;
    pStep->present = val;
    pStep->getDone = pStep->delayDone = pStep->delayPending = 0;
    pStep->getPending = 1;
    stat = ca_put(DBR_DOUBLE, pStep->pCh, &pStep->present);
    assertAlways(stat == ECA_NORMAL);
    stat = ca_array_get_callback(DBR_DOUBLE, 1, pStep->pCh, reqStepGet, pStep);
    assertAlways(stat == ECA_NORMAL);
#if 0
    ca_flush_io();
#endif
}

/*+/subr**********************************************************************
* NAME	reqTrigArm - arm a trigger condition
*
* DESCRIPTION
*	Arms a trigger condition, so that testing it can occur.  When it is
*	created, a trigger condition is in the disarmed state.
*
*	Arming a trigger does NOT affect any of the timing information,
*	including the "at least every" timer.  Thus, if the trigger has
*	been disarmed for 30 seconds and the condition is a 10 second
*	period, then calling reqTrigTest immediately following reqTrigArm
*	will result in a true return.
*
*	A trigger type of REQ_TYPE_TIMED is a slight exception to the
*	above.  When armed, the "next time" is set to be the current time
*	plus the current time.
*
*	Arming a REQ_TYPE_EVENT trigger sets up for detecting a transition
*	from false to true for the trigger condition.  Thus, if the
*	condition is true at the time reqTrigArm is called, reqTrigTest
*	will return false until the condition has gone false and then
*	true again.
*
* RETURNS
*	void
*
*-*/
void
reqTrigArm(pTrig, nowTs)
REQ_TRIG *pTrig;	/* I pointer to trigger structure */
TS_STAMP nowTs;		/* I current time */
{
    pTrig->everyCount = 0;
    pTrig->eventHappened = 0;
    pTrig->notYetPrimed = 1;
    if (pTrig->type != REQ_TYPE_TIMED && pTrig->type != REQ_TYPE_PERIODIC)
	pTrig->notYetTested = 1;
    pTrig->everyPrimed = 0;
    if (pTrig->type == REQ_TYPE_TIMED)
	TsAddStamps(&pTrig->timeNext, &nowTs, &pTrig->timeStamp);
    if (pTrig->atleastType == REQ_TYPE_TIMED)
	TsAddStamps(&pTrig->atleastNext, &nowTs, &pTrig->atleastStamp);
    pTrig->armed = 1;
    pTrig->demand = 0;
    pTrig->doneFlag = 0;
    if (pTrig->type == REQ_TYPE_EVENT)
	reqEventTest(pTrig, NULL);
}

/*+/subr**********************************************************************
* NAME	reqTrigInit - initialize a trigger for operations
*
*-*/
void
reqTrigInit(pTrig)
REQ_TRIG *pTrig;	/* I pointer to trigger structure */
{
    pTrig->notYetInit = 1;
    pTrig->armed = 0;
    pTrig->demand = 0;
    pTrig->doneFlag = 0;
    pTrig->everyCount = 0;
    pTrig->eventHappened = 0;
    pTrig->eventNoValYet = 1;
    pTrig->notYetTested = 1;
    pTrig->notYetPrimed = 1;
    pTrig->enumGetGR = 0;
    pTrig->timeNext.secPastEpoch = pTrig->timeNext.nsec = 0;
    pTrig->atleastNext.secPastEpoch = pTrig->atleastNext.nsec = 0;
    if (pTrig->eventCond != REQ_EV_UNDEF) {
	int	stat;
	if (pTrig->event_pCh == NULL) {
	    stat = ca_build_and_connect(pTrig->eventName, TYPENOTCONN, 0,
			&pTrig->event_pCh, NULL, reqEventConn, pTrig);
	    assertAlways(stat == ECA_NORMAL);
#if 0
	    ca_flush_io();
	    ca_pend_event(.05);
#else
#endif
	}
    }
    else {
	pTrig->notYetInit = 0;
	pTrig->event_pCh = NULL;
    }
}

/*+/subr**********************************************************************
* NAME	reqTrigTest - test a trigger condition
*
* DESCRIPTION
*	Tests a trigger condition.  If the trigger isn't armed, then the
*	test will always be false.  When a test returns true, the trigger
*	is disarmed by this routine.  The caller is responsible for
*	calling reqTrigArm to rearm the trigger.
*
*	When a trigger is REQ_TYPE_DEMAND and a demand has occurred, this
*	routine updates the "at least every" parameters and returns true.
*	When reqTrigArm is subsequently called to rearm the trigger, the
*	demand flag is reset, so that a new demand will be required to
*	cause a new trigger.
*
*	For periodic triggers, scheduling is independent of how long the
*	action takes.  For example, if the period is 30 seconds and there
*	is an elapsed time between trigger and rearm of 10 seconds, the
*	next trigger will still occur 30 seconds after the prior
*	trigger.  (This can cause problems if the processing time is
*	greater than the period, but the caller is responsible for detecting
*	and coping with such problems.)
*
*	For event triggers, if an event occurs it will always result in
*	at least one trigger--i.e., transient events latch.  The latch
*	is reset when the trigger is tested or armed.
*
*	If the trigger condition has never been tested before, and if
*	the trigger type (in the structure) is one of the following, then
*	the first test will return true:
*		REQ_TYPE_ALWAYS
*		REQ_TYPE_PERIODIC
*
*	The "at least every" timer is updated each time the trigger
*	condition is satisfied.  The "periodic" timer is upated ONLY
*	when the condition is satisfied due to the expiration of the
*	timer.
*
*	A special trigger type of REQ_TYPE_DONE allows a program to
*	use the trigger mechanism for `internal' triggers.  To set the
*	flag, call reqTrig_setDone.
*
* RETURNS
*	1	if trigger condition is satisfied, or
*	0
*
*-*/
int
reqTrigTest(pTrig, nowTs)
REQ_TRIG *pTrig;	/* I pointer to trigger structure */
TS_STAMP nowTs;		/* I current time */
{
    int		doIt=0;		/* trigger condition is T,F for 1,0 */

    if (pTrig->armed == 0)
	return doIt;
    if (pTrig->demand)
	doIt = 1;
    else if (pTrig->type == REQ_TYPE_ALWAYS)
	doIt = 1;
    else if (pTrig->notYetTested && pTrig->type == REQ_TYPE_PERIODIC) {
	doIt = 1;
	TsAddStamps(&pTrig->timeNext, &nowTs, &pTrig->timeStamp);
	pTrig->timeLast = nowTs;
    }
    else if (pTrig->type == REQ_TYPE_DONE) {
	if (pTrig->doneFlag == 1)
	    doIt = 1;
    }
    else if (pTrig->type == REQ_TYPE_EVENT) {
	if (pTrig->eventHappened)	/* flag set by Monitor? */
	    doIt = 1;
	else if (reqEventTest(pTrig, NULL) == 1)
	    doIt = 1;
    }
    else if (pTrig->type == REQ_TYPE_IF) {
	if (reqEventTest(pTrig, NULL) == 1)
	    doIt = 1;
    }
    else if (pTrig->type==REQ_TYPE_PERIODIC || pTrig->type==REQ_TYPE_TIMED) {
	if (TsCmpStampsGE(&nowTs, &pTrig->timeNext)) {
	    doIt = 1;
	    TsAddStamps(&pTrig->timeNext, &pTrig->timeNext, &pTrig->timeStamp);
	    pTrig->timeLast = nowTs;
	}
    }
    if (doIt == 0 && pTrig->atleastType == REQ_TYPE_TIMED) {
	if (TsCmpStampsGE(&nowTs, &pTrig->atleastNext))
	    doIt = 1;
    }
    if (doIt) {
	pTrig->armed = 0;
	if (pTrig->atleastType == REQ_TYPE_TIMED) {
	    pTrig->atleastLast = nowTs;
	    TsAddStamps(&pTrig->atleastNext, &nowTs, &pTrig->atleastStamp);
	}
    }
    pTrig->notYetTested = 0;
    return doIt;
}

/*+/subr**********************************************************************
* NAME	reqTrig_setDone - set the `user process' done flag
*
*-*/
void
reqTrig_setDone(pTrig)
REQ_TRIG *pTrig;	/* I pointer to trigger structure */
{
    pTrig->doneFlag = 1;
}

/*+/subr**********************************************************************
* NAME	reqTrigWrap - wrap up a trigger when operations are complete
*
*-*/
void
reqTrigWrap(pTrig)
REQ_TRIG *pTrig;	/* I pointer to trigger structure */
{
    if (pTrig->event_pCh != NULL)
	ca_clear_channel(pTrig->event_pCh);
}
